Package: cclustr
Title: Consensus Clustering Methods for Multiple Imputed Data
Version: 0.1.2
Authors@R: c(
    person("Anhuar", "Duran Mendoza", email = "asteyduran15@gmail.com", role = "aut"),
    person("Andres", "Montenegro Lemus", email = "andresfemole@gmail.com", role = c("aut", "cre")),
    person("Mario", "Pacheco Lopez", email = "mariopachecolopez@gmail.com", role = "aut")
  )
Description: Provides tools for performing consensus clustering on multiple 
    imputed datasets. The package supports a range of clustering algorithms 
    across imputations, including hierarchical methods (e.g., Ward, single, 
    complete, average) and partition-based approaches such as k-means, 
    k-medoids (PAM), fuzzy clustering, model-based clustering ('mclust'), 
    and methods for mixed or categorical data (k-modes and k-prototypes). 
    A co-assignment matrix is constructed to quantify agreement between 
    partitions, and consensus solutions are derived via hierarchical 
    clustering applied to the resulting dissimilarity matrix. Additional 
    functions are provided for validation and visualization of clustering 
    results, facilitating robust analysis in the presence of missing data.
    Consensus clustering framework is based on Monti et al. (2003) 
    <doi:10.1023/A:1023949509487>, rank aggregation methods follow 
    Pihur et al. (2007) <doi:10.1093/bioinformatics/btm158>, and the 
    PAC (Proportion of Ambiguous Clustering) metric is based on 
    Senbabaoglu et al. (2014) <doi:10.1038/srep06207>.
License: MIT + file LICENSE
URL: https://github.com/andrews06ml/cclustr
BugReports: https://github.com/andrews06ml/cclustr/issues
Depends: R (>= 4.0)
Imports: cluster, e1071, fpc, graphics, stats, viridisLite, klaR,
        clustMixType, proxy, mclust
Suggests: knitr, mice, mlbench (>= 2.1-11), rmarkdown, spelling,
        testthat (>= 3.0.0), utils
Encoding: UTF-8
Language: en-US
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-09-03 01:00:06 UTC; andre
Author: Anhuar Duran Mendoza [aut],
  Andres Montenegro Lemus [aut, cre],
  Mario Pacheco Lopez [aut]
Maintainer: Andres Montenegro Lemus <andresfemole@gmail.com>
Repository: CRAN
Date/Publication: 2026-09-03 02:20:02 UTC
Built: R 4.5.2; ; 2026-09-03 03:10:43 UTC; unix

# Generated by roxygen2: do not edit by hand

export(as_mild_list)
export(choose_best_clustering)
export(cluster_imputations)
export(consensus_clustering)
export(plot_consensus_dendrogram)
export(plot_consensus_matrix)
export(plot_validation_metrics)
export(run_mi_clustering)
export(validate_clustering)

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  warning   = FALSE,
  message   = FALSE
)

## ----load-data----------------------------------------------------------------
library(cclustr)
library(mice)

if (!requireNamespace("mlbench", quietly = TRUE) ||
    packageVersion("mlbench") < "2.1-11") {
  stop("Package 'mlbench' (>= 2.1-11) is required for this vignette. ",
       "Install it with: install.packages('mlbench')")
}

data("SynthDiabetes2", package = "mlbench")

# Remove diabetes from the clustering dataset
df <- SynthDiabetes2[, -9]

cat("Dimensions:", nrow(df), "x", ncol(df), "\n")
cat("\nMissing values per variable:\n")
print(colSums(is.na(df)))
cat("\nTotal missing values:", sum(is.na(df)),
    "(", round(mean(is.na(df)) * 100, 1), "%)\n")

## ----imputation---------------------------------------------------------------

imp <- mice(df, m = 20, maxit = 20, method = "pmm", printFlag = FALSE, seed = 202604)

# Convert to cclustr standard format
mi_synth <- as_mild_list(imp)

cat("Number of imputed datasets :", length(mi_synth), "\n")
cat("Dimensions of each dataset :", nrow(mi_synth[[1]]),
    "x", ncol(mi_synth[[1]]), "\n")
cat("Any missing values remaining:", any(sapply(mi_synth, anyNA)), "\n")

## ----clustering---------------------------------------------------------------
set.seed(202604)
partitions <- cluster_imputations(
  imp_list   = mi_synth,
  method     = "kmeans",
  k          = 2:4,
  scale_data = "global"
)

cat("Available k values:", names(partitions), "\n")
cat("Number of partitions per k:", length(partitions$k2), "\n")

## ----consensus----------------------------------------------------------------
consensus_results <- consensus_clustering(
  partitions     = partitions,
  cluster_method = "ward.D2",
  consensus_method = "classic"
)

cat("Consensus solutions available:", names(consensus_results), "\n")

## ----validation---------------------------------------------------------------
val_table <- validate_clustering(
  partitions        = partitions,
  consensus_results = consensus_results
)

print(val_table)

## ----selection----------------------------------------------------------------
best <- choose_best_clustering(
  validation_table  = val_table,
  consensus_results = consensus_results,
  prefer_stability  = TRUE,
  tie_breaker       = "silhouette"
)

cat("Optimal k selected:", best$best_k, "\n")
cat("\nConsensus cluster sizes:\n")
print(table(best$best_consensus))
cat("\nWeighted rank scores:\n")
print(best$scores_table[, c("k", "pac", "silhouette_mean",
                             "ari_mean_between_imputations", "score")])

## ----validation-plot, fig.cap = "Validation metrics across k = 2, 3, 4. The optimal value per metric is highlighted in red."----
plot_validation_metrics(best)

## ----heatmap, fig.cap = "Co-assignment matrix for the optimal k. A clear block-diagonal structure indicates a stable consensus."----
plot_consensus_matrix(
  consensus_result = best,
  reorder          = TRUE,
  viridis_option   = "D",
  show_labels = FALSE
)

## ----dendrogram, fig.cap = "Consensus dendrogram for the optimal k, derived from the dissimilarity matrix 1 - C."----
plot_consensus_dendrogram(
  consensus_result = best,
  rect             = TRUE,
  labels           = FALSE
)

## ----cluster-summary----------------------------------------------------------
synth_clustered          <- mi_synth[[1]]
synth_clustered$cluster  <- factor(best$best_consensus)

# Mean values per cluster
num_vars <- c("pregnant", "glucose", "pressure",
              "triceps", "insulin", "mass", "pedigree", "age")

cat("Mean clinical values by cluster:\n")
print(
  aggregate(
    synth_clustered[, num_vars],
    by  = list(Cluster = synth_clustered$cluster),
    FUN = function(x) round(mean(x, na.rm = TRUE), 2)
  )
)

## ----boxplots, fig.cap = "Distribution of glucose and BMI by consensus cluster."----
par(mfrow = c(1, 3), mar = c(4, 4, 3, 1))

boxplot(
  glucose ~ cluster,
  data = synth_clustered,
  main = "Plasma Glucose by Cluster",
  xlab = "Cluster",
  ylab = "Glucose (mg/dl)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)

boxplot(
  mass ~ cluster,
  data = synth_clustered,
  main = "Body Mass Index by Cluster",
  xlab = "Cluster",
  ylab = "BMI (kg/m²)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)

boxplot(
  insulin ~ cluster,
  data = synth_clustered,
  main = "Insulin",
  xlab = "Cluster",
  ylab = "Insulin (mu U/ml)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)
par(mfrow = c(1, 1))

## ----full-pipeline------------------------------------------------------------
set.seed(202604)

result <- run_mi_clustering(
  data             = imp,
  method           = "kmeans",
  k                = 2:4,
  scale_data       = "global",
  consensus_method = "classic",
  prefer_stability = TRUE,
  tie_breaker      = "silhouette"
)

cat("Best k selected:", result$best_k, "\n")
cat("\nCluster sizes:\n")
print(table(result$best_consensus))

## ----weighted-ari, eval = requireNamespace("mclust", quietly = TRUE)----------
set.seed(202604)

result_w <- run_mi_clustering(
  data             = imp,
  method           = "kmeans",
  k                = 2:4,
  scale_data       = "global",
  consensus_method = "weighted_ari",
  prefer_stability = TRUE,
  tie_breaker      = "silhouette"
)

cat("Best k (ARI-weighted consensus):", result_w$best_k, "\n")
print(table(result_w$best_consensus))

#' Perform clustering on multiple imputed datasets
#'
#' @description
#' Applies a clustering algorithm to each completed dataset in a standardized
#' imputation list (as produced by \code{\link{as_mild_list}}). Supports
#' hierarchical clustering, k-means, PAM, fuzzy c-means, model-based clustering,
#' k-modes (for categorical data), and k-prototypes (for mixed data).
#' Accepts a single value of \code{k} or a range, optionally scales the data
#' prior to clustering, and supports multiple distance metrics including
#' Gower, Jaccard, Simple Matching Coefficient, and user-defined distances.
#'
#' @param imp_list A named list of \code{data.frame} objects, as returned by
#'   \code{\link{as_mild_list}}. All datasets must have identical dimensions
#'   and column names.
#'
#' @param method A character string specifying the clustering algorithm.
#'   For hierarchical clustering, accepted values are \code{"ward.D"},
#'   \code{"ward.D2"}, \code{"single"}, \code{"complete"}, \code{"average"},
#'   \code{"centroid"}, \code{"median"}, and \code{"mcquitty"}.
#'   Additional options are:
#'   \itemize{
#'     \item \code{"kmeans"}: k-means clustering for numeric data.
#'     \item \code{"pam"}: partitioning around medoids; supports mixed data
#'           when used with \code{distance = "gower"}.
#'     \item \code{"fuzzy"}: fuzzy c-means clustering.
#'     \item \code{"mclust"}: model-based clustering via Gaussian mixtures.
#'     \item \code{"kmodes"}: clustering for purely categorical data.
#'     \item \code{"kprototypes"}: clustering for mixed data (numeric and
#'           categorical).
#'   }
#'   Default is \code{"ward.D2"}.
#'
#' @param k A single integer or an integer vector specifying the number of
#'   clusters. If a vector is provided, clustering is performed for each
#'   value of \code{k}.
#'
#' @param scale_data Character string specifying the scaling strategy for
#'   continuous numeric variables (binary variables coded as 0/1 are
#'   automatically excluded from scaling). Options are:
#'   \itemize{
#'     \item \code{"global"}: (default) variables are scaled using pooled
#'           mean and standard deviation computed across all imputations
#'           without stacking them in memory (combinatorial variance formula).
#'           Ensures comparability of distance scales between datasets and
#'           is recommended for consensus clustering.
#'     \item \code{"within"}: each imputed dataset is scaled independently
#'           using its own mean and standard deviation. Useful for
#'           exploratory analyses but may introduce inconsistencies in
#'           distance scales across datasets.
#'     \item \code{"none"}: no scaling is applied.
#'   }
#'   Scaling is automatically disabled for \code{distance = "gower"},
#'   \code{"jaccard"}, or \code{"simple_matching"}, and for
#'   \code{method = "kmodes"} and \code{"kprototypes"}, as these handle
#'   variable scales internally.
#'
#' @param distance A character string specifying the distance metric used for
#'   hierarchical clustering and \code{method = "pam"}. Options are:
#'   \itemize{
#'     \item \code{"euclidean"}: (default) standard Euclidean distance.
#'           Suitable for continuous numeric data after scaling.
#'     \item \code{"manhattan"}: sum of absolute differences. Suitable for
#'           continuous numeric data; more robust to outliers than Euclidean.
#'     \item \code{"gower"}: Gower distance via \code{cluster::daisy()}.
#'           Supports mixed data types (numeric, categorical, binary).
#'           When used with hierarchical clustering, numeric variables are
#'           normalized using global min/max computed across all imputations,
#'           ensuring cross-imputation comparability. Requires package
#'           \pkg{cluster}.
#'     \item \code{"jaccard"}: Jaccard dissimilarity. Requires all variables
#'           to be binary (0/1 or two-level factor). Requires package
#'           \pkg{proxy}.
#'     \item \code{"simple_matching"}: Simple Matching Coefficient (SMC).
#'           Requires all variables to be binary. Requires package
#'           \pkg{proxy}.
#'     \item \code{"custom"}: user-defined distance function supplied via
#'           \code{dist_fun}. Must return an object of class \code{dist}.
#'   }
#'   Ignored for \code{method = "kmeans"}, \code{"fuzzy"}, \code{"mclust"},
#'   \code{"kmodes"}, and \code{"kprototypes"}, which compute distances
#'   internally.
#'
#'   For \code{method = "pam"}, all distance options are fully supported
#'   as PAM operates on a precomputed dissimilarity matrix.
#'
#' @param dist_fun A function that takes a \code{data.frame} and returns a
#'   \code{dist} object. Required when \code{distance = "custom"};
#'   ignored otherwise.
#'
#' @param dist_args A named list of additional arguments passed to the distance
#'   function. For \code{distance = "gower"}, arguments are forwarded to
#'   \code{cluster::daisy()}; for \code{distance = "jaccard"} or
#'   \code{"simple_matching"}, to \code{proxy::dist()}; for
#'   \code{distance = "euclidean"} or \code{"manhattan"}, to
#'   \code{stats::dist()}; also for \code{distance = "custom"}. Default is \code{list()}.
#'
#' @param ... Additional arguments passed to the underlying clustering
#'   function: \code{stats::kmeans}, \code{e1071::cmeans},
#'   \code{mclust::Mclust}, \code{klaR::kmodes}, or
#'   \code{clustMixType::kproto}. Not used for \code{method = "pam"}, which
#'   operates on a precomputed dissimilarity matrix.
#'
#' @return
#' \itemize{
#'   \item If \code{k} is a single value: a named list of integer vectors,
#'         one per imputed dataset, containing the cluster assignment for
#'         each observation.
#'   \item If \code{k} is a vector: a named list of lists, where each
#'         top-level element corresponds to one value of \code{k} (named
#'         \code{"k2"}, \code{"k3"}, etc.), and each inner element is a
#'         named list of cluster assignment vectors, one per imputed dataset.
#' }
#'
#' @details
#' \strong{Scaling strategy and cross-imputation comparability:}
#'
#' For methods that rely on Euclidean or Manhattan distances
#' (\code{"kmeans"}, \code{"fuzzy"}, \code{"mclust"}, hierarchical with
#' \code{distance = "euclidean"} or \code{"manhattan"}), \code{scale_data =
#' "global"} is strongly recommended. It computes the pooled mean and
#' standard deviation across all imputations using the combinatorial variance
#' formula — without creating a stacked copy of the data in memory — and
#' applies the same transformation to every imputation. This ensures that
#' distances are on the same scale across datasets, which is essential for
#' a meaningful co-assignment matrix.
#'
#' Binary variables coded as 0/1 numeric are automatically excluded from
#' scaling because standardizing them loses their binary semantics.
#'
#' \strong{Gower distance and global normalization:}
#'
#' Gower distance normalizes each numeric variable by its range. When
#' computed independently per imputation, different imputations may produce
#' slightly different ranges, leading to incomparable distance scales. When
#' \code{distance = "gower"} is used with hierarchical clustering or PAM,
#' the function pre-normalizes numeric variables using global min/max
#' computed across all imputations (without stacking), so that Gower
#' distances are on a consistent scale across datasets. Categorical and
#' factor variables are left untouched and handled normally by
#' \code{cluster::daisy()}. Additional arguments to \code{cluster::daisy()}
#' can be passed via \code{dist_args}.
#'
#' \strong{Hierarchical clustering optimization:}
#'
#' When \code{method} is hierarchical and \code{k} is a vector, the
#' distance matrix and dendrogram are computed only once per imputed dataset
#' and then cut at each requested \code{k}, which avoids redundant
#' computation.
#'
#' \strong{Data type compatibility:}
#' \itemize{
#'   \item Numeric-only: hierarchical, \code{"kmeans"}, \code{"fuzzy"},
#'         \code{"mclust"}.
#'   \item Categorical-only: \code{"kmodes"}.
#'   \item Mixed (numeric + categorical): \code{"pam"} with
#'         \code{distance = "gower"} or \code{"kprototypes"}.
#'   \item Binary-only: hierarchical with \code{distance = "jaccard"} or
#'         \code{"simple_matching"}.
#' }
#'
#' Character variables are internally converted to factors.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Numeric data, hierarchical clustering
#' # ------------------------------------------------------------
#' set.seed(123)
#' imp_list <- replicate(3, {
#'   data.frame(x = rnorm(20), y = rnorm(20))
#' }, simplify = FALSE)
#'
#' # Single k, global scaling, euclidean distance (default)
#' res <- cluster_imputations(imp_list, method = "ward.D2", k = 2)
#' str(res)
#'
#' # Range of k, manhattan distance
#' res_multi <- cluster_imputations(imp_list, method = "ward.D2",
#'                                  k = 2:4, distance = "manhattan")
#' str(res_multi)
#'
#' # ------------------------------------------------------------
#' # Example 2: Mixed data, PAM with Gower distance
#' # ------------------------------------------------------------
#' set.seed(123)
#' imp_mixed <- replicate(3, {
#'   data.frame(
#'     x    = rnorm(20),
#'     grup = factor(sample(letters[1:3], 20, replace = TRUE))
#'   )
#' }, simplify = FALSE)
#'
#' res_pam <- cluster_imputations(imp_mixed, method = "pam",
#'                                k = 2, distance = "gower")
#' str(res_pam)
#'
#' # ------------------------------------------------------------
#' # Example 3: Binary data, Jaccard distance
#' # ------------------------------------------------------------
#' set.seed(123)
#' imp_bin <- replicate(3, {
#'   data.frame(
#'     a = sample(0:1, 20, replace = TRUE),
#'     b = sample(0:1, 20, replace = TRUE),
#'     c = sample(0:1, 20, replace = TRUE)
#'   )
#' }, simplify = FALSE)
#'
#' \donttest{
#' if (requireNamespace("proxy", quietly = TRUE)) {
#'   res_jac <- cluster_imputations(imp_bin, method = "ward.D2",
#'                                  k = 2, distance = "jaccard")
#'   str(res_jac)
#' }
#'
#' # ------------------------------------------------------------
#' # Example 4: Full pipeline with mice
#' # ------------------------------------------------------------
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#'   df_pre       <- mice::nhanes
#'   df_pre$age   <- factor(df_pre$age)
#'   df_pre$hyp   <- factor(df_pre$hyp)
#'
#'   imp  <- mice::mice(df_pre, m = 3, printFlag = FALSE, seed = 123)
#'   mild <- as_mild_list(imp)
#'
#'   # PAM + Gower with global range normalization
#'   parts <- cluster_imputations(mild, method = "pam",
#'                                k = 2:4, distance = "gower")
#'   str(parts)
#' }
#' }
#'
#' @references
#' Kaufman, L., & Rousseeuw, P. J. (1990).
#' \emph{Finding Groups in Data: An Introduction to Cluster Analysis}. Wiley.
#' \doi{10.1002/9780470316801}
#'
#' Bezdek, J. C. (1981).
#' \emph{Pattern Recognition with Fuzzy Objective Function Algorithms}.
#' Plenum Press.
#'
#' Huang, Z. (1998).
#' Extensions to the k-means algorithm for clustering large data sets
#' with categorical values.
#' \emph{Data Mining and Knowledge Discovery}, \strong{2}(3), 283-304.
#' \doi{10.1023/A:1009769707641}
#'
#' Gower, J. C. (1971).
#' A general coefficient of similarity and some of its properties.
#' \emph{Biometrics}, \strong{27}(4), 857-871.
#' \doi{10.2307/2528823}
#'
#' Scrucca, L., Fop, M., Murphy, T. B., & Raftery, A. E. (2016).
#' mclust 5: Clustering, classification and density estimation using
#' Gaussian finite mixture models.
#' \emph{The R Journal}, \strong{8}(1), 289-317.
#' \doi{10.32614/RJ-2016-021}
#'
#' @seealso \code{\link{as_mild_list}}, \code{\link{consensus_clustering}}
#'
#' @export
cluster_imputations <- function(imp_list,
                                method = "ward.D2",
                                k,
                                scale_data = c("global", "within", "none"),
                                distance = c("euclidean", "manhattan", "gower",
                                             "jaccard", "simple_matching", "custom"),
                                dist_fun = NULL,
                                dist_args = list(), # For distance
                                ...) {    # For clustering function

  scale_data <- match.arg(scale_data)
  distance   <- match.arg(distance)
  cluster_args <- list(...)

  # --------------------------------------------------------
  # Supported methods
  # --------------------------------------------------------
  hierarchical_methods <- c(
    "ward.D", "ward.D2", "single", "complete",
    "average", "centroid", "median", "mcquitty"
  )

  vector_methods <- c("kmeans", "fuzzy", "mclust")
  categorical_methods <- c("kmodes")
  mixed_methods <- c("pam", "kprototypes")

  if (!(method %in% c(hierarchical_methods, vector_methods,
                      categorical_methods, mixed_methods))) {
    stop("Unsupported clustering method.")
  }

  # --------------------------------------------------------
  # Detect variable types
  # --------------------------------------------------------

  imp_list <- lapply(imp_list, function(df) {
    df[] <- lapply(df, function(col) {
      if (is.character(col)) as.factor(col) else col
    })
    df
  })

  .detect_types <- function(df) {
    list(
      numeric = vapply(df, is.numeric, logical(1)),
      factor  = vapply(df, is.factor, logical(1)),
      binary  = vapply(df, function(x) length(unique(stats::na.omit(x))) <= 2, logical(1))
    )
  }

  types <- .detect_types(imp_list[[1]])
  has_factors <- any(types$factor)
  all_factors <- all(types$factor)
  all_binary  <- all(types$binary)

  # --------------------------------------------------------
  # Validations
  # --------------------------------------------------------
  if (method %in% vector_methods && has_factors) {
    stop("The dataset contains categorical variables but method = '", method, "' requires numeric data only. ")
  }

  if (distance %in% c("jaccard", "simple_matching") && !all_binary) {
    stop(
      "distance = '", distance, "' requires all variables to be binary ",
      "(numeric 0/1 or two-level factor)."
    )
  }

  if (method %in% hierarchical_methods) {
    if (has_factors && distance %in% c("euclidean", "manhattan")) {
      stop("Categorical data requires distance = 'gower', 'jaccard', or 'custom'.")
    }
  }

  if (all_factors && method == "kprototypes") {
    warning("All variables are categorical. Switching method to 'kmodes'.")
    method <- "kmodes"
  }

  if (method == "kmodes" && !all_factors) {
    stop("kmodes requires categorical data.")
  }

  if (method == "pam" && has_factors && distance != "gower") {
    warning("Mixed data detected. Consider distance = 'gower'.")
  }

  if (any(k < 2)) stop("k must be >= 2.")
  n_obs <- nrow(imp_list[[1]])
  if (any(k >= n_obs)) stop("k must be < number of observations.")

  if (method == "kprototypes" && distance != "euclidean") {
    warning("kprototypes handles distance internally. The 'distance' parameter is ignored.")
  }

  if (method %in% vector_methods && distance != "euclidean") {
    warning(paste0("The 'distance' parameter is ignored for method = '", method, "'."))
  }

  if (distance == "custom" && is.null(dist_fun)) {
    stop("distance = 'custom' requires a function supplied via dist_fun.")
  }

  # --------------------------------------------------------
  # Disable scaling when not applicable
  # --------------------------------------------------------
  if (distance %in% c("gower", "jaccard", "simple_matching")) {
    if (scale_data != "none") {
      if (distance == "gower") {
        warning(
          "The 'scale_data' parameter is ignored for distance = 'gower'. ",
          "Numeric variables are instead normalized using global min/max ",
          "computed across all imputations to ensure cross-imputation ",
          "comparability before computing Gower distances."
        )
      } else {
        warning(
          "The 'scale_data' parameter is ignored for distance = '", distance, "'. ",
          "This metric handles normalization internally."
        )
      }
      scale_data <- "none"
    }
  }

  if (method %in% categorical_methods && scale_data != "none") {
    warning("Scaling disabled for categorical-only method.")
    scale_data <- "none"
  }

  # kprototypes: handles numeric/categorical balance internally
  if (method == "kprototypes" && scale_data != "none") {
    warning(
      "Scaling disabled for method = 'kprototypes'. ",
      "kprototypes handles variable scales internally. ",
      "Variables with larger numeric ranges may still dominate; ",
      "consider pre-scaling manually if needed."
    )
    scale_data <- "none"
  }

  # --------------------------------------------------------
  # Scaling (only numeric)
  # --------------------------------------------------------

  num_cols <- types$numeric & !types$binary

  if (!any(num_cols)) {
    database <- imp_list
  } else if (scale_data == "none") {
    database <- imp_list
  } else if (scale_data == "within") {

    database <- lapply(imp_list, function(df) {
      df[, num_cols] <- scale(df[, num_cols, drop = FALSE])
      df
    })

  } else if (scale_data == "global") {

    # Pooled mean and SD without stacking (combinatorial variance formula)
    # This avoids creating a large stacked object in memory.
    imp_ns    <- vapply(imp_list, nrow, integer(1))
    imp_means <- lapply(imp_list,
                        function(df) colMeans(df[, num_cols, drop = FALSE]))
    imp_vars  <- lapply(imp_list,
                        function(df) apply(df[, num_cols, drop = FALSE], 2,
                                           stats::var))

    # Grand (pooled) mean: weighted average of per-imputation means
    grand_mean <- Reduce("+",
                         mapply(function(mu, ni) mu * ni,
                                imp_means, imp_ns, SIMPLIFY = FALSE)
    ) / sum(imp_ns)

    # Pooled variance: between- + within-imputation components
    pool_var <- Reduce("+",
                       mapply(function(v, mu, ni) {
                         (ni - 1) * v + ni * (mu - grand_mean) ^ 2
                       }, imp_vars, imp_means, imp_ns, SIMPLIFY = FALSE)
    ) / (sum(imp_ns) - 1)

    grand_sd          <- sqrt(pool_var)
    grand_sd[grand_sd == 0] <- 1   # avoid division by zero

    database <- lapply(imp_list, function(df) {
      df[, num_cols] <- sweep(df[, num_cols, drop = FALSE], 2, grand_mean, "-")
      df[, num_cols] <- sweep(df[, num_cols, drop = FALSE], 2, grand_sd,   "/")
      df
    })
  } else {
    database <- imp_list
  }

  # --------------------------------------------------------
  # Global range normalization for Gower distance
  # Pre-normalize numeric variables using pooled min/max so
  # that Gower distances are comparable across imputations.
  # daisy() will then see numeric variables already in [0, 1]
  # and compute its own range as ~1, leaving the scale intact.
  # Factor/categorical columns are not touched.
  # --------------------------------------------------------

  if (distance == "gower") {

    num_cols_gwr <- types$numeric & !types$binary

    if (any(num_cols_gwr)) {

      # Global min and max without stacking: element-wise pmin/pmax
      global_min <- do.call(pmin,
                            lapply(database,
                                   function(df) apply(df[, num_cols_gwr, drop = FALSE],
                                                      2, min, na.rm = TRUE)))
      global_max <- do.call(pmax,
                            lapply(database,
                                   function(df) apply(df[, num_cols_gwr, drop = FALSE],
                                                      2, max, na.rm = TRUE)))

      # Apply global range normalization to each dataset
      database <- lapply(database, function(df) {
        for (col in names(which(num_cols_gwr))) {
          rng <- global_max[col] - global_min[col]
          if (rng == 0) rng <- 1
          df[[col]] <- (df[[col]] - global_min[col]) / rng
        }
        df
      })
    }
  }

  # --------------------------------------------------------
  # Distance computation
  # --------------------------------------------------------
  .compute_distance <- function(df) {

    if (distance == "gower") {
      if (!requireNamespace("cluster", quietly = TRUE)) {
        stop("Package 'cluster' is required for Gower distance.")
      }
      return(do.call(cluster::daisy, c(list(x = df,
                                            metric = "gower"), dist_args)
      ))
    }

    if (distance == "jaccard") {
      if (!all_binary) stop("Jaccard requires binary variables.")
      if (!requireNamespace("proxy", quietly = TRUE)) {
        stop("Package 'proxy' is required.")
      }
      return(do.call(proxy::dist,
                     c(list(x = df,
                            method = "Jaccard"), dist_args)
      ))
    }

    if (distance == "simple_matching") {
      if (!all_binary) stop("simple matching requires binary variables.")
      if (!requireNamespace("proxy", quietly = TRUE)) {
        stop("Package 'proxy' is required.")
      }
      return(do.call(proxy::dist,
                     c(list(x = df,
                            method = "simple matching"), dist_args)
      ))
    }

    if (distance == "custom") {
      result <- do.call(dist_fun, c(list(df), dist_args))
      if (!inherits(result, "dist")) {
        stop("dist_fun must return an object of class 'dist'.")
      }
      return(result)
    }

    # euclidean / manhattan
    do.call(stats::dist,
            c(list(x = df,
                   method = distance), dist_args))
  }

  # --------------------------------------------------------
  # Clustering
  # --------------------------------------------------------
  .run_clustering <- function(df, k_i) {

    if (method %in% hierarchical_methods) {
      d <- .compute_distance(df)
      hc <- stats::hclust(d, method = method)
      return(stats::cutree(hc, k = k_i))
    }

    if (method == "kmeans") {
      return(do.call(stats::kmeans,
                     c(list(x= df, centers = k_i), cluster_args))$cluster)
    }

    if (method == "pam") {
      d <- .compute_distance(df)
      return(cluster::pam(d, k_i, diss = TRUE)$clustering)
    }

    if (method == "fuzzy") {
      return(do.call(e1071::cmeans,
                     c(list(x = df, centers = k_i), cluster_args))$cluster)
    }

    if (method == "mclust") {
      return(do.call(mclust::Mclust,
                     c(list(data = df, G = k_i), cluster_args))$classification)
    }

    if (method == "kmodes") {
      return(do.call(klaR::kmodes,
                     c(list(data = df, modes = k_i), cluster_args))$cluster)
    }

    if (method == "kprototypes") {
      return(do.call(clustMixType::kproto,
                     c(list(x = df, k = k_i), cluster_args))$cluster)
    }
  }

  # --------------------------------------------------------
  # Single k
  # --------------------------------------------------------
  if (length(k) == 1) {
    partitions <- lapply(database, function(df) {
      .run_clustering(df, k)
    })

    names(partitions) <- names(imp_list)
    return(partitions)
  }

  # --------------------------------------------------------
  # Multiple k
  # --------------------------------------------------------

  if (length(k) > 1) {

    if (method %in% hierarchical_methods) {
      hc_list <- lapply(database, function(df) {
        d <- .compute_distance(df)
        stats::hclust(d, method = method)
      })

      partitions_by_k <- lapply(k, function(k_i) {
        parts <- lapply(hc_list, function(hc) stats::cutree(hc, k = k_i))
        names(parts) <- names(imp_list)
        parts
      })

    } else {

      partitions_by_k <- lapply(k, function(k_i) {
        partitions <- lapply(database, function(df) {
          .run_clustering(df, k_i)
        })
        names(partitions) <- names(imp_list)
        partitions
      })
    }

    names(partitions_by_k) <- paste0("k", k)
    return(partitions_by_k)
  }
}

# tests/testthat/test-consensus.R
# Tests for validate_clustering() and choose_best_clustering()

# -------------------------------------------------------
# Reusable Helpers
# -------------------------------------------------------

# Perfectly separated partitions (two distinct groups)
make_perfect_parts <- function(n = 20, m = 3) {
  labels <- rep(1:2, each = n / 2)
  replicate(m, labels, simplify = FALSE) |>
    setNames(paste0("imp", seq_len(m)))
}

# Random partitions (pure noise)
make_random_parts <- function(n = 20, m = 3, k = 2, seed = 1) {
  set.seed(seed)
  replicate(m, sample(seq_len(k), n, replace = TRUE), simplify = FALSE) |>
    setNames(paste0("imp", seq_len(m)))
}

# Synthetic validation table for choose_best_clustering
make_val_table <- function() {
  data.frame(
    k                            = 2:4,
    pac                          = c(0.05, 0.20, 0.35),
    silhouette_mean              = c(0.70, 0.50, 0.30),
    ari_mean_between_imputations = c(0.90, 0.70, 0.50),
    ari_consensus_mean           = c(0.88, 0.65, 0.45),
    calinski_harabasz_mean       = c(200,  120,   80),
    davies_bouldin_mean          = c(0.30, 0.60,  0.90),
    dunn_index                   = c(0.50, 0.30,  0.20)
  )
}

# Synthetic consensus results for choose_best_clustering
make_cons_results <- function(n = 20) {
  set.seed(1)
  make_one <- function(k) list(
    k            = k,
    consensus    = sample(seq_len(k), n, replace = TRUE),
    coassignment = matrix(runif(n * n), n, n)
  )
  list(k2 = make_one(2), k3 = make_one(3), k4 = make_one(4))
}

# ============================================================
# validate_clustering — output structure
# ============================================================

test_that("returns a data.frame with one row per evaluated k", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  expect_s3_class(val, "data.frame")
  expect_equal(nrow(val), 1L)
})

test_that("contains all expected columns", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  expected_cols <- c("k", "pac", "silhouette_mean",
                     "ari_mean_between_imputations", "ari_consensus_mean",
                     "calinski_harabasz_mean", "davies_bouldin_mean", "dunn_index")
  expect_true(all(expected_cols %in% names(val)))
})

test_that("multiple k: returns one row for each evaluated k", {
  set.seed(1)
  imp   <- replicate(3, data.frame(x = rnorm(20), y = rnorm(20)), simplify = FALSE)
  parts <- cluster_imputations(imp, method = "ward.D2", k = 2:4)
  cons  <- consensus_clustering(parts)
  val   <- validate_clustering(parts, cons)
  expect_equal(nrow(val), 3L)
  expect_equal(sort(val$k), 2:4)
})

# ============================================================
# validate_clustering — metric ranges and properties
# ============================================================

test_that("pac is in [0, 1]", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  expect_gte(val$pac, 0)
  expect_lte(val$pac, 1)
})

test_that("silhouette_mean is in [-1, 1]", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$silhouette_mean)) {
    expect_gte(val$silhouette_mean, -1)
    expect_lte(val$silhouette_mean, 1)
  }
})

test_that("ari_mean_between_imputations is in [-1, 1]", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$ari_mean_between_imputations)) {
    expect_gte(val$ari_mean_between_imputations, -1)
    expect_lte(val$ari_mean_between_imputations,  1)
  }
})

test_that("ari_consensus_mean is in [-1, 1]", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$ari_consensus_mean)) {
    expect_gte(val$ari_consensus_mean, -1)
    expect_lte(val$ari_consensus_mean,  1)
  }
})

test_that("calinski_harabasz_mean is positive or NA", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$calinski_harabasz_mean)) {
    expect_gt(val$calinski_harabasz_mean, 0)
  }
})

test_that("davies_bouldin_mean is non-negative or NA", {
  parts <- make_random_parts(k = 2)
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$davies_bouldin_mean)) {
    expect_gte(val$davies_bouldin_mean, 0)
  }
})

# ============================================================
# validate_clustering — quality signals with ideal data
# ============================================================

test_that("perfect partitions produce low pac", {
  parts <- make_perfect_parts()
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  # With perfect clusters PAC should be very low
  expect_lt(val$pac, 0.1)
})

test_that("perfect partitions produce high silhouette", {
  parts <- make_perfect_parts()
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  if (!is.na(val$silhouette_mean)) {
    expect_gt(val$silhouette_mean, 0.5)
  }
})

test_that("perfect partitions produce ARI between imputations = 1", {
  parts <- make_perfect_parts()
  cons  <- consensus_clustering(parts, k = 2)
  val   <- validate_clustering(parts, cons)
  expect_equal(val$ari_mean_between_imputations, 1, tolerance = 1e-10)
})

# ============================================================
# validate_clustering — input validations
# ============================================================

test_that("fails if partitions is empty", {
  cons <- consensus_clustering(
    list(imp1 = c(1, 2, 1), imp2 = c(1, 2, 2)), k = 2
  )
  expect_error(validate_clustering(list(), cons), "non-empty")
})

test_that("fails if consensus_results is not a list", {
  parts <- make_random_parts()
  expect_error(validate_clustering(parts, "not_a_list"), "list")
})

# ============================================================
# choose_best_clustering — output structure
# ============================================================

test_that("returns list with all expected elements", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(choose_best_clustering(val, cons))
  expected <- c("best_k", "best_consensus", "best_coassignment",
                "best_consensus_result", "scores_table", "weights", "tie_breaker")
  expect_true(all(expected %in% names(best)))
})

test_that("best_k is one of the evaluated k", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_true(best$best_k %in% 2:4)
})

test_that("scores_table has score column", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_true("score" %in% names(best$scores_table))
})

test_that("scores_table is sorted by k ascending", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_equal(best$scores_table$k, sort(best$scores_table$k))
})

test_that("best_consensus has length n", {
  val  <- make_val_table()
  cons <- make_cons_results(n = 20)
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_length(best$best_consensus, 20)
})

test_that("best_coassignment is a square matrix", {
  val  <- make_val_table()
  cons <- make_cons_results(n = 20)
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_true(is.matrix(best$best_coassignment))
  expect_equal(nrow(best$best_coassignment), ncol(best$best_coassignment))
})

# ============================================================
# choose_best_clustering — selection logic
# ============================================================

test_that("selects k=2 when it is clearly better", {
  # k=2 dominates in all metrics of the synthetic table
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(choose_best_clustering(val, cons))
  expect_equal(best$best_k, 2L)
})

test_that("prefer_stability = TRUE uses stability weights", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(
    choose_best_clustering(val, cons, prefer_stability = TRUE)
  )
  expect_equal(best$weights[["pac"]], 2)
  expect_equal(best$weights[["ari_between"]], 2)
})

test_that("prefer_stability = FALSE uses compactness weights", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(
    choose_best_clustering(val, cons, prefer_stability = FALSE)
  )
  expect_equal(best$weights[["silhouette"]], 2)
  expect_equal(best$weights[["ch"]], 2)
})

test_that("prefer_stability = NULL uses equal weights", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(
    choose_best_clustering(val, cons, prefer_stability = NULL)
  )
  expect_true(all(best$weights == 1))
})

test_that("custom weights are respected", {
  val  <- make_val_table()
  cons <- make_cons_results()
  w    <- c(pac = 5, silhouette = 1, ari_between = 1,
            ari_consensus = 1, ch = 1, db = 1, dunn = 1)
  best <- suppressMessages(choose_best_clustering(val, cons, weights = w))
  expect_equal(best$weights[["pac"]], 5)
})

test_that("tie_breaker is recorded in the output", {
  val  <- make_val_table()
  cons <- make_cons_results()
  best <- suppressMessages(
    choose_best_clustering(val, cons, tie_breaker = "dunn")
  )
  expect_equal(best$tie_breaker, "dunn")
})

# ============================================================
# choose_best_clustering — input validations
# ============================================================

test_that("fails if columns are missing in validation_table", {
  val_bad <- data.frame(k = 2:3, pac = c(0.1, 0.2))
  cons    <- make_cons_results()
  expect_error(
    choose_best_clustering(val_bad, cons),
    "missing"
  )
})

test_that("fails with negative weights", {
  val  <- make_val_table()
  cons <- make_cons_results()
  w    <- c(pac = -1, silhouette = 1, ari_between = 1,
            ari_consensus = 1, ch = 1, db = 1, dunn = 1)
  expect_error(
    choose_best_clustering(val, cons, weights = w),
    ">= 0"
  )
})

test_that("fails with all weights at zero", {
  val  <- make_val_table()
  cons <- make_cons_results()
  w    <- c(pac = 0, silhouette = 0, ari_between = 0,
            ari_consensus = 0, ch = 0, db = 0, dunn = 0)
  expect_error(
    choose_best_clustering(val, cons, weights = w),
    "positive"
  )
})

test_that("fails with non-logical prefer_stability", {
  val  <- make_val_table()
  cons <- make_cons_results()
  expect_error(
    choose_best_clustering(val, cons, prefer_stability = "yes"),
    "prefer_stability"
  )
})

# tests/testthat/test-cluster.R
# Tests para cluster_imputations() y consensus_clustering()

# -------------------------------------------------------
# Reusable Helpers
# -------------------------------------------------------
make_num_imp <- function(n = 20, m = 3, seed = 1) {
  set.seed(seed)
  replicate(m, data.frame(x = rnorm(n), y = rnorm(n)), simplify = FALSE)
}

make_mixed_imp <- function(n = 20, m = 3, seed = 1) {
  set.seed(seed)
  replicate(m, data.frame(
    x    = rnorm(n),
    group = factor(sample(c("a", "b", "c"), n, replace = TRUE))
  ), simplify = FALSE)
}

make_bin_imp <- function(n = 20, m = 3, seed = 1) {
  set.seed(seed)
  replicate(m, data.frame(
    a = sample(0:1, n, replace = TRUE),
    b = sample(0:1, n, replace = TRUE),
    c = sample(0:1, n, replace = TRUE)
  ), simplify = FALSE)
}

# ============================================================
# cluster_imputations — output structure (single k)
# ============================================================

test_that("single k: returns a list of integer vectors", {
  set.seed(1)
  res <- cluster_imputations(make_num_imp(), method = "ward.D2", k = 2)
  expect_type(res, "list")
  expect_true(all(vapply(res, is.integer, logical(1))))
})

test_that("single k: number of partitions equals number of imputations", {
  imp <- make_num_imp(m = 4)
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2)
  expect_length(res, 4)
})

test_that("single k: each partition has as many elements as observations", {
  imp <- make_num_imp(n = 25, m = 3)
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 3)
  expect_true(all(vapply(res, length, integer(1)) == 25))
})

test_that("single k: cluster values are in {1, ..., k}", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 3)
  expect_true(all(vapply(res, function(z) all(z %in% 1:3), logical(1))))
})

test_that("single k: output names match input list names", {
  imp <- make_num_imp(m = 3)
  names(imp) <- c("imp1", "imp2", "imp3")
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2)
  expect_equal(names(res), names(imp))
})

# ============================================================
# cluster_imputations — output structure (multiple k)
# ============================================================

test_that("multiple k: returns a list of lists named k2, k3, ...", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2:4)
  expect_named(res, c("k2", "k3", "k4"))
  expect_true(all(vapply(res, is.list, logical(1))))
})

test_that("multiple k: each sublist has as many partitions as imputations", {
  imp <- make_num_imp(m = 3)
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2:3)
  expect_true(all(vapply(res, length, integer(1)) == 3))
})

test_that("multiple k: cluster values are correct for each k", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2:3)
  expect_true(all(vapply(res$k2, function(z) all(z %in% 1:2), logical(1))))
  expect_true(all(vapply(res$k3, function(z) all(z %in% 1:3), logical(1))))
})

# ============================================================
# cluster_imputations — hierarchical methods
# ============================================================

test_that("alternative hierarchical methods work", {
  imp <- make_num_imp()
  for (mth in c("ward.D", "single", "complete", "average")) {
    set.seed(1)
    res <- cluster_imputations(imp, method = mth, k = 2)
    expect_length(res, 3)
  }
})

# ============================================================
# cluster_imputations — kmeans
# ============================================================

test_that("kmeans: single k produces valid partitions", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "kmeans", k = 2)
  expect_length(res, 3)
  expect_true(all(vapply(res, function(z) all(z %in% 1:2), logical(1))))
})

test_that("kmeans: fails with categorical data", {
  imp <- make_mixed_imp()
  expect_error(
    cluster_imputations(imp, method = "kmeans", k = 2),
    "categorical"
  )
})

# ============================================================
# cluster_imputations — PAM with Gower distance (mixed data)
# ============================================================

test_that("PAM + Gower: works with mixed data", {
  imp <- make_mixed_imp()
  set.seed(1)
  expect_warning(
    res <- cluster_imputations(imp, method = "pam", k = 2, distance = "gower"),
    "ignored"
  )
  expect_length(res, 3)
  expect_true(all(vapply(res, function(z) all(z %in% 1:2), logical(1))))
})

test_that("PAM + Gower: multiple k returns correct structure", {
  imp <- make_mixed_imp()
  set.seed(1)
  expect_warning(
    res <- cluster_imputations(imp, method = "pam", k = 2:3, distance = "gower"),
    "ignored"
  )
  expect_named(res, c("k2", "k3"))
})

# ============================================================
# cluster_imputations — input validations
# ============================================================

test_that("fails with k = 1", {
  imp <- make_num_imp()
  expect_error(cluster_imputations(imp, method = "ward.D2", k = 1), "k must be")
})

test_that("fails with k >= n_obs", {
  imp <- make_num_imp(n = 10)
  expect_error(cluster_imputations(imp, method = "ward.D2", k = 10), "k must be")
})

test_that("fails with unsupported method", {
  imp <- make_num_imp()
  expect_error(
    cluster_imputations(imp, method = "non_existent_method", k = 2),
    "Unsupported"
  )
})

test_that("fails with distance = 'custom' without dist_fun", {
  imp <- make_num_imp()
  expect_error(
    cluster_imputations(imp, method = "ward.D2", k = 2, distance = "custom"),
    "dist_fun"
  )
})

test_that("fails with categorical data and euclidean distance in hierarchical", {
  imp <- make_mixed_imp()
  expect_error(
    cluster_imputations(imp, method = "ward.D2", k = 2, distance = "euclidean"),
    "Categorical data"
  )
})

test_that("fails with jaccard on non-binary data", {
  imp <- make_num_imp()
  expect_error(
    cluster_imputations(imp, method = "ward.D2", k = 2, distance = "jaccard"),
    "binary"
  )
})

# ============================================================
# cluster_imputations — global vs within scaling
# ============================================================

test_that("scale_data = 'global' does not change the number of partitions", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2, scale_data = "global")
  expect_length(res, 3)
})

test_that("scale_data = 'within' produces output equivalent in structure", {
  imp <- make_num_imp()
  set.seed(1)
  res <- cluster_imputations(imp, method = "ward.D2", k = 2, scale_data = "within")
  expect_length(res, 3)
  expect_true(all(vapply(res, function(z) all(z %in% 1:2), logical(1))))
})

# ============================================================
# consensus_clustering — output structure (single k)
# ============================================================

test_that("single k: returns list with mandatory elements", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 20, replace = TRUE),
    imp2 = sample(1:2, 20, replace = TRUE),
    imp3 = sample(1:2, 20, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2)
  expect_named(res, c("consensus_method", "k", "consensus",
                      "coassignment", "hclust", "weights"),
               ignore.order = TRUE)
})

test_that("single k: consensus has length n", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 20, replace = TRUE),
    imp2 = sample(1:2, 20, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2)
  expect_length(res$consensus, 20)
})

test_that("single k: coassignment is an n x n matrix with values in [0, 1]", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 20, replace = TRUE),
    imp2 = sample(1:2, 20, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2)
  expect_true(is.matrix(res$coassignment))
  expect_equal(dim(res$coassignment), c(20, 20))
  expect_true(all(res$coassignment >= 0 & res$coassignment <= 1))
})

test_that("single k: coassignment diagonal is 1 (each obs. with itself)", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:3, 15, replace = TRUE),
    imp2 = sample(1:3, 15, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 3)
  expect_true(all(diag(res$coassignment) == 1))
})

test_that("single k: coassignment is symmetric", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 15, replace = TRUE),
    imp2 = sample(1:2, 15, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2)
  expect_equal(res$coassignment, t(res$coassignment))
})

test_that("single k: weights sum to 1 (classic method)", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 15, replace = TRUE),
    imp2 = sample(1:2, 15, replace = TRUE),
    imp3 = sample(1:2, 15, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2, consensus_method = "classic")
  expect_equal(sum(res$weights), 1, tolerance = 1e-10)
})

test_that("single k: hclust is an object of class hclust", {
  set.seed(1)
  parts <- list(
    imp1 = sample(1:2, 15, replace = TRUE),
    imp2 = sample(1:2, 15, replace = TRUE)
  )
  res <- consensus_clustering(parts, k = 2)
  expect_s3_class(res$hclust, "hclust")
})

# ============================================================
# consensus_clustering — output structure (multiple k)
# ============================================================

test_that("multiple k: returns a list named k2, k3, ...", {
  imp <- make_num_imp()
  set.seed(1)
  parts <- cluster_imputations(imp, method = "ward.D2", k = 2:4)
  res   <- consensus_clustering(parts)
  expect_named(res, c("k2", "k3", "k4"))
})

test_that("multiple k: each element has the single k structure", {
  imp <- make_num_imp()
  set.seed(1)
  parts <- cluster_imputations(imp, method = "ward.D2", k = 2:3)
  res   <- consensus_clustering(parts)
  for (ki in res) {
    expect_true(all(c("consensus", "coassignment", "hclust", "weights") %in% names(ki)))
  }
})

test_that("multiple k: the k field of each element is correct", {
  imp <- make_num_imp()
  set.seed(1)
  parts <- cluster_imputations(imp, method = "ward.D2", k = 2:4)
  res   <- consensus_clustering(parts)
  expect_equal(res$k2$k, 2L)
  expect_equal(res$k3$k, 3L)
  expect_equal(res$k4$k, 4L)
})

# ============================================================
# consensus_clustering — input validations
# ============================================================

test_that("fails with empty list", {
  expect_error(consensus_clustering(list(), k = 2), "non-empty")
})

test_that("fails with single k and no k argument", {
  parts <- list(imp1 = c(1, 2, 1), imp2 = c(1, 1, 2))
  expect_error(consensus_clustering(parts), "k")
})

test_that("fails with k < 2", {
  parts <- list(imp1 = c(1, 2, 1), imp2 = c(1, 1, 2))
  expect_error(consensus_clustering(parts, k = 1), "k must be")
})

test_that("fails with unsupported cluster_method", {
  parts <- list(imp1 = c(1, 2, 1), imp2 = c(1, 1, 2))
  expect_error(
    consensus_clustering(parts, k = 2, cluster_method = "bad_method"),
    "cluster_method"
  )
})

test_that("fails if partitions have different lengths", {
  parts <- list(imp1 = c(1, 2, 1), imp2 = c(1, 2))
  expect_error(consensus_clustering(parts, k = 2), "same length")
})

#' Build a consensus partition from multiple imputation clustering results
#'
#' @description
#' Constructs a consensus clustering solution from a collection of partitions
#' obtained across multiple imputed datasets (as produced by
#' \code{\link{cluster_imputations}}). The function builds a co-assignment
#' matrix reflecting how frequently each pair of observations is assigned to
#' the same cluster, and derives the final consensus partition via hierarchical
#' clustering on the resulting dissimilarity. Supports both a single value of
#' \code{k} and a range of values.
#'
#' @param partitions A list of cluster assignment vectors (one per imputed
#'   dataset) for a single \code{k}, or a named list of such lists (one per
#'   \code{k} value, named \code{"k2"}, \code{"k3"}, etc.) as returned by
#'   \code{\link{cluster_imputations}} when \code{k} is a vector.
#' @param k An integer specifying the number of consensus clusters. Required
#'   when \code{partitions} is a flat list of vectors (single-k input).
#'   Ignored when \code{partitions} is a list-of-lists with names following
#'   the pattern \code{"k2"}, \code{"k3"}, etc., as \code{k} is inferred
#'   automatically from the names.
#' @param cluster_method A character string specifying the agglomeration
#'   method passed to \code{\link[stats]{hclust}} for the consensus stage.
#'   Accepted values are \code{"ward.D"}, \code{"ward.D2"}, \code{"single"},
#'   \code{"complete"}, \code{"average"}, \code{"centroid"}, \code{"median"},
#'   and \code{"mcquitty"}. Default is \code{"ward.D2"}.
#' @param consensus_method A character string specifying how the
#'   co-assignment matrix is built. Options are:
#'   \itemize{
#'     \item \code{"classic"}: unweighted co-assignment; all partitions
#'           contribute equally.
#'     \item \code{"weighted_ari"}: partitions are weighted by their
#'           pairwise ARI centrality, so more consistent partitions
#'           contribute more to the consensus. Requires \pkg{mclust}.
#'   }
#'   Default is \code{"classic"}.
#'
#' @return
#' \itemize{
#'   \item If \code{partitions} is a flat list (single \code{k}): a named
#'         list with the following elements:
#'     \itemize{
#'       \item \code{consensus_method}: character string with the method used.
#'       \item \code{k}: integer, number of clusters requested.
#'       \item \code{consensus}: integer vector of consensus cluster labels,
#'             one per observation.
#'       \item \code{coassignment}: numeric matrix of size \eqn{n \times n}
#'             with co-assignment probabilities between 0 and 1.
#'       \item \code{hclust}: the \code{hclust} object from the consensus
#'             stage.
#'       \item \code{weights}: named numeric vector of partition weights.
#'     }
#'   \item If \code{partitions} is a list-of-lists (multiple \code{k}):
#'         a named list where each element corresponds to one value of
#'         \code{k} and contains the structure described above.
#' }
#'
#' @details
#' The co-assignment matrix \eqn{C} is defined such that \eqn{C_{ij}}
#' represents the weighted proportion of partitions in which observations
#' \eqn{i} and \eqn{j} are assigned to the same cluster. The final
#' dissimilarity is computed as \eqn{1 - C} and passed to
#' \code{\link[stats]{hclust}}.
#'
#' When \code{consensus_method = "weighted_ari"}, the weight of each
#' partition is proportional to its mean ARI against all other partitions,
#' reflecting its centrality within the ensemble. Negative centrality values
#' are floored at zero. If all centralities are zero, the function falls back
#' to equal weights with a warning.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Basic consensus clustering
#' # ------------------------------------------------------------
#' set.seed(123)
#' # simulate 3 partitions
#' partitions <- list(
#'   imp1 = sample(1:2, 10, replace = TRUE),
#'   imp2 = sample(1:2, 10, replace = TRUE),
#'   imp3 = sample(1:2, 10, replace = TRUE)
#' )
#'
#' # Classic consensus
#' cons <- consensus_clustering(partitions, k = 2)
#' str(cons)
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: consensus method with mice
#' # ------------------------------------------------------------
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#'   set.seed(123)
#'
#'   imp    <- mice::mice(mice::nhanes, m = 3, printFlag = FALSE)
#'   mild   <- as_mild_list(imp)
#'   parts  <- cluster_imputations(mild, method = "ward.D2", k = 3)
#'
#'   # -------------------------------------------------------------
#'   # Single k, classic consensus
#'   # -------------------------------------------------------------
#'   cons <- consensus_clustering(parts, k = 3)
#'
#'   # -------------------------------------------------------------
#'   # Single k, ARI-weighted consensus
#'  # -------------------------------------------------------------
#'   if (requireNamespace("mclust", quietly = TRUE)) {
#'     cons_w <- consensus_clustering(parts, k = 3,
#'                                   consensus_method = "weighted_ari")
#' }
#'   # -------------------------------------------------------------
#'   # Multiple k values
#'   # -------------------------------------------------------------
#'   parts_multi <- cluster_imputations(mild, method = "ward.D2", k = 2:4)
#'   cons_multi  <- consensus_clustering(parts_multi)
#' }
#' }
#'
#' @references
#' Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
#' Consensus clustering: A resampling-based method for class discovery
#' and visualization of gene expression microarray data.
#' \emph{Machine Learning}, \strong{52}(1-2), 91-118.
#' \doi{10.1023/A:1023949509487}
#'
#' Hubert, L., & Arabie, P. (1985).
#' Comparing partitions.
#' \emph{Journal of Classification}, \strong{2}, 193-218.
#' \doi{10.1007/BF01908075}
#'
#' @seealso \code{\link{cluster_imputations}}, \code{\link{validate_clustering}}
#'
#' @export
consensus_clustering <- function(partitions,
                                 k = NULL,
                                 cluster_method = "ward.D2",
                                 consensus_method = c("classic", "weighted_ari")) {

  #Control verbose
  verbose <- getOption("cclustr.verbose", FALSE)

  # --------------------------------------------------------
  # Supported methods for hclust consensus stage
  # --------------------------------------------------------
  hierarchical_methods <- c(
    "ward.D", "ward.D2", "single", "complete",
    "average", "centroid", "median", "mcquitty"
  )

  if (!(cluster_method %in% hierarchical_methods)) {
    stop("cluster_method not supported for hclust().")
  }

  # --------------------------------------------------------
  # Input validation: partitions must be a list
  # --------------------------------------------------------
  if (!is.list(partitions) || length(partitions) == 0) {
    stop("Partitions must be provided as a non-empty list.")
  }

  consensus_method <- match.arg(consensus_method)

  # --------------------------------------------------------
  # Detect input structure
  # Case A: list of vectors  -> single k
  # Case B: list of lists    -> multiple k (output of cluster_imputations when k is a range)
  # --------------------------------------------------------
  is_list_of_lists <- all(vapply(partitions, is.list, logical(1)))

  # --------------------------------------------------------
  # Helper: run consensus for a single k (core logic)
  # --------------------------------------------------------
  .consensus_single_k <- function(partitions_k, k_value) {

    # partitions_k must be a list of vectors
    if (!is.list(partitions_k) || length(partitions_k) == 0) {
      stop("Each k element must be a non-empty list of vectors.")
    }

    # Convert partitions to atomic vectors (defensive)
    partitions_k <- lapply(partitions_k, function(z) {
      if (is.data.frame(z) || is.matrix(z)) z <- as.vector(z)
      if (is.factor(z)) z <- as.character(z)
      if (!is.atomic(z)) stop("Each partition must be an atomic vector.")
      z
    })

    # n observations and m partitions
    n <- length(partitions_k[[1]])
    m <- length(partitions_k)

    if (n == 0) stop("Partitions contain empty vectors.")

    # Consistency check: all partitions must have same length
    if (!all(vapply(partitions_k, length, integer(1)) == n)) {
      stop("All partitions must have the same length.")
    }

    # Validate k
    if (length(k_value) != 1 || !is.numeric(k_value) || is.na(k_value) || k_value < 2) {
      stop("k must be a single numeric value >= 2.")
    }
    k_value <- as.integer(k_value)

    # --------------------------------------------------------
    # Compute weights
    # --------------------------------------------------------

    # Initialize equal weights (classic consensus baseline)
    weights <- rep(1 / m, m)
    names(weights) <- names(partitions_k)

    if (consensus_method == "classic") {
      # do nothing; equal weights
      if (verbose) message(paste0("Detected consensus method: classic (k = ", k_value, ")"))

    } else if (consensus_method == "weighted_ari") {
      if (verbose) message(paste0("Detected consensus method: weighted_ari (k = ", k_value, ")"))

      if (!requireNamespace("mclust", quietly = TRUE)) {
        stop("Package 'mclust' is required for consensus_method = 'weighted_ari'.")
      }

      # Pairwise ARI matrix (m x m)
      ari_mat <- matrix(NA_real_, nrow = m, ncol = m)

      for (i in seq_len(m)) {
        ari_mat[i, i] <- 1
        for (j in seq_len(m)) {
          if (j > i) {
            ari <- mclust::adjustedRandIndex(partitions_k[[i]], partitions_k[[j]])
            ari_mat[i, j] <- ari
            ari_mat[j, i] <- ari
          }
        }
      }

      # Centrality score: mean ARI vs others (exclude self)
      centrality <- vapply(seq_len(m), function(i) mean(ari_mat[i, -i], na.rm = TRUE), numeric(1))
      centrality <- pmax(centrality, 0) # no negative weights

      if (all(centrality == 0)) {
        warning(paste0("All ARI centralities are 0 for k = ", k_value, ". Falling back to equal weights."))
        weights <- rep(1 / m, m)
      } else {
        weights <- centrality / sum(centrality)
      }

      names(weights) <- names(partitions_k)
    }

    # --------------------------------------------------------
    # Build (weighted) co-assignment matrix
    # --------------------------------------------------------
    coassignment <- matrix(0, n, n)

    for (idx in seq_len(m)) {
      z <- partitions_k[[idx]]
      coassignment <- coassignment + weights[idx] * outer(z, z, FUN = "==")
    }

    coassignment <- pmin(pmax(coassignment, 0), 1)

    # --------------------------------------------------------
    # Consensus clustering based on dissimilarity
    # --------------------------------------------------------
    diss <- stats::as.dist(1 - coassignment)
    hc <- stats::hclust(diss, method = cluster_method)
    consensus <- stats::cutree(hc, k = k_value)

    return(list(
      consensus_method = consensus_method,
      k = k_value,
      consensus = consensus,
      coassignment = coassignment,
      hclust = hc,
      weights = weights
    ))
  }

  # --------------------------------------------------------
  # Case A: partitions is a list of vectors (single k)
  # --------------------------------------------------------
  if (!is_list_of_lists) {

    if (is.null(k)) {
      stop("When 'partitions' is a list of vectors (single k), you must provide k.")
    }

    res <- .consensus_single_k(partitions, k)
    if (verbose) message("Consensus clustering completed")
    return(res)
  }

  # --------------------------------------------------------
  # Case B: partitions is a list-of-lists by k (multiple k)
  # Output of cluster_imputations() when k is a range: names like 'k2','k3',...
  # --------------------------------------------------------
  # Infer k values from names if possible; otherwise require k
  k_names <- names(partitions)

  if (!is.null(k_names) && all(grepl("^k[0-9]+$", k_names))) {
    k_values <- as.integer(sub("^k", "", k_names))
  } else {
    if (is.null(k)) {
      stop("For multiple-k input (list-of-lists), provide k or name elements as 'k2','k3',...")
    }
    k_values <- as.integer(k)
  }

  # Run consensus for each k
  results_by_k <- vector("list", length(partitions))
  names(results_by_k) <- names(partitions)

  for (idx in seq_along(partitions)) {
    k_val <- k_values[idx]
    results_by_k[[idx]] <- .consensus_single_k(partitions[[idx]], k_val)
  }

  if (verbose) message("Consensus clustering completed for multiple k values")
  return(results_by_k)
}

#' Select the optimal number of clusters from a validation table
#'
#' @description
#' Identifies the best value of \code{k} from the validation metrics produced
#' by \code{\link{validate_clustering}}, using a weighted rank aggregation
#' strategy. Each metric is ranked independently and then combined into a
#' single score using user-defined weights. The \code{k} with the lowest
#' weighted rank score is selected as optimal. In case of ties, a secondary
#' metric is used as a tiebreaker.
#'
#' @param validation_table A \code{data.frame} as returned by
#'   \code{\link{validate_clustering}}, containing one row per evaluated
#'   \code{k} and columns \code{pac}, \code{silhouette_mean},
#'   \code{ari_mean_between_imputations}, \code{ari_consensus_mean},
#'   \code{calinski_harabasz_mean}, \code{davies_bouldin_mean}, and
#'   \code{dunn_index}.
#' @param consensus_results A list as returned by
#'   \code{\link{consensus_clustering}}, either for a single \code{k} or
#'   for multiple \code{k} values. Used to retrieve the consensus labels
#'   and co-assignment matrix for the selected \code{k}.
#' @param weights A named numeric vector specifying the relative importance
#'   of each metric in the rank aggregation. Expected names are \code{pac},
#'   \code{silhouette}, \code{ari_between}, \code{ari_consensus}, \code{ch},
#'   \code{db}, and \code{dunn}. If \code{NULL} (default), weights are set
#'   automatically based on \code{prefer_stability}. When provided
#'   explicitly, \code{prefer_stability} is ignored.
#' @param prefer_stability Controls the default weighting strategy when
#'   \code{weights} is \code{NULL}. Accepted values are:
#'   \itemize{
#'     \item \code{NULL} (default): all metrics receive equal weight,
#'           providing a balanced selection that does not favor any
#'           particular aspect of clustering quality.
#'     \item \code{TRUE}: stability metrics (\code{pac},
#'           \code{ari_between}, \code{ari_consensus}) receive higher
#'           weights.
#'     \item \code{FALSE}: internal compactness metrics
#'           (\code{silhouette}, \code{ch}, \code{db}, \code{dunn})
#'           receive higher weights.
#'   }
#' @param tie_breaker A character string specifying the secondary metric
#'   used to break ties in the weighted rank score. One of
#'   \code{"silhouette"} (default), \code{"pac"}, \code{"dunn"},
#'   \code{"ch"}, \code{"db"}, \code{"ari_between"}, or
#'   \code{"ari_consensus"}.
#'
#' @return A named list with the following elements:
#'   \itemize{
#'     \item \code{best_k}: integer, the selected optimal number of clusters.
#'     \item \code{best_consensus}: integer vector of consensus cluster
#'           labels for the selected \code{k}, one per observation.
#'     \item \code{best_coassignment}: numeric matrix of size
#'           \eqn{n \times n} with co-assignment probabilities for the
#'           selected \code{k}.
#'     \item \code{best_consensus_result}: the full consensus result object
#'           for the selected \code{k}, as returned by
#'           \code{\link{consensus_clustering}}.
#'     \item \code{scores_table}: the input \code{validation_table} with an
#'           additional \code{score} column containing the weighted rank
#'           score for each \code{k}, sorted by \code{k}.
#'     \item \code{weights}: named numeric vector of weights effectively
#'           used in the aggregation.
#'     \item \code{tie_breaker}: character string indicating the tiebreaker
#'           metric used.
#'   }
#'
#' @details
#' The rank aggregation proceeds as follows:
#' \enumerate{
#'   \item Each metric is ranked from best to worst (rank 1 = best),
#'         accounting for the direction of optimality: lower is better for
#'         \code{pac} and \code{davies_bouldin_mean}; higher is better for
#'         all remaining metrics.
#'   \item For each \code{k}, the weighted average rank is computed using
#'         the provided or default weights. Rows with \code{NA} in some
#'         metrics are handled by renormalizing the weights over available
#'         metrics only.
#'   \item The \code{k} with the lowest weighted rank score is selected.
#'         If multiple \code{k} values share the minimum score, the
#'         tiebreaker metric is used; if still tied, the smallest \code{k}
#'         is preferred.
#' }
#'
#' Default weights when \code{prefer_stability = NULL} (default):
#' \code{pac = 1}, \code{silhouette = 1}, \code{ari_between = 1},
#' \code{ari_consensus = 1}, \code{ch = 1}, \code{db = 1}, \code{dunn = 1}.
#'
#' Default weights when \code{prefer_stability = TRUE}:
#' \code{pac = 2}, \code{silhouette = 1.5}, \code{ari_between = 2},
#' \code{ari_consensus = 2}, \code{ch = 1}, \code{db = 1}, \code{dunn = 1}.
#'
#' Default weights when \code{prefer_stability = FALSE}:
#' \code{pac = 1}, \code{silhouette = 2}, \code{ari_between = 1},
#' \code{ari_consensus = 1}, \code{ch = 2}, \code{db = 2}, \code{dunn = 2}.
#'
#' No automatic selection method replaces domain knowledge. The
#' \code{scores_table} and \code{\link{plot_validation_metrics}} are
#' provided so that the user can inspect individual metrics and make an
#' informed decision.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Basic validation with simulated partitions
#' # ------------------------------------------------------------
#'
#' # 1. Simulated validation table for k = 2, 3, 4
#' val_table <- data.frame(
#'   k                            = 2:4,
#'   pac                          = c(0.15, 0.08, 0.20),
#'   silhouette_mean              = c(0.42, 0.61, 0.38),
#'   ari_mean_between_imputations = c(0.80, 0.91, 0.75),
#'   ari_consensus_mean           = c(0.82, 0.93, 0.77),
#'   calinski_harabasz_mean       = c(120,  198,  105),
#'   davies_bouldin_mean          = c(0.85, 0.54, 0.97),
#'   dunn_index                   = c(0.30, 0.48, 0.27)
#' )
#'
#' # 2. Simulated consensus result for each k
#' set.seed(123)
#' n <- 30
#' make_cons <- function(k) {
#'   list(
#'     k            = k,
#'     consensus    = sample(seq_len(k), n, replace = TRUE),
#'     coassignment = matrix(runif(n * n), n, n)
#'   )
#' }
#' cons_results <- list(k2 = make_cons(2),
#'                      k3 = make_cons(3),
#'                      k4 = make_cons(4))
#'
#' # 3. Default selection (equal weights)
#' best <- choose_best_clustering(val_table, cons_results)
#' best$best_k
#' head(best$scores_table)
#'
#'
#' # 4. Custom weights
#' best_custom <- choose_best_clustering(
#'   val_table, cons_results,
#'   weights = c(pac = 3, silhouette = 1, ari_between = 3,
#'               ari_consensus = 3, ch = 1, db = 1, dunn = 1)
#' )
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: Full pipeline with mice
#' # ------------------------------------------------------------
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#' set.seed(123)
#' imp   <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#' mild  <- as_mild_list(imp)
#' parts <- cluster_imputations(mild, method = "ward.D2", k = 2:5)
#' cons  <- consensus_clustering(parts)
#' val   <- validate_clustering(parts, cons)
#'
#' # Default selection (equal weights)
#' best <- choose_best_clustering(val, cons)
#' best$best_k
#'
#' # Stability-focused selection
#' best2 <- choose_best_clustering(val, cons,
#'                                 prefer_stability = TRUE)
#'
#' # Compactness-focused selection
#' best3 <- choose_best_clustering(val, cons,
#'                                 prefer_stability = FALSE,
#'                                 tie_breaker      = "dunn")
#'
#' # Custom weights
#' best4 <- choose_best_clustering(val, cons,
#'                                 weights = c(pac = 3, silhouette = 1,
#'                                             ari_between = 3, ari_consensus = 3,
#'                                             ch = 1, db = 1, dunn = 1))
#' }
#' }
#'
#' @references
#' Pihur, V., Datta, S., & Datta, S. (2007). Weighted rank aggregation
#' of cluster validation measures: a Monte Carlo cross-entropy approach.
#' \emph{Bioinformatics}, \strong{23}(13), 1607-1615.
#' \doi{10.1093/bioinformatics/btm158}
#'
#' Hubert, L., & Arabie, P. (1985).
#' Comparing partitions.
#' \emph{Journal of Classification}, \strong{2}, 193-218.
#' \doi{10.1007/BF01908075}
#'
#' @seealso \code{\link{validate_clustering}}, \code{\link{consensus_clustering}},
#'   \code{\link{plot_validation_metrics}}
#'
#' @export
choose_best_clustering <- function(validation_table,
                                   consensus_results,
                                   weights          = NULL,
                                   prefer_stability = NULL,
                                   tie_breaker      = c("silhouette", "pac", "dunn", "ch", "db",
                                                        "ari_between", "ari_consensus")) {

  #Control verbosity
  verbose <- getOption("cclustr.verbose", FALSE)

  tie_breaker <- match.arg(tie_breaker)

  # Validate prefer_stability type
  if (!is.null(prefer_stability) && !is.logical(prefer_stability))
    stop("'prefer_stability' must be TRUE, FALSE, or NULL.")

  # Required columns
  req  <- c("k", "pac", "silhouette_mean",
            "ari_mean_between_imputations", "ari_consensus_mean",
            "calinski_harabasz_mean", "davies_bouldin_mean", "dunn_index")

  miss <- setdiff(req, names(validation_table))
  if (length(miss) > 0) {
    stop(paste0("validation_table is missing: ", paste(miss, collapse = ", ")))
  }

  # default weights
  if (is.null(weights)) {
    if (isTRUE(prefer_stability)) {
      if (verbose) message("Weighting strategy: stability-focused")
      weights <- c(pac = 2, silhouette = 1.5,
                   ari_between = 2, ari_consensus = 2,
                   ch = 1, db = 1, dunn = 1)

    } else if (isFALSE(prefer_stability)) {
      if (verbose) message("Weighting strategy: compactness-focused")
      weights <- c(pac = 1, silhouette = 2,
                   ari_between = 1, ari_consensus = 1,
                   ch = 2, db = 2, dunn = 2)

    } else {
      if (verbose) message("Weighting strategy: equal weights")
      weights <- c(pac = 1, silhouette = 1,
                   ari_between = 1, ari_consensus = 1,
                   ch = 1, db = 1, dunn = 1)
    }
  } else {
    # Validate user-supplied weights
    expected_w <- c("pac", "silhouette", "ari_between", "ari_consensus",
                    "ch", "db", "dunn")
    unknown_w  <- setdiff(names(weights), expected_w)
    missing_w  <- setdiff(expected_w, names(weights))

    if (length(unknown_w) > 0)
      warning("Unknown names in 'weights' will be ignored: ",
              paste(unknown_w, collapse = ", "))
    if (length(missing_w) > 0)
      warning("Metrics without a weight will be treated as 0: ",
              paste(missing_w, collapse = ", "))
    if (any(weights < 0, na.rm = TRUE))
      stop("All values in 'weights' must be >= 0.")
    if (sum(weights, na.rm = TRUE) == 0)
      stop("At least one weight in 'weights' must be positive.")

    # Fill missing weight names with zero
    weights[missing_w] <- 0
  }

  .rank_best1 <- function(x, higher_is_better = TRUE) {
    if (all(is.na(x))) return(rep(NA_real_, length(x)))
    if (higher_is_better) rank(-x, ties.method = "average", na.last = "keep")
    else                  rank( x, ties.method = "average", na.last = "keep")
  }

  vt <- validation_table

  r_pac  <- .rank_best1(vt$pac,                          higher_is_better = FALSE)
  r_sil  <- .rank_best1(vt$silhouette_mean,              higher_is_better = TRUE)
  r_ab   <- .rank_best1(vt$ari_mean_between_imputations, higher_is_better = TRUE)
  r_ac   <- .rank_best1(vt$ari_consensus_mean,           higher_is_better = TRUE)
  r_ch   <- .rank_best1(vt$calinski_harabasz_mean,       higher_is_better = TRUE)
  r_db   <- .rank_best1(vt$davies_bouldin_mean,          higher_is_better = FALSE)
  r_dunn <- .rank_best1(vt$dunn_index,                   higher_is_better = TRUE)

  # Compute weighted average rank per k, renormalizing over available metrics
  score <- numeric(nrow(vt))
  for (i in seq_len(nrow(vt))) {
    vals <- c(pac           = r_pac[i],
              silhouette    = r_sil[i],
              ari_between   = r_ab[i],
              ari_consensus = r_ac[i],
              ch            = r_ch[i],
              db            = r_db[i],
              dunn          = r_dunn[i])
    ok <- !is.na(vals)
    if (!any(ok)) { score[i] <- NA_real_; next }
    wi       <- weights[names(vals)][ok]
    wi       <- wi / sum(wi)
    score[i] <- sum(wi * vals[ok])
  }
  vt$score <- score

  tb <- switch(tie_breaker,
               silhouette    =  vt$silhouette_mean,
               pac           = -vt$pac,
               dunn          =  vt$dunn_index,
               ch            =  vt$calinski_harabasz_mean,
               db            = -vt$davies_bouldin_mean,
               ari_between   =  vt$ari_mean_between_imputations,
               ari_consensus =  vt$ari_consensus_mean)

  best_idx <- order(vt$score, -tb, vt$k)[1]
  best_k   <- vt$k[best_idx]

  best_obj <- NULL
  if (!is.null(consensus_results$k) && !is.null(consensus_results$consensus)) {
    best_obj <- consensus_results
  } else {
    key <- paste0("k", best_k)
    if (!is.null(consensus_results[[key]])) {
      best_obj <- consensus_results[[key]]
    } else {
      hit <- which(vapply(consensus_results, function(z) !is.null(z$k) && z$k == best_k, logical(1)))
      if (length(hit) == 1) best_obj <- consensus_results[[hit]]
    }
  }

  if (is.null(best_obj) || is.null(best_obj$consensus) || is.null(best_obj$coassignment)) {
    stop("Could not retrieve consensus labels/coassignment for the selected k from consensus_results.")
  }

  message("Best k selected: ", best_k)

  list(
    best_k                = best_k,
    best_consensus        = best_obj$consensus,
    best_coassignment     = best_obj$coassignment,
    best_consensus_result = best_obj,
    scores_table          = vt[order(vt$k), ],
    weights               = weights,
    tie_breaker           = tie_breaker
  )
}

#' Plot validation metrics across candidate numbers of clusters
#'
#' @description
#' Produces a multi-panel line plot displaying one or more clustering
#' validation metrics as a function of the number of clusters \code{k},
#' as computed by \code{\link{validate_clustering}} and optionally extended
#' with a \code{score} column by \code{\link{choose_best_clustering}}. For
#' each metric, the optimal value is highlighted in red.
#'
#' @param validation_table A \code{data.frame} as returned by
#'   \code{\link{validate_clustering}} or \code{$scores_table} from
#'   \code{\link{choose_best_clustering}}. Must contain a column \code{k}
#'   and at least one of the requested metric columns.
#' @param metrics A character vector specifying which metrics to plot.
#'   Only metrics present in \code{validation_table} are plotted; others
#'   are silently ignored. Default includes all standard metrics plus
#'   \code{score}:
#'   \code{"pac"}, \code{"silhouette_mean"},
#'   \code{"ari_mean_between_imputations"}, \code{"ari_consensus_mean"},
#'   \code{"calinski_harabasz_mean"}, \code{"davies_bouldin_mean"},
#'   \code{"dunn_index"}, and \code{"score"}.
#'
#' @return Invisibly returns \code{NULL}. The function is called for its
#'   side effect of producing a plot.
#'
#' @details
#' The number of panels is determined automatically from the number of
#' available metrics using \code{ceiling(sqrt(n))} rows and
#' \code{ceiling(n / nrow)} columns, producing a layout as square as
#' possible.
#'
#' The direction of optimality is accounted for when highlighting the best
#' value: lower is better for \code{pac}, \code{davies_bouldin_mean}, and
#' \code{score}; higher is better for all remaining metrics. The optimal
#' point for each metric is highlighted in red.
#'
#' Original graphical parameters are restored on exit via
#' \code{\link[graphics]{par}}.
#'
#' When \code{score} is present (i.e., \code{validation_table} comes from
#' \code{\link{choose_best_clustering}}), the panel for \code{score} shows
#' the overall weighted rank aggregation - the \code{k} with the lowest
#' score is the recommended solution.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Validation metrics with simulated data
#' # ------------------------------------------------------------
#' val_table <- data.frame(
#'   k                            = 2:4,
#'   pac                          = c(0.15, 0.08, 0.20),
#'   silhouette_mean              = c(0.42, 0.61, 0.38),
#'   ari_mean_between_imputations = c(0.80, 0.91, 0.75),
#'   ari_consensus_mean           = c(0.82, 0.93, 0.77),
#'   calinski_harabasz_mean       = c(120,  198,  105),
#'   davies_bouldin_mean          = c(0.85, 0.54, 0.97),
#'   dunn_index                   = c(0.30, 0.48, 0.27)
#' )
#'
#' # Plot all metrics
#' plot_validation_metrics(val_table)
#'
#' # Plot only stability metrics
#' plot_validation_metrics(val_table,
#'                         metrics = c("pac",
#'                                     "ari_mean_between_imputations",
#'                                     "ari_consensus_mean"))
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: metrics validation with mice
#' # ------------------------------------------------------------
#'
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#' set.seed(123)
#' imp    <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#' mild   <- as_mild_list(imp)
#' parts  <- cluster_imputations(mild, method = "ward.D2", k = 2:5)
#' cons   <- consensus_clustering(parts)
#' val    <- validate_clustering(parts, cons)
#'
#' # Plot all available metrics (no score column yet)
#' plot_validation_metrics(val)
#'
#' # Include score column from choose_best_clustering
#' best <- choose_best_clustering(val, cons)
#' plot_validation_metrics(best)
#'
#' # Plot only stability metrics
#' plot_validation_metrics(val,
#'                         metrics = c("pac",
#'                                     "ari_mean_between_imputations",
#'                                     "ari_consensus_mean"))
#' }
#' }
#'
#' @seealso
#' \code{\link{validate_clustering}}, \code{\link{choose_best_clustering}},
#' \code{\link{plot_consensus_dendrogram}}, \code{\link{plot_consensus_matrix}}
#'
#' @export
plot_validation_metrics <- function(validation_table,
                                    metrics = c("pac",
                                                "silhouette_mean",
                                                "ari_mean_between_imputations",
                                                "ari_consensus_mean",
                                                "calinski_harabasz_mean",
                                                "davies_bouldin_mean",
                                                "dunn_index",
                                                "score")) {

  # Auto-extract scores_table from mi_clustering_result or choose_best_clustering output
  if (inherits(validation_table, "mi_clustering_result")) {
    validation_table <- validation_table$scores_table

  } else if (is.list(validation_table) && !is.data.frame(validation_table)) {

    if (!is.null(validation_table$scores_table)) {
      validation_table <- validation_table$scores_table
    } else {
      stop("Could not extract a valid scores_table from the provided list.")
    }
  }

  if (!is.data.frame(validation_table)) {
    stop("validation_table must be a data.frame.")
  }

  if (!("k" %in% names(validation_table))) {
    stop("validation_table must contain column 'k'.")
  }

  available_metrics <- intersect(metrics, names(validation_table))

  if (length(available_metrics) == 0) {
    stop("None of the requested metrics are available in validation_table.")
  }

  old_par <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(old_par))

  n         <- length(available_metrics)
  nrow_plot <- ceiling(sqrt(n))
  ncol_plot <- ceiling(n / nrow_plot)

  graphics::par(mfrow = c(nrow_plot, ncol_plot))

  # labels for each metric
  metric_labels <- c(
    pac                          = "PAC",
    silhouette_mean              = "Silhouette",
    ari_mean_between_imputations = "ARI (between imputations)",
    ari_consensus_mean           = "ARI (consensus)",
    calinski_harabasz_mean       = "Calinski-Harabasz",
    davies_bouldin_mean          = "Davies-Bouldin",
    dunn_index                   = "Dunn Index",
    score                        = "Weighted Score"
  )

  for (metric in available_metrics) {

    y <- validation_table[[metric]]

    label <- if (!is.na(metric_labels[metric])) metric_labels[metric] else gsub("_", " ", metric)

    plot(validation_table$k, y,
         type = "b",
         pch  = 19,
         xlab = "Number of clusters (k)",
         ylab = label,
         main = label,
         xaxt = "n")

    graphics::axis(1, at = validation_table$k)

    best_idx <- which.max(
      if (metric %in% c("pac", "davies_bouldin_mean", "score")) -y else y
    )
    if (length(best_idx) == 0) next

    graphics::points(validation_table$k[best_idx], y[best_idx],
           pch = 19, cex = 1.4, col = "red")
  }
  invisible(NULL)
}

#' Plot a consensus clustering dendrogram
#'
#' @description
#' Displays the hierarchical clustering dendrogram derived from the consensus
#' dissimilarity matrix \eqn{1 - C}, where \eqn{C} is the co-assignment matrix
#' produced by \code{\link{consensus_clustering}}. Optionally draws colored
#' rectangles around the clusters defined by the requested \code{k}.
#'
#' @param consensus_result A named list as returned by
#'   \code{\link{consensus_clustering}} for a single \code{k}. Must contain
#'   at least \code{$hclust} (an \code{hclust} object) and \code{$k} (an
#'   integer specifying the number of clusters).
#' @param rect Logical. If \code{TRUE} (default), colored rectangles are drawn
#'   around the clusters via \code{\link[stats]{rect.hclust}}.
#' @param k Integer. The number of clusters to display. Only required when
#'   \code{consensus_result} contains results for multiple \code{k} values.
#'   Ignored if a single-k result is passed directly.
#' @param hang Numeric. The fraction of the plot height by which labels hang
#'   below the rest of the plot. A negative value (default \code{-1}) causes
#'   all labels to hang down from zero.
#' @param labels Logical or character vector. Controls the labels displayed
#'   at the leaves of the dendrogram. If \code{NULL} (default), observation
#'   labels are shown. If \code{FALSE}, no labels are displayed — recommended
#'   for large datasets where labels become unreadable. A character vector
#'   of the same length as the number of observations can also be supplied
#'   to use custom labels.
#'
#' @return Invisibly returns \code{NULL}. The function is called for its
#'   side effect of producing a plot.
#'
#' @details
#' The dendrogram is built from the \code{hclust} object stored in
#' \code{consensus_result$hclust}, which was computed by
#' \code{\link{consensus_clustering}} on the consensus dissimilarity matrix
#' \eqn{1 - C}. The colored rectangles highlight the \code{k} groups
#' obtained by cutting the dendrogram at the level that yields exactly
#' \code{k} clusters.
#'
#' This function is intended to be called on a single \code{k}. To compare
#' dendrograms across multiple values of \code{k}, call the function
#' iteratively and manage the graphics layout externally via
#' \code{\link[graphics]{par}}.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: consensus dendrogram with simulated results
#' # ------------------------------------------------------------
#' set.seed(123)
#' hc <- hclust(dist(matrix(rnorm(60), nrow = 20)), method = "ward.D2")
#' cons <- list(hclust = hc, k = 3)
#'
#' plot_consensus_dendrogram(cons)
#' plot_consensus_dendrogram(cons, rect = FALSE)
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: Dendrogram with mice
#' # ------------------------------------------------------------
#'
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#' set.seed(123)
#' imp    <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#' mild   <- as_mild_list(imp)
#' parts  <- cluster_imputations(mild, method = "ward.D2", k = 3)
#' cons   <- consensus_clustering(parts, k = 3)
#'
#' # Single dendrogram
#' plot_consensus_dendrogram(cons)
#'
#' # Without rectangles
#' plot_consensus_dendrogram(cons, rect = FALSE)
#'
#' # Compare multiple k values
#' parts_multi <- cluster_imputations(mild, method = "ward.D2", k = 2:4)
#' cons_multi  <- consensus_clustering(parts_multi)
#'
#' par(mfrow = c(1, 3), mar = c(3, 2, 2, 1))
#' for (k_i in 2:4) {
#'   plot_consensus_dendrogram(cons_multi[[paste0("k", k_i)]])
#' }
#' par(mfrow = c(1, 1))
#' }
#' }
#'
#' @seealso
#' \code{\link{consensus_clustering}}, \code{\link{plot_consensus_matrix}},
#' \code{\link{plot_validation_metrics}}
#'
#' @export
plot_consensus_dendrogram <- function(consensus_result,
                                      k = NULL,
                                      rect = TRUE,
                                      hang = -1,
                                      labels = NULL) {

  # --------------------------------------------------------
  # Auto-extract best_consensus_result from mi_clustering_result
  # --------------------------------------------------------
  if (inherits(consensus_result, "mi_clustering_result")) {
    consensus_result <- consensus_result$best_consensus_result
  }

  # --------------------------------------------------------
  # Auto-extract best_consensus_result from choose_best_clustering()
  # --------------------------------------------------------
  if (is.list(consensus_result) &&
      !is.null(consensus_result$best_consensus_result) &&
      is.null(consensus_result$coassignment)) {
    consensus_result <- consensus_result$best_consensus_result
  }

  # --------------------------------------------------------
  # Detect if multiple k (list of lists)
  # --------------------------------------------------------
  is_multi_k <- is.list(consensus_result) &&
    all(vapply(consensus_result, function(x) is.list(x) && !is.null(x$hclust), logical(1)))

  # --------------------------------------------------------
  # k: Available options
  # --------------------------------------------------------
  if (is_multi_k && is.null(k)) {
    available_k <- names(consensus_result)
    stop(paste0("Multiple k detected. Available options: ",
                paste(available_k, collapse = ", "),
                ". Please specify k."))
  }

  # --------------------------------------------------------
  # Case 1: multiple k
  # --------------------------------------------------------
  if (is_multi_k) {

    k_name <- paste0("k", k)

    if (!k_name %in% names(consensus_result))
      stop("Requested k not found in consensus_result.")

    consensus_result <- consensus_result[[k_name]]
  }

  # --------------------------------------------------------
  # Validate structure
  # --------------------------------------------------------
  if (is.null(consensus_result$hclust)) {
    stop("consensus_result must contain 'hclust'.")
  }

  if (is.null(consensus_result$k)) {
    stop("consensus_result must contain 'k'.")
  }

  # --------------------------------------------------------
  # Plot
  # --------------------------------------------------------
  # Build plot arguments dynamically
  plot_args <- list(
    x    = consensus_result$hclust,
    hang = hang,
    main = paste("Consensus dendrogram (k =", consensus_result$k, ")"),
    xlab = "",
    sub  = ""
  )

  # Only pass labels if explicitly specified by the user
  if (!is.null(labels)) {
    plot_args$labels <- labels
  }

  do.call(plot, plot_args)

  if (rect) {
    stats::rect.hclust(consensus_result$hclust,
                       k      = consensus_result$k,
                       border = 2:6)
  }
  invisible(NULL)
}

#' Standardize multiple imputation outputs into a unified list
#'
#' @description
#' Converts multiple imputation outputs from different packages into a
#' standardized list of completed datasets, where each element corresponds
#' to one imputed dataset. Supported formats include \code{mids} objects
#' from \pkg{mice}, long-format data frames, \code{amelia} objects,
#' \code{imputationList} objects from \pkg{mitools}, and plain lists of
#' data frames. After conversion, consistency and data quality checks are
#' applied across all imputed datasets.
#'
#' This function is designed as a preprocessing step for clustering
#' multiple imputed datasets, ensuring strict comparability across imputations.
#'
#' @param x An imputation object. Accepted classes and formats:
#' \itemize{
#'   \item \code{mids}: output of \code{mice::mice()}.
#'   \item \code{data.frame}: long-format output of
#'         \code{mice::complete(x, action = "long")}. Must contain a
#'         \code{.imp} column.
#'   \item \code{amelia}: output of \code{Amelia::amelia()}.
#'   \item \code{imputationList}: output of \code{mitools::imputationList()}.
#'   \item \code{list}: a plain list where every element is a
#'         \code{data.frame} representing one completed dataset.
#' }
#'
#' @return A named list of \code{data.frame} objects, one per imputed
#' dataset. All datasets are guaranteed to have identical dimensions and
#' column names, and to contain no \code{NA}, \code{NaN}, \code{Inf}, or
#' all equal values in a column.
#'
#' @details
#' The function performs the following consistency checks before returning:
#' \itemize{
#'   \item Identical number of rows across all imputed datasets.
#'   \item Identical number of columns across all imputed datasets.
#'   \item Identical column names across all imputed datasets.
#'   \item Absence of \code{NA}, \code{NaN}, and \code{Inf} values.
#'   \item Absence of numeric columns with all values equal,
#'         as such variables provide no information for distance-based
#'         clustering and may lead to degenerate solutions.
#' }
#' All datasets are coerced to base \code{data.frame} to avoid
#' compatibility issues with \code{tibble} or other subclasses.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Simple list of completed datasets
#' # ------------------------------------------------------------
#' set.seed(123)
#'
#' imp_list <- replicate(3, {
#'   data.frame(
#'     x = rnorm(10),
#'     y = rnorm(10)
#'   )
#' }, simplify = FALSE)
#'
#' mild <- as_mild_list(imp_list)
#'
#' length(mild)      # number of imputations
#' str(mild[[1]])    # structure of one dataset
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: List of completed datasets with mice
#' # ------------------------------------------------------------
#'
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#'   set.seed(123)
#'
#'   # Example 1: Using a mids object
#'   imp <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#'   mild <- as_mild_list(imp)
#'
#'   length(mild)        # number of imputations
#'   str(mild[[1]])      # structure of one completed dataset
#'
#'   # Example 2: Using long-format data
#'   long_df <- mice::complete(imp, action = "long", include = FALSE)
#'   mild2 <- as_mild_list(long_df)
#'
#'   length(mild2)
#' }
#' }
#'
#' @references
#' Rubin, D. B. (1976).
#' Inference and missing data.
#' \emph{Biometrika}, \strong{63}(3), 581-592.
#' \doi{10.1093/biomet/63.3.581}
#'
#' Rubin, D. B. (1987).
#' \emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.
#'
#' van Buuren, S., & Groothuis-Oudshoorn, K. (2011).
#' mice: Multivariate Imputation by Chained Equations in R.
#' \emph{Journal of Statistical Software}, \strong{45}(3), 1-67.
#' \doi{10.18637/jss.v045.i03}
#'
#' Honaker, J., King, G., & Blackwell, M. (2011).
#' Amelia II: A Program for Missing Data.
#' \emph{Journal of Statistical Software}, \strong{45}(7), 1-47.
#' \doi{10.18637/jss.v045.i07}
#'
#' @seealso \code{\link{cluster_imputations}}, \code{\link{consensus_clustering}}
#'
#' @export
as_mild_list <- function(x) {

  #Control verbosity
  verbose <- getOption("cclustr.verbose", FALSE)

  # --------------------------------------------------------
  # Case 1: 'mids' object (output of mice())
  # --------------------------------------------------------
  if (inherits(x, "mids")) {
    if (verbose) message("Detected: mice (mids object)")
    imp_list <- mice::complete(x, action = "all")
  }

  # --------------------------------------------------------
  # Case 2: mice Long format (output of mice::complete(..., "long"))
  # --------------------------------------------------------
  # Each imputation is identified by the '.imp' column
  else if (is.data.frame(x)) {

    if (!(".imp" %in% names(x))) {
      stop("The data.frame does not contain the '.imp' column.")
    }

    if (verbose) message("Detected: long-format imputations")

    # Identify imputation indices (excluding original incomplete data: .imp = 0)
    imp_ids <- sort(unique(x$.imp))
    imp_ids <- imp_ids[imp_ids != 0]  # delete original data

    # Split the long data.frame into a list of completed datasets
    imp_list <- lapply(imp_ids, function(i) {
      subset(x, .imp == i)[ , !(names(x) %in% c(".imp", ".id")), drop = FALSE]
    })

    # Assign standardized names to each imputed dataset
    names(imp_list) <- paste0("imp", imp_ids)
  }

  # --------------------------------------------------------
  # Case 3: 'amelia' object (output of amelia())
  # --------------------------------------------------------
  else if (inherits(x, "amelia")) {
    if (verbose) message("Detected: Amelia object")
    imp_list <- x$imputations
  }

  #-------------------------------------------------------
  # Case 4: 'imputationList' object (output of mitools())
  #-------------------------------------------------------
  else if (inherits(x, "imputationList")) {
    if (verbose) message("Detected: mitools imputationList")
    imp_list <- x$imputations
  }

  # --------------------------------------------------------
  # Case 5: already a list of completed datasets
  # --------------------------------------------------------
  # Each element must be a completed dataset
  else if (is.list(x)) {

    if (!all(vapply(x, is.data.frame, logical(1)))) {
      stop("The input list must contain only data.frames.")
    }

    imp_list <- x
  }

  #-------------------------------------------------------
  # Case 6: Unsupported input type
  #-------------------------------------------------------
  else {
    stop(
      paste(
        "Unsupported imputation format.\n\n",
        "Supported inputs are:\n",
        " - mids object (mice)\n",
        " - long-format data.frame from mice::complete(..., 'long')\n",
        " - amelia object (Amelia package)\n",
        " - imputationList object (mitools package)\n",
        " - list of completed data.frames\n\n"
        )
    )
  }

  #-------------------------------------------------------
  # Final consistency checks
  #-------------------------------------------------------
  # All imputations must have the same number of rows
  if (length(unique(vapply(imp_list, nrow, integer(1)))) != 1) {
    stop("All imputations must have the same number of rows.")
  }

  # All imputations must have the same number of columns
  if (length(unique(vapply(imp_list, ncol, integer(1)))) != 1) {
    stop("All imputations must have the same number of columns.")
  }

  # All imputations must have identical column names
  ref_names <- names(imp_list[[1]])
  if (!all(vapply(imp_list, function(df) identical(names(df), ref_names), logical(1)))) {
    stop("All imputations must have identical column names.")
  }

  # Convert to data.frame (avoid tibble issues)
  imp_list <- lapply(imp_list, as.data.frame)

  #-------------------------------------------------------
  # check missing, infinite and equal values
  #-------------------------------------------------------

  for (i in seq_along(imp_list)) {

    df <- imp_list[[i]]

    # 1) Check NA values
    if (anyNA(df)) {
      stop(paste("Imputation", i, "contains NA values."))
    }

    # 2) Check NaN values
    if (any(is.nan(as.matrix(df)))) {
      stop(paste("Imputation", i, "contains NaN values."))
    }

    # 3) Check Inf values
    if (any(is.infinite(as.matrix(df)))) {
      stop(paste("Imputation", i, "contains Inf values."))
    }

    # 4) Check columns with the same values
    same_cols <- names(df)[
      vapply(names(df), function(nm) {
        col <- df[[nm]]
        is.numeric(col) && length(unique(col)) == 1
      }, logical(1))
    ]
    if (length(same_cols) > 0) {
      const_vals <- sapply(same_cols, function(nm) unique(df[[nm]]))
      details <- paste(same_cols, "=", const_vals, collapse = ", ")
      stop(paste("Imputation", i,
                 "has constant numeric columns (all values identical):",
                 details))
    }
  }

  if (verbose) {
    message("All datasets passed validation")
    message(paste("Created list of", length(imp_list), "imputed datasets"))
  }

  # Return the standardized list of imputed datasets
  if (is.null(names(imp_list))) {
    names(imp_list) <- paste0("imp", seq_along(imp_list))
  }

  return(imp_list)
}

# tests/testthat/test-imputation.R
# Tests for as_mild_list()

# -------------------------------------------------------
# Helper: valid list of data.frames with no NAs
# -------------------------------------------------------
make_imp <- function(n = 15, m = 3, seed = 1) {
  set.seed(seed)
  replicate(m, data.frame(x = rnorm(n), y = rnorm(n)), simplify = FALSE)
}

# ============================================================
# Output structure
# ============================================================

test_that("returns a named list of data.frames", {
  result <- as_mild_list(make_imp())
  expect_type(result, "list")
  expect_true(all(vapply(result, is.data.frame, logical(1))))
  expect_false(is.null(names(result)))
})

test_that("number of elements matches the number of input imputations", {
  result <- as_mild_list(make_imp(m = 5))
  expect_length(result, 5)
})

test_that("all datasets have the same dimensions", {
  result <- as_mild_list(make_imp(n = 20, m = 4))
  nrows  <- vapply(result, nrow, integer(1))
  ncols  <- vapply(result, ncol, integer(1))
  expect_true(length(unique(nrows)) == 1)
  expect_true(length(unique(ncols)) == 1)
})

test_that("all datasets have identical column names", {
  result    <- as_mild_list(make_imp())
  col_names <- lapply(result, names)
  expect_true(all(vapply(col_names, identical, logical(1), col_names[[1]])))
})

test_that("datasets are base data.frames, not tibbles", {
  result <- as_mild_list(make_imp())
  expect_true(all(vapply(result,
                         function(df) identical(class(df), "data.frame"),
                         logical(1))))
})

# ============================================================
# Long-format input with .imp column
# ============================================================

test_that("accepts a long-format data.frame with .imp column", {
  set.seed(42)
  long_df <- data.frame(
    .imp = rep(1:3, each = 10),
    x    = rnorm(30),
    y    = rnorm(30)
  )
  result <- as_mild_list(long_df)
  expect_length(result, 3)
  expect_true(all(vapply(result, is.data.frame, logical(1))))
})

test_that("excludes rows with .imp == 0 (original incomplete data)", {
  set.seed(42)
  long_df <- data.frame(
    .imp = rep(0:2, each = 10),
    x    = rnorm(30),
    y    = rnorm(30)
  )
  result <- as_mild_list(long_df)
  expect_length(result, 2)
})

test_that("drops .imp and .id columns from the output", {
  set.seed(42)
  long_df <- data.frame(
    .imp = rep(1:2, each = 10),
    .id  = rep(1:10, 2),
    x    = rnorm(20),
    y    = rnorm(20)
  )
  result <- as_mild_list(long_df)
  expect_false(".imp" %in% names(result[[1]]))
  expect_false(".id"  %in% names(result[[1]]))
})

# ============================================================
# Data quality checks
# ============================================================

test_that("fails if any imputation contains NA", {
  imp <- make_imp()
  imp[[2]][3, 1] <- NA
  expect_error(as_mild_list(imp), "NA")
})

test_that("fails if any imputation contains NaN", {
  imp <- make_imp()
  imp[[1]][1, 1] <- NaN
  expect_error(as_mild_list(imp), "NA values")
})

test_that("fails if any imputation contains Inf", {
  imp <- make_imp()
  imp[[1]][2, 2] <- Inf
  expect_error(as_mild_list(imp), "Inf")
})

test_that("fails if a numeric column is constant across all rows", {
  imp <- make_imp()
  imp[[1]]$x <- 5   # constant column
  expect_error(as_mild_list(imp), "constant")
})

test_that("fails if imputations have different numbers of rows", {
  set.seed(1)
  imp <- list(
    data.frame(x = rnorm(10)),
    data.frame(x = rnorm(12))
  )
  expect_error(as_mild_list(imp), "rows")
})

test_that("fails if imputations have different numbers of columns", {
  set.seed(1)
  imp <- list(
    data.frame(x = rnorm(10), y = rnorm(10)),
    data.frame(x = rnorm(10))
  )
  expect_error(as_mild_list(imp), "columns")
})

test_that("fails if imputations have different column names", {
  set.seed(1)
  imp <- list(
    data.frame(x = rnorm(10), y = rnorm(10)),
    data.frame(x = rnorm(10), z = rnorm(10))
  )
  expect_error(as_mild_list(imp), "column names")
})

# ============================================================
# Unsupported input types
# ============================================================

test_that("fails with unsupported input types", {
  expect_error(as_mild_list(42),      "Unsupported")
  expect_error(as_mild_list("text"),  "Unsupported")
  expect_error(as_mild_list(TRUE),    "Unsupported")
})

test_that("fails if the list contains non-data.frame elements", {
  bad_list <- list(data.frame(x = 1:5), matrix(1:5))
  expect_error(as_mild_list(bad_list), "data.frames")
})

test_that("fails with a data.frame that has no .imp column", {
  df <- data.frame(x = 1:10, y = rnorm(10))
  expect_error(as_mild_list(df), "\\.imp")
})

#' Plot the consensus co-assignment matrix
#'
#' @description
#' Displays a heatmap of the co-assignment matrix \eqn{C} produced by
#' \code{\link{consensus_clustering}}, where each cell \eqn{C_{ij}} represents
#' the weighted proportion of partitions in which observations \eqn{i} and
#' \eqn{j} were assigned to the same cluster. Values close to 1 indicate
#' that two observations are consistently grouped together across all
#' imputed datasets; values close to 0 indicate the opposite. Rows and
#' columns can be reordered by consensus cluster assignment to make the
#' block structure visually apparent.
#'
#' @param consensus_result A named list as returned by
#'   \code{\link{consensus_clustering}} for a single \code{k}. Must contain
#'   at least \code{$coassignment} (a numeric \eqn{n \times n} matrix) and,
#'   when \code{reorder = TRUE}, \code{$consensus} (an integer vector of
#'   cluster labels).
#' @param reorder Logical. If \code{TRUE} (default), rows and columns are
#'   reordered by the consensus cluster assignment stored in
#'   \code{consensus_result$consensus}, making the block structure of the
#'   matrix visually apparent.
#' @param show_labels Logical. If \code{TRUE}, observation indices are
#'   displayed on both axes, ordered according to the consensus cluster
#'   assignment. Recommended only for small datasets (n < 50) as labels
#'   become unreadable for large \eqn{n}. If \code{FALSE} (default),
#'   tick marks are shown without labels to preserve the axis frame
#'   without visual clutter.
#' @param viridis_option A single character specifying the color palette
#'   passed to \code{\link[viridisLite]{viridis}}. Accepted values are
#'   \code{"A"} (magma), \code{"B"} (inferno), \code{"C"} (plasma), and
#'   \code{"D"} (viridis, default). Low co-assignment values are mapped to
#'   dark colors and high values to bright colors.
#'
#' @return Invisibly returns \code{NULL}. The function is called for its
#'   side effect of producing a plot.
#'
#' @details
#' The matrix is rendered using \code{\link[graphics]{image}} on the
#' consensus dissimilarity \eqn{1 - C} reordered by cluster membership.
#' Original graphical parameters are restored on exit via
#' \code{\link[graphics]{par}}.
#'
#' A well-defined consensus solution produces a block-diagonal pattern in
#' the heatmap, where bright blocks along the diagonal correspond to
#' observations that are consistently assigned to the same cluster, and
#' dark off-diagonal regions indicate observations that are rarely grouped
#' together.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Consensus matrix with simulated consensus result
#' # ------------------------------------------------------------
#' set.seed(123)
#' n <- 20
#' true_labels <- rep(1:2, each = n / 2)
#' M <- matrix(0.1, n, n)
#' for (i in seq_len(n))
#'   for (j in seq_len(n))
#'     if (true_labels[i] == true_labels[j]) M[i, j] <- 0.9
#'
#' cons <- list(coassignment = M,
#'              consensus    = true_labels,
#'              k            = 2)
#'
#' plot_consensus_matrix(cons, viridis_option = "A")
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: Consensus matrix with mice
#' # ------------------------------------------------------------
#'
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#' imp   <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#' mild  <- as_mild_list(imp)
#' parts <- cluster_imputations(mild, method = "ward.D2", k = 3)
#' cons  <- consensus_clustering(parts, k = 3)
#'
#' # Default plot with viridis palette
#' plot_consensus_matrix(cons)
#'
#' # Without reordering
#' plot_consensus_matrix(cons, reorder = FALSE)
#'
#' # Compare color palettes
#' par(mfrow = c(2, 2), mar = c(4, 4, 2, 2))
#' for (opt in c("A", "B", "C", "D")) {
#'   plot_consensus_matrix(cons, viridis_option = opt)
#' }
#' par(mfrow = c(1, 1))
#' }
#' }
#'
#' @seealso
#' \code{\link{consensus_clustering}}, \code{\link{plot_consensus_dendrogram}},
#' \code{\link{plot_validation_metrics}}
#'
#' @export
plot_consensus_matrix <- function(consensus_result,
                                  reorder        = TRUE,
                                  show_labels    = FALSE,
                                  viridis_option = "D") {

  # Auto-extract best_consensus_result from mi_clustering_result
  if (inherits(consensus_result, "mi_clustering_result")) {
    consensus_result <- consensus_result$best_consensus_result
  }

  # Auto-extract best_consensus_result from choose_best_clustering()
  if (is.list(consensus_result) &&
      !is.null(consensus_result$best_consensus_result) &&
      is.null(consensus_result$coassignment)) {
    consensus_result <- consensus_result$best_consensus_result
  }

  # Validate that the co-assignment matrix is present
  if (is.null(consensus_result$coassignment)) {
    stop("consensus_result must contain 'coassignment'.")
  }

  M <- consensus_result$coassignment

  # Ensure the co-assignment object is a matrix
  if (!is.matrix(M)) {
    stop("'coassignment' must be a matrix.")
  }

  # Reorder rows and columns by consensus cluster assignment
  # to make the block-diagonal structure visually apparent
  ord <- seq_len(nrow(M))

  if (reorder) {
    if (!is.null(consensus_result$consensus)) {
      ord <- order(consensus_result$consensus)
    }
  }

  M_ord <- M[ord, ord, drop = FALSE]

  # Save and restore graphical parameters on exit
  op <- graphics::par(no.readonly = TRUE)
  on.exit(graphics::par(op))

  # Delete axes
  graphics::par(mar  = c(4, 4, 2, 2),
                xaxt = "n",
                yaxt = "n")

  # Validate the viridis color palette option
  if (!viridis_option %in% c("A", "B", "C", "D", "E", "F", "G", "H"))
    stop("'viridis_option' must be one of 'A', 'B', 'C', 'D', 'E', 'F', 'G', or 'H'.")

  # Render the heatmap using image()
  graphics::image(seq_len(nrow(M_ord)), seq_len(ncol(M_ord)), t(M_ord[rev(seq_len(nrow(M_ord))), ]),
                  col  = viridisLite::viridis(100, option = viridis_option),
                  axes = FALSE,
                  xlab = "Observations",
                  ylab = "Observations",
                  main = paste("Consensus matrix (k =", consensus_result$k, ")"))

  if (show_labels) {
    # Restore axes to draw them with axis()
    graphics::par(xaxt = "s", yaxt = "s")
    graphics::axis(1, at = seq_len(nrow(M_ord)), labels = ord,      las = 2, cex.axis = 0.4)
    graphics::axis(2, at = seq_len(ncol(M_ord)), labels = rev(ord), las = 2, cex.axis = 0.4)
  } else {
    # Draw axes without any tick marks or labels for a clean appearance
    graphics::axis(1, at = seq_len(nrow(M_ord)), labels = FALSE, tick = FALSE)
    graphics::axis(2, at = seq_len(ncol(M_ord)), labels = FALSE, tick = FALSE)
  }

  graphics::box()
  invisible(NULL)
}

#' Run the full multiple-imputation clustering pipeline
#'
#' @description
#' Executes the complete multiple-imputation clustering pipeline in a single
#' call, integrating imputation standardization, per-imputation clustering,
#' consensus construction, validation, and optimal \code{k} selection.
#' Designed as a convenience wrapper around \code{\link{as_mild_list}},
#' \code{\link{cluster_imputations}}, \code{\link{consensus_clustering}},
#' \code{\link{validate_clustering}}, and \code{\link{choose_best_clustering}}.
#' When a single value of \code{k} is supplied, the selection step is skipped
#' and the unique solution is returned directly.
#'
#' @param data An imputation object accepted by \code{\link{as_mild_list}}.
#'   Supported formats include \code{mids} objects from \pkg{mice},
#'   long-format data frames with a \code{.imp} column, \code{amelia} objects,
#'   \code{imputationList} objects from \pkg{mitools}, and plain lists of
#'   \code{data.frame}s.
#' @param method A character string specifying the clustering algorithm applied
#'   to each imputed dataset. For hierarchical clustering, accepted values are
#'   \code{"ward.D"}, \code{"ward.D2"}, \code{"single"}, \code{"complete"},
#'   \code{"average"}, \code{"centroid"}, \code{"median"}, and
#'   \code{"mcquitty"}. Additional options are \code{"kmeans"}, \code{"pam"},
#'   \code{"fuzzy"}, \code{"mclust"}, \code{"kmodes"}, and
#'   \code{"kprototypes"}. Default is \code{"ward.D2"}.
#' @param k A single integer or an integer vector specifying the number(s) of
#'   clusters to evaluate. If a single value is supplied, the pipeline skips
#'   the \code{\link{choose_best_clustering}} step and returns that solution
#'   directly. If a vector is supplied, all values are evaluated and the
#'   optimal \code{k} is selected via weighted rank aggregation.
#' @param scale_data A character string specifying how numeric columns are
#'   standardized prior to clustering. Accepted values are \code{"global"}
#'   (default, pooled mean and standard deviation computed across all
#'   imputations without stacking), \code{"within"} (each imputed dataset
#'   scaled independently), and \code{"none"} (no scaling applied). Scaling
#'   is automatically disabled for \code{distance = "gower"},
#'   \code{"jaccard"}, or \code{"simple_matching"}, and for
#'   \code{method = "kmodes"} and \code{"kprototypes"}.
#' @param distance A character string specifying the distance metric used for
#'   hierarchical clustering and \code{method = "pam"}. Options are
#'   \code{"euclidean"} (default), \code{"manhattan"}, \code{"gower"}
#'   (via \code{cluster::daisy()}, supports mixed data), \code{"jaccard"}
#'   (binary data, requires \pkg{proxy}), \code{"simple_matching"} (binary
#'   data, requires \pkg{proxy}), and \code{"custom"} (user-defined function
#'   via \code{dist_fun}). Ignored for \code{method = "kmeans"},
#'   \code{"fuzzy"}, \code{"mclust"}, \code{"kmodes"}, and
#'   \code{"kprototypes"}, which compute distances internally.
#' @param dist_fun A function that takes a \code{data.frame} and returns a
#'   \code{dist} object. Required when \code{distance = "custom"};
#'   ignored otherwise.
#' @param dist_args A named list of additional arguments passed to the
#'   distance function. For \code{distance = "gower"}, forwarded to
#'   \code{cluster::daisy()}; for \code{"jaccard"} or
#'   \code{"simple_matching"}, to \code{proxy::dist()}; for
#'   \code{"euclidean"} or \code{"manhattan"}, to \code{stats::dist()};
#'   also for \code{"custom"}. Default is \code{list()}.
#' @param consensus_method A character string specifying the method used to
#'   build the co-assignment matrix. \code{"classic"} assigns equal weight to
#'   all partitions; \code{"weighted_ari"} weights partitions by their
#'   pairwise ARI centrality. Default is \code{"classic"}.
#' @param cluster_method_consensus A character string specifying the
#'   agglomeration method passed to \code{\link[stats]{hclust}} during the
#'   consensus stage. Accepted values are the same as for \code{method} when
#'   using hierarchical clustering. Default is \code{"ward.D2"}.
#' @param pac_lower A numeric value between 0 and 1 specifying the lower bound
#'   of the ambiguous assignment region for the PAC metric. Default is
#'   \code{0.1}.
#' @param pac_upper A numeric value between 0 and 1 specifying the upper bound
#'   of the ambiguous assignment region for the PAC metric. Default is
#'   \code{0.9}.
#' @param weights A named numeric vector specifying the relative importance of
#'   each metric in the rank aggregation used by
#'   \code{\link{choose_best_clustering}}. Expected names are \code{pac},
#'   \code{silhouette}, \code{ari_between}, \code{ari_consensus}, \code{ch},
#'   \code{db}, and \code{dunn}. If \code{NULL} (default), weights are set
#'   automatically based on \code{prefer_stability}. Ignored when \code{k} is
#'   a single value.
#' @param prefer_stability Logical. If \code{TRUE} (default), stability
#'   metrics (\code{pac}, \code{ari_between}, \code{ari_consensus}) receive
#'   higher weights during \code{k} selection. If \code{FALSE}, internal
#'   compactness metrics (\code{silhouette}, \code{ch}, \code{db},
#'   \code{dunn}) receive higher weights. Ignored when \code{k} is a single
#'   value or when \code{weights} is provided explicitly.
#' @param tie_breaker A character string specifying the secondary metric used
#'   to break ties during \code{k} selection. One of \code{"silhouette"}
#'   (default), \code{"pac"}, \code{"dunn"}, \code{"ch"}, \code{"db"},
#'   \code{"ari_between"}, or \code{"ari_consensus"}. Ignored when \code{k}
#'   is a single value.
#' @param ... Additional arguments passed to the underlying clustering
#'   function via \code{\link{cluster_imputations}}: \code{stats::kmeans},
#'   \code{e1071::cmeans}, \code{mclust::Mclust}, \code{klaR::kmodes}, or
#'   \code{clustMixType::kproto}. Not used for \code{method = "pam"}.
#'
#' @return An object of class \code{"mi_clustering_result"}: a named list
#'   with the following elements:
#'   \itemize{
#'     \item \code{call}: the matched call.
#'     \item \code{input_k}: the value(s) of \code{k} supplied by the user.
#'     \item \code{clustering_method}: character string with the clustering
#'           method used.
#'     \item \code{consensus_method}: character string with the consensus
#'           method used.
#'     \item \code{scale_data}: A character string specifying how numeric columns
#'           are standardized prior to clustering
#'     \item \code{imputations}: standardized list of imputed datasets, as
#'           returned by \code{\link{as_mild_list}}.
#'     \item \code{partitions}: clustering assignments per imputed dataset,
#'           as returned by \code{\link{cluster_imputations}}.
#'     \item \code{consensus_results}: consensus clustering output, as
#'           returned by \code{\link{consensus_clustering}}.
#'     \item \code{validation_table}: validation metrics per \code{k}, as
#'           returned by \code{\link{validate_clustering}}.
#'     \item \code{best_k}: integer, the selected optimal number of clusters.
#'     \item \code{best_consensus}: integer vector of consensus cluster labels
#'           for the selected \code{k}, one per observation.
#'     \item \code{best_coassignment}: numeric matrix of co-assignment
#'           probabilities for the selected \code{k}.
#'     \item \code{best_consensus_result}: the full consensus result object
#'           for the selected \code{k}.
#'     \item \code{scores_table}: validation table with an additional
#'           \code{score} column when \code{k} is a vector; the plain
#'           validation table when \code{k} is a single value.
#'     \item \code{selection}: the full output of
#'           \code{\link{choose_best_clustering}} when \code{k} is a vector,
#'           or a minimal equivalent list when \code{k} is a single value.
#'   }
#'
#' @details
#' The pipeline proceeds as follows:
#' \enumerate{
#'   \item \code{\link{as_mild_list}} standardizes the imputation object and
#'         validates data quality.
#'   \item \code{\link{cluster_imputations}} applies the chosen algorithm to
#'         each imputed dataset, optionally scaling the data beforehand.
#'   \item \code{\link{consensus_clustering}} builds a co-assignment matrix
#'         and derives the consensus partition for each requested \code{k}.
#'   \item \code{\link{validate_clustering}} computes internal and stability
#'         metrics for each candidate \code{k}.
#'   \item \code{\link{choose_best_clustering}} selects the optimal \code{k}
#'         via weighted rank aggregation (only when \code{length(k) > 1}).
#' }
#'
#' When \code{length(k) == 1}, step 5 is bypassed and the unique solution
#' is returned directly, with \code{weights} and \code{tie_breaker} set to
#' \code{NULL} in the \code{selection} element.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Simulated list of imputed datasets
#' # ------------------------------------------------------------
#' set.seed(123)
#' make_imputation <- function() {
#'   data.frame(
#'     x1 = rnorm(30),
#'     x2 = rnorm(30),
#'     x3 = rnorm(30)
#'   )
#' }
#'
#' imp_list <- lapply(1:5, function(i) make_imputation())
#'
#' # Single k
#' res <- run_mi_clustering(imp_list, method = "ward.D2", k = 3)
#' res$best_k
#' res$best_consensus
#'
#' # Range of k values (automatic selection)
#' res_multi <- run_mi_clustering(imp_list, method = "ward.D2", k = 2:4)
#' res_multi$best_k
#' res_multi$validation_table
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: Full pipeline with mice
#' # ------------------------------------------------------------
#'if (requireNamespace("mice", quietly = TRUE)) {
#'
#' set.seed(123)
#' imp <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
#'
#' # Single k
#' res <- run_mi_clustering(imp, method = "ward.D2", k = 3)
#' res$best_k
#' res$best_consensus
#'
#' # Range of k values (automatic selection)
#' res_multi <- run_mi_clustering(imp, method = "ward.D2", k = 2:5)
#' res_multi$best_k
#' res_multi$validation_table
#'
#' # PAM with Gower distance (mixed data)
#' res_gower <- run_mi_clustering(imp, method = "pam", k = 2:4,
#'                                distance = "gower")
#'
#' # ARI-weighted consensus, compactness-focused selection
#' res_w <- run_mi_clustering(imp, method = "ward.D2", k = 2:5,
#'                            consensus_method  = "weighted_ari",
#'                            prefer_stability  = FALSE,
#'                            tie_breaker       = "dunn")
#' }
#' }
#'
#' @references
#' Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
#' Consensus clustering: A resampling-based method for class discovery
#' and visualization of gene expression microarray data.
#' \emph{Machine Learning}, \strong{52}(1-2), 91-118.
#' \doi{10.1023/A:1023949509487}
#'
#' Rubin, D. B. (1987).
#' \emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.
#'
#' Pihur, V., Datta, S., & Datta, S. (2007).
#' Weighted rank aggregation of cluster validation measures.
#' \emph{Bioinformatics}, \strong{23}(13), 1607-1615.
#' \doi{10.1093/bioinformatics/btm158}
#'
#' @seealso
#' \code{\link{as_mild_list}}, \code{\link{cluster_imputations}},
#' \code{\link{consensus_clustering}}, \code{\link{validate_clustering}},
#' \code{\link{choose_best_clustering}}
#'
#' @export
run_mi_clustering <- function(data,
                              method = "ward.D2",
                              k,
                              scale_data = "global",
                              distance = "euclidean",
                              dist_fun = NULL,
                              dist_args = list(),
                              consensus_method = c("classic", "weighted_ari"),
                              cluster_method_consensus = "ward.D2",
                              pac_lower = 0.1,
                              pac_upper = 0.9,
                              weights = NULL,
                              prefer_stability = TRUE,
                              tie_breaker = c("silhouette", "pac", "dunn", "ch", "db",
                                              "ari_between", "ari_consensus"),
                              ...) {

  # --------------------------------------------------------
  # Match arguments
  # --------------------------------------------------------
  consensus_method <- match.arg(consensus_method)
  tie_breaker <- match.arg(tie_breaker)

  # --------------------------------------------------------
  # 1) Standardize imputations
  # --------------------------------------------------------
  imp_list <- as_mild_list(data)

  # --------------------------------------------------------
  # 2) Cluster each imputation
  # --------------------------------------------------------
  partitions <- cluster_imputations(
    imp_list   = imp_list,
    method     = method,
    k          = k,
    scale_data = scale_data,
    distance   = distance,
    dist_fun   = dist_fun,
    dist_args  = dist_args,
    ...
  )

  # --------------------------------------------------------
  # 3) Consensus clustering
  # --------------------------------------------------------
  consensus_results <- consensus_clustering(
    partitions = partitions,
    k = if (length(k) == 1) k else NULL,
    cluster_method = cluster_method_consensus,
    consensus_method = consensus_method
  )

  # --------------------------------------------------------
  # 4) Validation
  # --------------------------------------------------------
  validation_table <- validate_clustering(
    partitions = partitions,
    consensus_results = consensus_results,
    pac_lower = pac_lower,
    pac_upper = pac_upper
  )

  # --------------------------------------------------------
  # 5) Choose best solution (solo si k es un rango)
  # --------------------------------------------------------
  if (length(k) > 1) {
    best_solution <- choose_best_clustering(
      validation_table = validation_table,
      consensus_results = consensus_results,
      weights = weights,
      prefer_stability = prefer_stability,
      tie_breaker = tie_breaker
    )
  } else {
    # Con k único, la "mejor" solución es la única disponible
    best_solution <- list(
      best_k             = k,
      best_consensus     = consensus_results$consensus,
      best_coassignment  = consensus_results$coassignment,
      best_consensus_result = consensus_results,
      scores_table       = validation_table,
      weights            = NULL,
      tie_breaker        = NULL
    )
  }
  # --------------------------------------------------------
  # Return all results
  # --------------------------------------------------------
  out <- list(
    call = match.call(),
    input_k = k,
    clustering_method = method,
    consensus_method = consensus_method,
    scale_data = scale_data,
    imputations = imp_list,
    partitions = partitions,
    consensus_results = consensus_results,
    validation_table = validation_table,
    best_k = best_solution$best_k,
    best_consensus = best_solution$best_consensus,
    best_coassignment = best_solution$best_coassignment,
    best_consensus_result = best_solution$best_consensus_result,
    scores_table = best_solution$scores_table,
    selection = best_solution
  )

  class(out) <- "mi_clustering_result"
  return(out)
}

#' Validate consensus clustering results across multiple imputed datasets
#'
#' @description
#' Computes a comprehensive set of internal and stability validation metrics
#' for consensus clustering results obtained from multiple imputed datasets.
#' All internal metrics are computed on the consensus dissimilarity matrix
#' \eqn{1 - C}, where \eqn{C} is the co-assignment matrix, making the
#' function robust to numeric, categorical, and mixed data types. Supports
#' both single-k and multi-k evaluation.
#'
#' @param partitions A list of cluster assignment vectors (one per imputed
#'   dataset) for a single \code{k}, or a named list of such lists (one per
#'   \code{k} value, named \code{"k2"}, \code{"k3"}, etc.) as returned by
#'   \code{\link{cluster_imputations}}.
#' @param consensus_results A list as returned by
#'   \code{\link{consensus_clustering}}, either for a single \code{k} or
#'   for multiple \code{k} values. Must contain at least \code{$consensus}
#'   and \code{$coassignment} for each \code{k}.
#' @param pac_lower A numeric value between 0 and 1 specifying the lower
#'   bound of the ambiguous assignment region for the PAC metric. Default
#'   is \code{0.1}.
#' @param pac_upper A numeric value between 0 and 1 specifying the upper
#'   bound of the ambiguous assignment region for the PAC metric. Default
#'   is \code{0.9}.
#'
#' @return A \code{data.frame} with one row per evaluated \code{k} and the
#'   following columns:
#'   \itemize{
#'     \item \code{k}: number of clusters evaluated.
#'     \item \code{pac}: Proportion of Ambiguous Clusterings. Lower values
#'           indicate a more stable consensus. Computed from the upper
#'           triangle of the co-assignment matrix.
#'     \item \code{silhouette_mean}: mean silhouette width computed on the
#'           consensus dissimilarity \eqn{1 - C}. Higher values indicate
#'           better-defined clusters.
#'     \item \code{ari_mean_between_imputations}: mean pairwise Adjusted
#'           Rand Index (ARI) across all pairs of imputed partitions.
#'           Higher values indicate greater clustering stability.
#'     \item \code{ari_consensus_mean}: mean ARI between the consensus
#'           partition and each individual imputed partition. Higher values
#'           indicate that the consensus is representative of the ensemble.
#'     \item \code{calinski_harabasz_mean}: Calinski-Harabasz index computed
#'           via \code{fpc::cluster.stats()}. Higher values indicate more
#'           compact and well-separated clusters.
#'     \item \code{davies_bouldin_mean}: Davies-Bouldin index computed using
#'           a medoid-based approach on the consensus dissimilarity. Lower
#'           values indicate better cluster separation.
#'     \item \code{dunn_index}: Dunn index computed via
#'           \code{fpc::cluster.stats()}. Higher values indicate better
#'           cluster compactness and separation.
#'   }
#'
#' @details
#' All internal validation metrics (silhouette, Calinski-Harabasz,
#' Davies-Bouldin, Dunn) are computed exclusively on the consensus
#' dissimilarity matrix \eqn{1 - C} rather than on the original feature
#' space. This design choice ensures compatibility with any data type and
#' avoids recomputing distances from the raw imputed datasets.
#'
#' The Davies-Bouldin index is computed using a custom medoid-based
#' implementation derived from the consensus dissimilarity, as no standard
#' R implementation accepts a precomputed dissimilarity matrix directly.
#'
#' Helper functions prefixed with a dot (e.g., \code{.pac_from_coassignment},
#' \code{.mean_silhouette_from_diss}) are internal and not exported.
#'
#' @examples
#' # ------------------------------------------------------------
#' # Example 1: Basic validation with simulated partitions
#' # ------------------------------------------------------------
#' set.seed(123)
#'
#' # Simulate partitions (3 imputations, k = 2)
#' parts <- list(
#'   imp1 = sample(1:2, 20, replace = TRUE),
#'   imp2 = sample(1:2, 20, replace = TRUE),
#'   imp3 = sample(1:2, 20, replace = TRUE)
#' )
#'
#' cons <- consensus_clustering(parts, k = 2)
#' val  <- validate_clustering(parts, cons)
#'
#' print(val)
#'
#' \donttest{
#' # ------------------------------------------------------------
#' # Example 2: Full pipeline with mice (multiple k)
#' # ------------------------------------------------------------
#' if (requireNamespace("mice", quietly = TRUE)) {
#'
#'   set.seed(123)
#'
#'   imp  <- mice::mice(mice::nhanes, m = 3, printFlag = FALSE)
#'   mild <- as_mild_list(imp)
#'
#'   parts <- cluster_imputations(mild, method = "ward.D2", k = 2:3)
#'   cons  <- consensus_clustering(parts)
#'
#'   val <- validate_clustering(parts, cons)
#'
#'   print(val)
#' }
#' }
#'
#' @references
#' Senbabaoglu, Y., Michailidis, G., & Li, J. Z. (2014).
#' Critical limitations of consensus clustering in class discovery.
#' \emph{Scientific Reports}, \strong{4}, 6207.
#' \doi{10.1038/srep06207}
#'
#' Rousseeuw, P. J. (1987).
#' Silhouettes: A graphical aid to the interpretation and validation
#' of cluster analysis.
#' \emph{Journal of Computational and Applied Mathematics}, \strong{20}, 53-65.
#' \doi{10.1016/0377-0427(87)90125-7}
#'
#' Calinski, T., & Harabasz, J. (1974).
#' A dendrite method for cluster analysis.
#' \emph{Communications in Statistics}, \strong{3}, 1-27.
#' \doi{10.1080/03610927408827101}
#'
#' Davies, D. L., & Bouldin, D. W. (1979).
#' A cluster separation measure.
#' \emph{IEEE Transactions on Pattern Analysis and Machine Intelligence},
#' \strong{PAMI-1}, 224-227.
#' \doi{10.1109/TPAMI.1979.4766909}
#'
#' Dunn, J. C. (1974).
#' Well-separated clusters and optimal fuzzy partitions.
#' \emph{Journal of Cybernetics}, \strong{4}, 95-104.
#' \doi{10.1080/01969727408546059}
#'
#' Hubert, L., & Arabie, P. (1985).
#' Comparing partitions.
#' \emph{Journal of Classification}, \strong{2}, 193-218.
#' \doi{10.1007/BF01908075}
#'
#' @seealso \code{\link{consensus_clustering}}, \code{\link{choose_best_clustering}}
#'
#' @export
validate_clustering <- function(partitions,
                                consensus_results,
                                pac_lower = 0.1,
                                pac_upper = 0.9) {

  # Required packages
  if (!requireNamespace("mclust", quietly = TRUE)) {
    stop("Package 'mclust' is required.")
  }
  if (!requireNamespace("cluster", quietly = TRUE)) {
    stop("Package 'cluster' is required.")
  }
  if (!requireNamespace("fpc", quietly = TRUE)) {
    stop("Package 'fpc' is required.")
  }

  # --------------------------------------------------------
  # First validation
  # --------------------------------------------------------

  if (!is.list(partitions) || length(partitions) == 0) {
    stop("partitions must be a non-empty list.")
  }

  if (!is.list(consensus_results)) {
    stop("consensus_results must be a list returned by consensus_clustering().")
  }

  # --------------------------------------------------------
  # Helpers
  # --------------------------------------------------------

  .pac_from_coassignment <- function(coassignment, lower, upper) {
    vals <- coassignment[upper.tri(coassignment, diag = FALSE)]
    mean(vals > lower & vals < upper)
  }

  .mean_silhouette_from_diss <- function(diss, labels) {
    if (length(unique(labels)) < 2) return(NA_real_)

    sil <- tryCatch(
      cluster::silhouette(as.integer(labels), stats::as.dist(diss)),
      error = function(e) NULL
    )

    if (is.null(sil)) return(NA_real_)

    mean(sil[, "sil_width"])
  }

  .ari_mean_between_partitions <- function(parts_list) {
    m <- length(parts_list)
    if (m < 2) return(NA_real_)
    ari_vals <- c()
    for (i in seq_len(m - 1)) {
      for (j in seq.int(i + 1, m)) {
        ari_vals <- c(ari_vals,
                      mclust::adjustedRandIndex(parts_list[[i]], parts_list[[j]]))
      }
    }
    mean(ari_vals)
  }

  .ari_consensus_mean <- function(consensus, parts_list) {
    vals <- vapply(parts_list, function(z) {
      if (length(unique(z)) < 2 || length(unique(consensus)) < 2) {
        return(NA_real_)
      }
      mclust::adjustedRandIndex(consensus, z)
    }, numeric(1))
    mean(vals, na.rm = TRUE)
  }

  # CH + Dunn from dissimilarity
  .ch_dunn_from_diss <- function(diss, labels) {
    cs <- tryCatch(
      fpc::cluster.stats(d = stats::as.dist(diss),
                         clustering = as.integer(labels)),
      error = function(e) NULL
    )
    if (is.null(cs)) {
      return(list(ch = NA_real_, dunn = NA_real_))
    }
    list(
      ch = as.numeric(cs$ch),
      dunn = as.numeric(cs$dunn)
    )
  }

  # Davies–Bouldin (medoid-based)
  .db_from_diss_medoids <- function(diss, labels) {

    labs <- as.integer(labels)
    cl_ids <- sort(unique(labs))
    k <- length(cl_ids)
    if (k < 2) return(NA_real_)

    medoid <- integer(k)
    S <- numeric(k)

    for (ii in seq_along(cl_ids)) {
      cl <- cl_ids[ii]
      idx <- which(labs == cl)

      if (length(idx) <= 1) {
        medoid[ii] <- idx[1]
        S[ii] <- 0
        next
      }

      Dsub <- diss[idx, idx, drop = FALSE]
      mpos <- which.min(rowSums(Dsub))
      medoid[ii] <- idx[mpos]
      S[ii] <- mean(diss[idx, medoid[ii]])
    }

    d_med <- diss[medoid, medoid, drop = FALSE]

    Rmax <- rep(NA_real_, k)
    for (i in seq_len(k)) {
      ratios <- rep(NA_real_, k)
      for (j in seq_len(k)) {
        if (j == i) next
        denom <- d_med[i, j]
        if (!is.na(denom) && denom > 0) {
          ratios[j] <- (S[i] + S[j]) / denom
        }
      }
      Rmax[i] <- suppressWarnings(max(ratios, na.rm = TRUE))
      if (!is.finite(Rmax[i])) Rmax[i] <- NA_real_
    }

    mean(Rmax, na.rm = TRUE)
  }

  # --------------------------------------------------------
  # Core computation per k
  # --------------------------------------------------------

  .one_k <- function(parts_k, cres_k) {

    if (is.null(cres_k$consensus) || is.null(cres_k$coassignment)) {
      stop("Each consensus result must contain 'consensus' and 'coassignment'.")
    }

    cons  <- cres_k$consensus
    coass <- cres_k$coassignment
    diss  <- 1 - coass

    # Robust protection
    if (length(unique(cons)) < 2) {
      return(data.frame(
        k = cres_k$k,
        pac = NA_real_,
        silhouette_mean = NA_real_,
        ari_mean_between_imputations = NA_real_,
        ari_consensus_mean = NA_real_,
        calinski_harabasz_mean = NA_real_,
        davies_bouldin_mean = NA_real_,
        dunn_index = NA_real_,
        stringsAsFactors = FALSE
      ))
    }

    chd <- .ch_dunn_from_diss(diss, cons)

    data.frame(
      k = cres_k$k,
      pac = .pac_from_coassignment(coass, pac_lower, pac_upper),
      silhouette_mean = .mean_silhouette_from_diss(diss, cons),
      ari_mean_between_imputations =
        .ari_mean_between_partitions(parts_k),
      ari_consensus_mean =
        .ari_consensus_mean(cons, parts_k),
      calinski_harabasz_mean = chd$ch,
      davies_bouldin_mean =
        .db_from_diss_medoids(diss, cons),
      dunn_index = chd$dunn,
      stringsAsFactors = FALSE
    )
  }

  # --------------------------------------------------------
  # Single-k vs Multi-k
  # --------------------------------------------------------

  is_multi_k <- all(vapply(partitions, is.list, logical(1)))

  if (!is_multi_k) {
    if (!all(vapply(partitions, is.atomic, logical(1)))) {
      stop("For single-k input, partitions must be a list of atomic vectors.")
    }
    return(.one_k(partitions, consensus_results))
  }

  k_names <- intersect(names(partitions), names(consensus_results))
  k_values <- suppressWarnings(as.integer(sub("^k", "", k_names)))
  k_names <- k_names[order(k_values)]

  rows <- lapply(k_names,
                 function(kk)
                   .one_k(partitions[[kk]],
                          consensus_results[[kk]]))

  do.call(rbind, rows)
}

#' cclustr: Consensus Clustering Methods for Multiple Imputed Data
#'
#' @description
#' The \pkg{cclustr} package provides a comprehensive framework for performing
#' clustering analysis on datasets treated with multiple imputation. It
#' addresses the uncertainty inherent in missing data by finding a stable
#' consensus partition across multiple imputed realizations.
#'
#' @details
#' Standard statistical workflows for imputed data often focus on supervised
#' learning or univariate statistics. \pkg{cclustr} extends this to
#' unsupervised learning by implementing a co-assignment matrix approach.
#'
#' @section Main Workflow:
#' To achieve a robust consensus solution, follow these steps:
#' \enumerate{
#'   \item \strong{Standardization:} Use \code{\link{as_mild_list}} to format
#'   imputed data (e.g., from \pkg{mice}) into a consistent structure.
#'   \item \strong{Clustering:} Apply \code{\link{cluster_imputations}} to run
#'   algorithms (K-means, PAM, K-prototypes, etc.) across all imputations.
#'   \item \strong{Consensus:} Generate the final partition using
#'   \code{\link{consensus_clustering}} based on co-assignment probabilities.
#'   \item \strong{Selection:} Evaluate model quality and optimal cluster
#'   counts with \code{\link{choose_best_clustering}}.
#' }
#'
#' @section Supported Data Types:
#' \itemize{
#'   \item \strong{Numeric:} Hierarchical clustering, K-means, Fuzzy C-means, and Mclust.
#'   \item \strong{Categorical:} K-modes.
#'   \item \strong{Mixed:} PAM with Gower distance or K-prototypes.
#' }
#'
#' @author
#' \strong{Maintainer}: Andres Montenegro Lemus \email{andresfemole@@gmail.com}
#'
#' Authors:
#' \itemize{
#'   \item Anhuar Duran Mendoza \email{asteyduran15@@gmail.com}
#'   \item Mario Pacheco Lopez \email{mariopachecolopez@@gmail.com}
#' }
#'
#' @references
#' Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
#' Consensus clustering: A resampling-based method for class discovery
#' and visualization of gene expression microarray data.
#' \emph{Machine Learning}, \strong{52}(1-2), 91-118.
#' \doi{10.1023/A:1023949509487}
#'
#' Senbabaoglu, Y., Michailidis, G., & Li, J. Z. (2014).
#' Critical limitations of consensus clustering in class discovery.
#' \emph{Scientific Reports}, \strong{4}, 6207.
#' \doi{10.1038/srep06207}
#'
#' Pihur, V., Datta, S., & Datta, S. (2007).
#' Weighted rank aggregation of cluster validation measures.
#' \emph{Bioinformatics}, \strong{23}(13), 1607-1615.
#' \doi{10.1093/bioinformatics/btm158}
#'
#' Rubin, D. B. (1987).
#' \emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.
#'
#' van Buuren, S., & Groothuis-Oudshoorn, K. (2011).
#' mice: Multivariate Imputation by Chained Equations in R.
#' \emph{Journal of Statistical Software}, \strong{45}(3), 1-67.
#' \doi{10.18637/jss.v045.i03}
#'
#' Hubert, L., & Arabie, P. (1985).
#' Comparing partitions.
#' \emph{Journal of Classification}, \strong{2}, 193-218.
#' \doi{10.1007/BF01908075}
#'
#' Rousseeuw, P. J. (1987).
#' Silhouettes: A graphical aid to the interpretation and validation
#' of cluster analysis.
#' \emph{Journal of Computational and Applied Mathematics}, \strong{20}, 53-65.
#' \doi{10.1016/0377-0427(87)90125-7}
#'
#' Calinski, T., & Harabasz, J. (1974).
#' A dendrite method for cluster analysis.
#' \emph{Communications in Statistics}, \strong{3}, 1-27.
#' \doi{10.1080/03610927408827101}
#'
#' Davies, D. L., & Bouldin, D. W. (1979).
#' A cluster separation measure.
#' \emph{IEEE Transactions on Pattern Analysis and Machine Intelligence},
#' \strong{PAMI-1}, 224-227.
#' \doi{10.1109/TPAMI.1979.4766909}
#'
#' Dunn, J. C. (1974).
#' Well-separated clusters and optimal fuzzy partitions.
#' \emph{Journal of Cybernetics}, \strong{4}, 95-104.
#' \doi{10.1080/01969727408546059}
#'
#' @seealso
#' Useful links:
#' \itemize{
#'   \item \url{https://github.com/andrews06ml/cclustr}
#'   \item Report bugs at \url{https://github.com/andrews06ml/cclustr/issues}
#' }
#'
#' @docType package
#' @keywords internal
"_PACKAGE"

---
title: "Consensus Clustering with Multiple Imputed Data using cclustr"
author: "Anhuar Duran Mendoza, Andres Montenegro Lemus, Mario Pacheco Lopez"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Consensus Clustering with Multiple Imputed Data using cclustr}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  fig.width = 7,
  fig.height = 5,
  warning   = FALSE,
  message   = FALSE
)
```

## Introduction

Missing data is a pervasive challenge in applied statistics. When performing
cluster analysis, the standard practice of imputing a single dataset and then
clustering ignores the uncertainty introduced by the imputation process,
potentially leading to unstable or misleading cluster assignments.

**cclustr** addresses this problem by applying clustering algorithms across
all imputed datasets and synthesizing the results into a single, robust
consensus partition. The core idea is the **co-assignment matrix**: a
symmetric matrix $C$ where each entry $C_{ij}$ represents the proportion of
imputed datasets in which observations $i$ and $j$ were assigned to the same
cluster. High values indicate stable, robust groupings across imputations;
low values indicate ambiguous assignments.

The main workflow of `cclustr` consists of four steps:

1. **Standardize** imputed data with `as_mild_list()`
2. **Cluster** each imputed dataset with `cluster_imputations()`
3. **Synthesize** results into a consensus partition with `consensus_clustering()`
4. **Validate and select** the optimal number of clusters with
   `validate_clustering()` and `choose_best_clustering()`

All four steps can also be executed in a single call using the convenience
wrapper `run_mi_clustering()`.

---

## Dataset: SynthDiabetes2

We use the **SynthDiabetes2 Database** (`SynthDiabetes2`) from
the `mlbench` package (version 2.1-11 or later), a synthetic diabetes case study with 768 observations.
Every value in the data is synthetic and no row corresponds to a real person; the dataset was generated using sequential robust synthesis with a random-forest backend, with the missingness pattern treated as a structural variable so that it mimics the pattern of missing values found in the original clinical study it is modeled on. The dataset presents **709 missing values** distributed across several variables, strongly motivating the use of multiple imputation before clustering.

### Variables

| Variable   | Type    | Description                                      |
|------------|---------|--------------------------------------------------|
| `pregnant` | Numeric | Number of times pregnant                         |
| `glucose`  | Numeric | Plasma glucose concentration (2h oral test)      |
| `pressure` | Numeric | Diastolic blood pressure (mm Hg)                 |
| `triceps`  | Numeric | Triceps skin fold thickness (mm)                 |
| `insulin`  | Numeric | 2-hour serum insulin (mu U/ml)                   |
| `mass`     | Numeric | Body mass index (kg/m²)                          |
| `pedigree` | Numeric | Diabetes pedigree function                       |
| `age`      | Numeric | Age in years                                     |
| `diabetes` | Factor  | Diabetes test result (positive/negative)         |

> **Note:** The `diabetes` variable is excluded from the clustering analysis.
> This avoids circular reasoning and respects the unsupervised nature of the
> analysis.

### Loading the data

```{r load-data}
library(cclustr)
library(mice)

if (!requireNamespace("mlbench", quietly = TRUE) ||
    packageVersion("mlbench") < "2.1-11") {
  stop("Package 'mlbench' (>= 2.1-11) is required for this vignette. ",
       "Install it with: install.packages('mlbench')")
}

data("SynthDiabetes2", package = "mlbench")

# Remove diabetes from the clustering dataset
df <- SynthDiabetes2[, -9]

cat("Dimensions:", nrow(df), "x", ncol(df), "\n")
cat("\nMissing values per variable:\n")
print(colSums(is.na(df)))
cat("\nTotal missing values:", sum(is.na(df)),
    "(", round(mean(is.na(df)) * 100, 1), "%)\n")
```

The high proportion of missing values — particularly in `insulin` (52.7%) and
`triceps` (32.6%) — makes this dataset a compelling real-world case study for
multiple-imputation clustering.

---

## Step 1: Multiple Imputation with mice

We generate 20 imputed datasets using `mice` with predictive mean matching
(`pmm`). The `as_mild_list()` function from `cclustr` then converts the
`mids` object into a standardized named list of completed data frames.

```{r imputation}

imp <- mice(df, m = 20, maxit = 20, method = "pmm", printFlag = FALSE, seed = 202604)

# Convert to cclustr standard format
mi_synth <- as_mild_list(imp)

cat("Number of imputed datasets :", length(mi_synth), "\n")
cat("Dimensions of each dataset :", nrow(mi_synth[[1]]),
    "x", ncol(mi_synth[[1]]), "\n")
cat("Any missing values remaining:", any(sapply(mi_synth, anyNA)), "\n")
```

`as_mild_list()` performs strict consistency checks across all imputed
datasets, ensuring identical dimensions, column names, and the absence of
`NA`, `NaN`, `Inf`, or constant-valued columns.

---

## Step 2: Clustering Each Imputed Dataset

Since all clustering variables are numeric, we use **k-means
clustering** and `scale_data = "global"`. The
global scaling option computes the pooled mean and standard deviation across
all imputations without stacking them in memory, ensuring that distances are
on the same scale across all imputed datasets — a key requirement for a
meaningful co-assignment matrix.

We evaluate $k = 2, 3, 4$ clusters simultaneously.

```{r clustering}
set.seed(202604)
partitions <- cluster_imputations(
  imp_list   = mi_synth,
  method     = "kmeans",
  k          = 2:4,
  scale_data = "global"
)

cat("Available k values:", names(partitions), "\n")
cat("Number of partitions per k:", length(partitions$k2), "\n")
```

The result is a named list of lists: one element per $k$ value (`k2`, `k3`,
`k4`), each containing one integer vector of cluster assignments per imputed
dataset.

---

## Step 3: Consensus Clustering

`consensus_clustering()` builds the **co-assignment matrix** $C$ and derives
a final partition from it. This process involves **two conceptually distinct
steps**:

1. **How the co-assignment matrix is constructed** (`consensus_method`)
2. **How the final clusters are extracted** (`cluster_method`)

### The `consensus_method` parameter

The `consensus_method` argument controls **how each partition contributes
to the co-assignment matrix** $C$.

| Method            | Description |
|------------------|------------|
| `"classic"`      | Unweighted co-assignment. All partitions contribute equally to $C$. This is the standard approach in consensus clustering literature. |
| `"weighted_ari"` | Partitions are weighted according to their **centrality** within the ensemble, measured by their mean Adjusted Rand Index (ARI) against all other partitions. More consistent partitions have greater influence on $C$. Requires the `mclust` package. |

### The `cluster_method` parameter

The `cluster_method` argument controls the **agglomeration method** used
in this second-stage hierarchical clustering step. It is important to
distinguish it from the `method` argument in `cluster_imputations()`:

- `method` (in `cluster_imputations`): the algorithm used to cluster each
  imputed dataset (e.g., `"kmeans"`, `"ward.D2"`, `"pam"`).
- `cluster_method` (in `consensus_clustering`): the linkage criterion used
  to hierarchically cluster the **co-assignment dissimilarity** $1 - C$ in
  order to derive the final consensus partition.

The available options for `cluster_method` are the standard agglomeration
methods from `stats::hclust()`:

| Method        | Description                                                         |
|---------------|---------------------------------------------------------------------|
| `"ward.D2"`   | (default) Minimizes total within-cluster variance. Tends to produce compact, roughly equal-sized clusters. Recommended for most cases. |
| `"ward.D"`    | Earlier Ward formulation; similar behavior to `"ward.D2"` but uses unsquared distances internally. |
| `"complete"`  | Maximum linkage. Tends to produce compact clusters of similar diameter. |
| `"average"`   | UPGMA. Compromise between single and complete; less sensitive to outliers than complete linkage. |
| `"single"`    | Minimum linkage. Sensitive to chaining effects; not recommended for consensus matrices. |
| `"centroid"`  | Uses the centroid of each cluster. Can produce inversions in the dendrogram. |
| `"median"`    | Weighted centroid. Also susceptible to inversions. |
| `"mcquitty"`  | WPGMA. Similar to average but uses equal weighting. |

For consensus matrices, `"ward.D2"` and `"complete"` are generally the
most appropriate choices, as they favor compact, well-separated groups
that match the block-diagonal structure of a stable co-assignment matrix.

```{r consensus}
consensus_results <- consensus_clustering(
  partitions     = partitions,
  cluster_method = "ward.D2",
  consensus_method = "classic"
)

cat("Consensus solutions available:", names(consensus_results), "\n")
```

---

## Step 4: Validation and Optimal k Selection

`validate_clustering()` computes a comprehensive set of internal and stability
metrics for each candidate $k$. All metrics are computed on the consensus
dissimilarity $1 - C$, making them robust to any data type.

```{r validation}
val_table <- validate_clustering(
  partitions        = partitions,
  consensus_results = consensus_results
)

print(val_table)
```

### Metric Reference

| Metric                          | Better when | Description                                    |
|---------------------------------|-------------|------------------------------------------------|
| `pac`                           | Lower ↓     | Proportion of Ambiguous Clusterings            |
| `silhouette_mean`               | Higher ↑    | Mean silhouette width                          |
| `ari_mean_between_imputations`  | Higher ↑    | Stability across imputed datasets              |
| `ari_consensus_mean`            | Higher ↑    | Agreement between consensus and each partition |
| `calinski_harabasz_mean`        | Higher ↑    | Cluster compactness and separation             |
| `davies_bouldin_mean`           | Lower ↓     | Average cluster similarity                     |
| `dunn_index`                    | Higher ↑    | Min inter-cluster / max intra-cluster distance |

`choose_best_clustering()` aggregates all metrics using **weighted rank
aggregation**: each metric is independently ranked and combined into a single
score. The $k$ with the lowest weighted rank score is selected as optimal.

```{r selection}
best <- choose_best_clustering(
  validation_table  = val_table,
  consensus_results = consensus_results,
  prefer_stability  = TRUE,
  tie_breaker       = "silhouette"
)

cat("Optimal k selected:", best$best_k, "\n")
cat("\nConsensus cluster sizes:\n")
print(table(best$best_consensus))
cat("\nWeighted rank scores:\n")
print(best$scores_table[, c("k", "pac", "silhouette_mean",
                             "ari_mean_between_imputations", "score")])
```

### Plotting Validation Metrics

```{r validation-plot, fig.cap = "Validation metrics across k = 2, 3, 4. The optimal value per metric is highlighted in red."}
plot_validation_metrics(best)
```

---

## Visualizing the Consensus Solution

All visualizations are based on the **optimal k** selected in the previous
step, accessed through `best$best_consensus_result`.

### Co-assignment Matrix

A well-defined consensus solution produces a **block-diagonal pattern**:
bright blocks along the diagonal indicate observations consistently assigned
to the same cluster across all imputed datasets, while dark off-diagonal
regions indicate observations rarely grouped together.

```{r heatmap, fig.cap = "Co-assignment matrix for the optimal k. A clear block-diagonal structure indicates a stable consensus."}
plot_consensus_matrix(
  consensus_result = best,
  reorder          = TRUE,
  viridis_option   = "D",
  show_labels = FALSE
)
```

### Consensus Dendrogram

The dendrogram is built from the hierarchical clustering of the consensus
dissimilarity $1 - C$. Colored rectangles highlight the groups obtained by
cutting the tree at the optimal $k$.

```{r dendrogram, fig.cap = "Consensus dendrogram for the optimal k, derived from the dissimilarity matrix 1 - C."}
plot_consensus_dendrogram(
  consensus_result = best,
  rect             = TRUE,
  labels           = FALSE
)
```

---

## Interpreting the Final Clusters

We attach the consensus labels to the first imputed dataset for a descriptive summary of each cluster's clinical profile.

```{r cluster-summary}
synth_clustered          <- mi_synth[[1]]
synth_clustered$cluster  <- factor(best$best_consensus)

# Mean values per cluster
num_vars <- c("pregnant", "glucose", "pressure",
              "triceps", "insulin", "mass", "pedigree", "age")

cat("Mean clinical values by cluster:\n")
print(
  aggregate(
    synth_clustered[, num_vars],
    by  = list(Cluster = synth_clustered$cluster),
    FUN = function(x) round(mean(x, na.rm = TRUE), 2)
  )
)
```

### Key Variables by Cluster

```{r boxplots, fig.cap = "Distribution of glucose and BMI by consensus cluster."}
par(mfrow = c(1, 3), mar = c(4, 4, 3, 1))

boxplot(
  glucose ~ cluster,
  data = synth_clustered,
  main = "Plasma Glucose by Cluster",
  xlab = "Cluster",
  ylab = "Glucose (mg/dl)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)

boxplot(
  mass ~ cluster,
  data = synth_clustered,
  main = "Body Mass Index by Cluster",
  xlab = "Cluster",
  ylab = "BMI (kg/m²)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)

boxplot(
  insulin ~ cluster,
  data = synth_clustered,
  main = "Insulin",
  xlab = "Cluster",
  ylab = "Insulin (mu U/ml)",
  col  = c("#4E79A7", "#F28E2B", "#E15759")
)
par(mfrow = c(1, 1))
```

---

## Complete Pipeline with run_mi_clustering()

All previous steps can be executed in a single call using `run_mi_clustering()`.
This is the recommended approach for most users.

```{r full-pipeline}
set.seed(202604)

result <- run_mi_clustering(
  data             = imp,
  method           = "kmeans",
  k                = 2:4,
  scale_data       = "global",
  consensus_method = "classic",
  prefer_stability = TRUE,
  tie_breaker      = "silhouette"
)

cat("Best k selected:", result$best_k, "\n")
cat("\nCluster sizes:\n")
print(table(result$best_consensus))
```

---

## Advanced: ARI-Weighted Consensus

By default all imputed partitions contribute equally to the co-assignment
matrix (`consensus_method = "classic"`). When partitions vary in quality,
`consensus_method = "weighted_ari"` weights each partition by its mean
Adjusted Rand Index (ARI) against all others — so more consistent partitions
contribute more to the final consensus.

```{r weighted-ari, eval = requireNamespace("mclust", quietly = TRUE)}
set.seed(202604)

result_w <- run_mi_clustering(
  data             = imp,
  method           = "kmeans",
  k                = 2:4,
  scale_data       = "global",
  consensus_method = "weighted_ari",
  prefer_stability = TRUE,
  tie_breaker      = "silhouette"
)

cat("Best k (ARI-weighted consensus):", result_w$best_k, "\n")
print(table(result_w$best_consensus))
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Dendrogram.R
\name{plot_consensus_dendrogram}
\alias{plot_consensus_dendrogram}
\title{Plot a consensus clustering dendrogram}
\usage{
plot_consensus_dendrogram(
  consensus_result,
  k = NULL,
  rect = TRUE,
  hang = -1,
  labels = NULL
)
}
\arguments{
\item{consensus_result}{A named list as returned by
\code{\link{consensus_clustering}} for a single \code{k}. Must contain
at least \code{$hclust} (an \code{hclust} object) and \code{$k} (an
integer specifying the number of clusters).}

\item{k}{Integer. The number of clusters to display. Only required when
\code{consensus_result} contains results for multiple \code{k} values.
Ignored if a single-k result is passed directly.}

\item{rect}{Logical. If \code{TRUE} (default), colored rectangles are drawn
around the clusters via \code{\link[stats]{rect.hclust}}.}

\item{hang}{Numeric. The fraction of the plot height by which labels hang
below the rest of the plot. A negative value (default \code{-1}) causes
all labels to hang down from zero.}

\item{labels}{Logical or character vector. Controls the labels displayed
at the leaves of the dendrogram. If \code{NULL} (default), observation
labels are shown. If \code{FALSE}, no labels are displayed — recommended
for large datasets where labels become unreadable. A character vector
of the same length as the number of observations can also be supplied
to use custom labels.}
}
\value{
Invisibly returns \code{NULL}. The function is called for its
side effect of producing a plot.
}
\description{
Displays the hierarchical clustering dendrogram derived from the consensus
dissimilarity matrix \eqn{1 - C}, where \eqn{C} is the co-assignment matrix
produced by \code{\link{consensus_clustering}}. Optionally draws colored
rectangles around the clusters defined by the requested \code{k}.
}
\details{
The dendrogram is built from the \code{hclust} object stored in
\code{consensus_result$hclust}, which was computed by
\code{\link{consensus_clustering}} on the consensus dissimilarity matrix
\eqn{1 - C}. The colored rectangles highlight the \code{k} groups
obtained by cutting the dendrogram at the level that yields exactly
\code{k} clusters.

This function is intended to be called on a single \code{k}. To compare
dendrograms across multiple values of \code{k}, call the function
iteratively and manage the graphics layout externally via
\code{\link[graphics]{par}}.
}
\examples{
# ------------------------------------------------------------
# Example 1: consensus dendrogram with simulated results
# ------------------------------------------------------------
set.seed(123)
hc <- hclust(dist(matrix(rnorm(60), nrow = 20)), method = "ward.D2")
cons <- list(hclust = hc, k = 3)

plot_consensus_dendrogram(cons)
plot_consensus_dendrogram(cons, rect = FALSE)

\donttest{
# ------------------------------------------------------------
# Example 2: Dendrogram with mice
# ------------------------------------------------------------

if (requireNamespace("mice", quietly = TRUE)) {

set.seed(123)
imp    <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
mild   <- as_mild_list(imp)
parts  <- cluster_imputations(mild, method = "ward.D2", k = 3)
cons   <- consensus_clustering(parts, k = 3)

# Single dendrogram
plot_consensus_dendrogram(cons)

# Without rectangles
plot_consensus_dendrogram(cons, rect = FALSE)

# Compare multiple k values
parts_multi <- cluster_imputations(mild, method = "ward.D2", k = 2:4)
cons_multi  <- consensus_clustering(parts_multi)

par(mfrow = c(1, 3), mar = c(3, 2, 2, 1))
for (k_i in 2:4) {
  plot_consensus_dendrogram(cons_multi[[paste0("k", k_i)]])
}
par(mfrow = c(1, 1))
}
}

}
\seealso{
\code{\link{consensus_clustering}}, \code{\link{plot_consensus_matrix}},
\code{\link{plot_validation_metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Imputation.R
\name{as_mild_list}
\alias{as_mild_list}
\title{Standardize multiple imputation outputs into a unified list}
\usage{
as_mild_list(x)
}
\arguments{
\item{x}{An imputation object. Accepted classes and formats:
\itemize{
\item \code{mids}: output of \code{mice::mice()}.
\item \code{data.frame}: long-format output of
\code{mice::complete(x, action = "long")}. Must contain a
\code{.imp} column.
\item \code{amelia}: output of \code{Amelia::amelia()}.
\item \code{imputationList}: output of \code{mitools::imputationList()}.
\item \code{list}: a plain list where every element is a
\code{data.frame} representing one completed dataset.
}}
}
\value{
A named list of \code{data.frame} objects, one per imputed
dataset. All datasets are guaranteed to have identical dimensions and
column names, and to contain no \code{NA}, \code{NaN}, \code{Inf}, or
all equal values in a column.
}
\description{
Converts multiple imputation outputs from different packages into a
standardized list of completed datasets, where each element corresponds
to one imputed dataset. Supported formats include \code{mids} objects
from \pkg{mice}, long-format data frames, \code{amelia} objects,
\code{imputationList} objects from \pkg{mitools}, and plain lists of
data frames. After conversion, consistency and data quality checks are
applied across all imputed datasets.

This function is designed as a preprocessing step for clustering
multiple imputed datasets, ensuring strict comparability across imputations.
}
\details{
The function performs the following consistency checks before returning:
\itemize{
\item Identical number of rows across all imputed datasets.
\item Identical number of columns across all imputed datasets.
\item Identical column names across all imputed datasets.
\item Absence of \code{NA}, \code{NaN}, and \code{Inf} values.
\item Absence of numeric columns with all values equal,
as such variables provide no information for distance-based
clustering and may lead to degenerate solutions.
}
All datasets are coerced to base \code{data.frame} to avoid
compatibility issues with \code{tibble} or other subclasses.
}
\examples{
# ------------------------------------------------------------
# Example 1: Simple list of completed datasets
# ------------------------------------------------------------
set.seed(123)

imp_list <- replicate(3, {
  data.frame(
    x = rnorm(10),
    y = rnorm(10)
  )
}, simplify = FALSE)

mild <- as_mild_list(imp_list)

length(mild)      # number of imputations
str(mild[[1]])    # structure of one dataset

\donttest{
# ------------------------------------------------------------
# Example 2: List of completed datasets with mice
# ------------------------------------------------------------

if (requireNamespace("mice", quietly = TRUE)) {

  set.seed(123)

  # Example 1: Using a mids object
  imp <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
  mild <- as_mild_list(imp)

  length(mild)        # number of imputations
  str(mild[[1]])      # structure of one completed dataset

  # Example 2: Using long-format data
  long_df <- mice::complete(imp, action = "long", include = FALSE)
  mild2 <- as_mild_list(long_df)

  length(mild2)
}
}

}
\references{
Rubin, D. B. (1976).
Inference and missing data.
\emph{Biometrika}, \strong{63}(3), 581-592.
\doi{10.1093/biomet/63.3.581}

Rubin, D. B. (1987).
\emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.

van Buuren, S., & Groothuis-Oudshoorn, K. (2011).
mice: Multivariate Imputation by Chained Equations in R.
\emph{Journal of Statistical Software}, \strong{45}(3), 1-67.
\doi{10.18637/jss.v045.i03}

Honaker, J., King, G., & Blackwell, M. (2011).
Amelia II: A Program for Missing Data.
\emph{Journal of Statistical Software}, \strong{45}(7), 1-47.
\doi{10.18637/jss.v045.i07}
}
\seealso{
\code{\link{cluster_imputations}}, \code{\link{consensus_clustering}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Heatmap.R
\name{plot_consensus_matrix}
\alias{plot_consensus_matrix}
\title{Plot the consensus co-assignment matrix}
\usage{
plot_consensus_matrix(
  consensus_result,
  reorder = TRUE,
  show_labels = FALSE,
  viridis_option = "D"
)
}
\arguments{
\item{consensus_result}{A named list as returned by
\code{\link{consensus_clustering}} for a single \code{k}. Must contain
at least \code{$coassignment} (a numeric \eqn{n \times n} matrix) and,
when \code{reorder = TRUE}, \code{$consensus} (an integer vector of
cluster labels).}

\item{reorder}{Logical. If \code{TRUE} (default), rows and columns are
reordered by the consensus cluster assignment stored in
\code{consensus_result$consensus}, making the block structure of the
matrix visually apparent.}

\item{show_labels}{Logical. If \code{TRUE}, observation indices are
displayed on both axes, ordered according to the consensus cluster
assignment. Recommended only for small datasets (n < 50) as labels
become unreadable for large \eqn{n}. If \code{FALSE} (default),
tick marks are shown without labels to preserve the axis frame
without visual clutter.}

\item{viridis_option}{A single character specifying the color palette
passed to \code{\link[viridisLite]{viridis}}. Accepted values are
\code{"A"} (magma), \code{"B"} (inferno), \code{"C"} (plasma), and
\code{"D"} (viridis, default). Low co-assignment values are mapped to
dark colors and high values to bright colors.}
}
\value{
Invisibly returns \code{NULL}. The function is called for its
side effect of producing a plot.
}
\description{
Displays a heatmap of the co-assignment matrix \eqn{C} produced by
\code{\link{consensus_clustering}}, where each cell \eqn{C_{ij}} represents
the weighted proportion of partitions in which observations \eqn{i} and
\eqn{j} were assigned to the same cluster. Values close to 1 indicate
that two observations are consistently grouped together across all
imputed datasets; values close to 0 indicate the opposite. Rows and
columns can be reordered by consensus cluster assignment to make the
block structure visually apparent.
}
\details{
The matrix is rendered using \code{\link[graphics]{image}} on the
consensus dissimilarity \eqn{1 - C} reordered by cluster membership.
Original graphical parameters are restored on exit via
\code{\link[graphics]{par}}.

A well-defined consensus solution produces a block-diagonal pattern in
the heatmap, where bright blocks along the diagonal correspond to
observations that are consistently assigned to the same cluster, and
dark off-diagonal regions indicate observations that are rarely grouped
together.
}
\examples{
# ------------------------------------------------------------
# Example 1: Consensus matrix with simulated consensus result
# ------------------------------------------------------------
set.seed(123)
n <- 20
true_labels <- rep(1:2, each = n / 2)
M <- matrix(0.1, n, n)
for (i in seq_len(n))
  for (j in seq_len(n))
    if (true_labels[i] == true_labels[j]) M[i, j] <- 0.9

cons <- list(coassignment = M,
             consensus    = true_labels,
             k            = 2)

plot_consensus_matrix(cons, viridis_option = "A")

\donttest{
# ------------------------------------------------------------
# Example 2: Consensus matrix with mice
# ------------------------------------------------------------

if (requireNamespace("mice", quietly = TRUE)) {

imp   <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
mild  <- as_mild_list(imp)
parts <- cluster_imputations(mild, method = "ward.D2", k = 3)
cons  <- consensus_clustering(parts, k = 3)

# Default plot with viridis palette
plot_consensus_matrix(cons)

# Without reordering
plot_consensus_matrix(cons, reorder = FALSE)

# Compare color palettes
par(mfrow = c(2, 2), mar = c(4, 4, 2, 2))
for (opt in c("A", "B", "C", "D")) {
  plot_consensus_matrix(cons, viridis_option = opt)
}
par(mfrow = c(1, 1))
}
}

}
\seealso{
\code{\link{consensus_clustering}}, \code{\link{plot_consensus_dendrogram}},
\code{\link{plot_validation_metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Selection.R
\name{choose_best_clustering}
\alias{choose_best_clustering}
\title{Select the optimal number of clusters from a validation table}
\usage{
choose_best_clustering(
  validation_table,
  consensus_results,
  weights = NULL,
  prefer_stability = NULL,
  tie_breaker = c("silhouette", "pac", "dunn", "ch", "db", "ari_between",
    "ari_consensus")
)
}
\arguments{
\item{validation_table}{A \code{data.frame} as returned by
\code{\link{validate_clustering}}, containing one row per evaluated
\code{k} and columns \code{pac}, \code{silhouette_mean},
\code{ari_mean_between_imputations}, \code{ari_consensus_mean},
\code{calinski_harabasz_mean}, \code{davies_bouldin_mean}, and
\code{dunn_index}.}

\item{consensus_results}{A list as returned by
\code{\link{consensus_clustering}}, either for a single \code{k} or
for multiple \code{k} values. Used to retrieve the consensus labels
and co-assignment matrix for the selected \code{k}.}

\item{weights}{A named numeric vector specifying the relative importance
of each metric in the rank aggregation. Expected names are \code{pac},
\code{silhouette}, \code{ari_between}, \code{ari_consensus}, \code{ch},
\code{db}, and \code{dunn}. If \code{NULL} (default), weights are set
automatically based on \code{prefer_stability}. When provided
explicitly, \code{prefer_stability} is ignored.}

\item{prefer_stability}{Controls the default weighting strategy when
\code{weights} is \code{NULL}. Accepted values are:
\itemize{
\item \code{NULL} (default): all metrics receive equal weight,
providing a balanced selection that does not favor any
particular aspect of clustering quality.
\item \code{TRUE}: stability metrics (\code{pac},
\code{ari_between}, \code{ari_consensus}) receive higher
weights.
\item \code{FALSE}: internal compactness metrics
(\code{silhouette}, \code{ch}, \code{db}, \code{dunn})
receive higher weights.
}}

\item{tie_breaker}{A character string specifying the secondary metric
used to break ties in the weighted rank score. One of
\code{"silhouette"} (default), \code{"pac"}, \code{"dunn"},
\code{"ch"}, \code{"db"}, \code{"ari_between"}, or
\code{"ari_consensus"}.}
}
\value{
A named list with the following elements:
\itemize{
\item \code{best_k}: integer, the selected optimal number of clusters.
\item \code{best_consensus}: integer vector of consensus cluster
labels for the selected \code{k}, one per observation.
\item \code{best_coassignment}: numeric matrix of size
\eqn{n \times n} with co-assignment probabilities for the
selected \code{k}.
\item \code{best_consensus_result}: the full consensus result object
for the selected \code{k}, as returned by
\code{\link{consensus_clustering}}.
\item \code{scores_table}: the input \code{validation_table} with an
additional \code{score} column containing the weighted rank
score for each \code{k}, sorted by \code{k}.
\item \code{weights}: named numeric vector of weights effectively
used in the aggregation.
\item \code{tie_breaker}: character string indicating the tiebreaker
metric used.
}
}
\description{
Identifies the best value of \code{k} from the validation metrics produced
by \code{\link{validate_clustering}}, using a weighted rank aggregation
strategy. Each metric is ranked independently and then combined into a
single score using user-defined weights. The \code{k} with the lowest
weighted rank score is selected as optimal. In case of ties, a secondary
metric is used as a tiebreaker.
}
\details{
The rank aggregation proceeds as follows:
\enumerate{
\item Each metric is ranked from best to worst (rank 1 = best),
accounting for the direction of optimality: lower is better for
\code{pac} and \code{davies_bouldin_mean}; higher is better for
all remaining metrics.
\item For each \code{k}, the weighted average rank is computed using
the provided or default weights. Rows with \code{NA} in some
metrics are handled by renormalizing the weights over available
metrics only.
\item The \code{k} with the lowest weighted rank score is selected.
If multiple \code{k} values share the minimum score, the
tiebreaker metric is used; if still tied, the smallest \code{k}
is preferred.
}

Default weights when \code{prefer_stability = NULL} (default):
\code{pac = 1}, \code{silhouette = 1}, \code{ari_between = 1},
\code{ari_consensus = 1}, \code{ch = 1}, \code{db = 1}, \code{dunn = 1}.

Default weights when \code{prefer_stability = TRUE}:
\code{pac = 2}, \code{silhouette = 1.5}, \code{ari_between = 2},
\code{ari_consensus = 2}, \code{ch = 1}, \code{db = 1}, \code{dunn = 1}.

Default weights when \code{prefer_stability = FALSE}:
\code{pac = 1}, \code{silhouette = 2}, \code{ari_between = 1},
\code{ari_consensus = 1}, \code{ch = 2}, \code{db = 2}, \code{dunn = 2}.

No automatic selection method replaces domain knowledge. The
\code{scores_table} and \code{\link{plot_validation_metrics}} are
provided so that the user can inspect individual metrics and make an
informed decision.
}
\examples{
# ------------------------------------------------------------
# Example 1: Basic validation with simulated partitions
# ------------------------------------------------------------

# 1. Simulated validation table for k = 2, 3, 4
val_table <- data.frame(
  k                            = 2:4,
  pac                          = c(0.15, 0.08, 0.20),
  silhouette_mean              = c(0.42, 0.61, 0.38),
  ari_mean_between_imputations = c(0.80, 0.91, 0.75),
  ari_consensus_mean           = c(0.82, 0.93, 0.77),
  calinski_harabasz_mean       = c(120,  198,  105),
  davies_bouldin_mean          = c(0.85, 0.54, 0.97),
  dunn_index                   = c(0.30, 0.48, 0.27)
)

# 2. Simulated consensus result for each k
set.seed(123)
n <- 30
make_cons <- function(k) {
  list(
    k            = k,
    consensus    = sample(seq_len(k), n, replace = TRUE),
    coassignment = matrix(runif(n * n), n, n)
  )
}
cons_results <- list(k2 = make_cons(2),
                     k3 = make_cons(3),
                     k4 = make_cons(4))

# 3. Default selection (equal weights)
best <- choose_best_clustering(val_table, cons_results)
best$best_k
head(best$scores_table)


# 4. Custom weights
best_custom <- choose_best_clustering(
  val_table, cons_results,
  weights = c(pac = 3, silhouette = 1, ari_between = 3,
              ari_consensus = 3, ch = 1, db = 1, dunn = 1)
)

\donttest{
# ------------------------------------------------------------
# Example 2: Full pipeline with mice
# ------------------------------------------------------------
if (requireNamespace("mice", quietly = TRUE)) {

set.seed(123)
imp   <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
mild  <- as_mild_list(imp)
parts <- cluster_imputations(mild, method = "ward.D2", k = 2:5)
cons  <- consensus_clustering(parts)
val   <- validate_clustering(parts, cons)

# Default selection (equal weights)
best <- choose_best_clustering(val, cons)
best$best_k

# Stability-focused selection
best2 <- choose_best_clustering(val, cons,
                                prefer_stability = TRUE)

# Compactness-focused selection
best3 <- choose_best_clustering(val, cons,
                                prefer_stability = FALSE,
                                tie_breaker      = "dunn")

# Custom weights
best4 <- choose_best_clustering(val, cons,
                                weights = c(pac = 3, silhouette = 1,
                                            ari_between = 3, ari_consensus = 3,
                                            ch = 1, db = 1, dunn = 1))
}
}

}
\references{
Pihur, V., Datta, S., & Datta, S. (2007). Weighted rank aggregation
of cluster validation measures: a Monte Carlo cross-entropy approach.
\emph{Bioinformatics}, \strong{23}(13), 1607-1615.
\doi{10.1093/bioinformatics/btm158}

Hubert, L., & Arabie, P. (1985).
Comparing partitions.
\emph{Journal of Classification}, \strong{2}, 193-218.
\doi{10.1007/BF01908075}
}
\seealso{
\code{\link{validate_clustering}}, \code{\link{consensus_clustering}},
\code{\link{plot_validation_metrics}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Coassignment.R
\name{consensus_clustering}
\alias{consensus_clustering}
\title{Build a consensus partition from multiple imputation clustering results}
\usage{
consensus_clustering(
  partitions,
  k = NULL,
  cluster_method = "ward.D2",
  consensus_method = c("classic", "weighted_ari")
)
}
\arguments{
\item{partitions}{A list of cluster assignment vectors (one per imputed
dataset) for a single \code{k}, or a named list of such lists (one per
\code{k} value, named \code{"k2"}, \code{"k3"}, etc.) as returned by
\code{\link{cluster_imputations}} when \code{k} is a vector.}

\item{k}{An integer specifying the number of consensus clusters. Required
when \code{partitions} is a flat list of vectors (single-k input).
Ignored when \code{partitions} is a list-of-lists with names following
the pattern \code{"k2"}, \code{"k3"}, etc., as \code{k} is inferred
automatically from the names.}

\item{cluster_method}{A character string specifying the agglomeration
method passed to \code{\link[stats]{hclust}} for the consensus stage.
Accepted values are \code{"ward.D"}, \code{"ward.D2"}, \code{"single"},
\code{"complete"}, \code{"average"}, \code{"centroid"}, \code{"median"},
and \code{"mcquitty"}. Default is \code{"ward.D2"}.}

\item{consensus_method}{A character string specifying how the
co-assignment matrix is built. Options are:
\itemize{
\item \code{"classic"}: unweighted co-assignment; all partitions
contribute equally.
\item \code{"weighted_ari"}: partitions are weighted by their
pairwise ARI centrality, so more consistent partitions
contribute more to the consensus. Requires \pkg{mclust}.
}
Default is \code{"classic"}.}
}
\value{
\itemize{
\item If \code{partitions} is a flat list (single \code{k}): a named
list with the following elements:
\itemize{
\item \code{consensus_method}: character string with the method used.
\item \code{k}: integer, number of clusters requested.
\item \code{consensus}: integer vector of consensus cluster labels,
one per observation.
\item \code{coassignment}: numeric matrix of size \eqn{n \times n}
with co-assignment probabilities between 0 and 1.
\item \code{hclust}: the \code{hclust} object from the consensus
stage.
\item \code{weights}: named numeric vector of partition weights.
}
\item If \code{partitions} is a list-of-lists (multiple \code{k}):
a named list where each element corresponds to one value of
\code{k} and contains the structure described above.
}
}
\description{
Constructs a consensus clustering solution from a collection of partitions
obtained across multiple imputed datasets (as produced by
\code{\link{cluster_imputations}}). The function builds a co-assignment
matrix reflecting how frequently each pair of observations is assigned to
the same cluster, and derives the final consensus partition via hierarchical
clustering on the resulting dissimilarity. Supports both a single value of
\code{k} and a range of values.
}
\details{
The co-assignment matrix \eqn{C} is defined such that \eqn{C_{ij}}
represents the weighted proportion of partitions in which observations
\eqn{i} and \eqn{j} are assigned to the same cluster. The final
dissimilarity is computed as \eqn{1 - C} and passed to
\code{\link[stats]{hclust}}.

When \code{consensus_method = "weighted_ari"}, the weight of each
partition is proportional to its mean ARI against all other partitions,
reflecting its centrality within the ensemble. Negative centrality values
are floored at zero. If all centralities are zero, the function falls back
to equal weights with a warning.
}
\examples{
# ------------------------------------------------------------
# Example 1: Basic consensus clustering
# ------------------------------------------------------------
set.seed(123)
# simulate 3 partitions
partitions <- list(
  imp1 = sample(1:2, 10, replace = TRUE),
  imp2 = sample(1:2, 10, replace = TRUE),
  imp3 = sample(1:2, 10, replace = TRUE)
)

# Classic consensus
cons <- consensus_clustering(partitions, k = 2)
str(cons)

\donttest{
# ------------------------------------------------------------
# Example 2: consensus method with mice
# ------------------------------------------------------------
if (requireNamespace("mice", quietly = TRUE)) {

  set.seed(123)

  imp    <- mice::mice(mice::nhanes, m = 3, printFlag = FALSE)
  mild   <- as_mild_list(imp)
  parts  <- cluster_imputations(mild, method = "ward.D2", k = 3)

  # -------------------------------------------------------------
  # Single k, classic consensus
  # -------------------------------------------------------------
  cons <- consensus_clustering(parts, k = 3)

  # -------------------------------------------------------------
  # Single k, ARI-weighted consensus
 # -------------------------------------------------------------
  if (requireNamespace("mclust", quietly = TRUE)) {
    cons_w <- consensus_clustering(parts, k = 3,
                                  consensus_method = "weighted_ari")
}
  # -------------------------------------------------------------
  # Multiple k values
  # -------------------------------------------------------------
  parts_multi <- cluster_imputations(mild, method = "ward.D2", k = 2:4)
  cons_multi  <- consensus_clustering(parts_multi)
}
}

}
\references{
Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
Consensus clustering: A resampling-based method for class discovery
and visualization of gene expression microarray data.
\emph{Machine Learning}, \strong{52}(1-2), 91-118.
\doi{10.1023/A:1023949509487}

Hubert, L., & Arabie, P. (1985).
Comparing partitions.
\emph{Journal of Classification}, \strong{2}, 193-218.
\doi{10.1007/BF01908075}
}
\seealso{
\code{\link{cluster_imputations}}, \code{\link{validate_clustering}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Cluster.R
\name{cluster_imputations}
\alias{cluster_imputations}
\title{Perform clustering on multiple imputed datasets}
\usage{
cluster_imputations(
  imp_list,
  method = "ward.D2",
  k,
  scale_data = c("global", "within", "none"),
  distance = c("euclidean", "manhattan", "gower", "jaccard", "simple_matching", "custom"),
  dist_fun = NULL,
  dist_args = list(),
  ...
)
}
\arguments{
\item{imp_list}{A named list of \code{data.frame} objects, as returned by
\code{\link{as_mild_list}}. All datasets must have identical dimensions
and column names.}

\item{method}{A character string specifying the clustering algorithm.
For hierarchical clustering, accepted values are \code{"ward.D"},
\code{"ward.D2"}, \code{"single"}, \code{"complete"}, \code{"average"},
\code{"centroid"}, \code{"median"}, and \code{"mcquitty"}.
Additional options are:
\itemize{
\item \code{"kmeans"}: k-means clustering for numeric data.
\item \code{"pam"}: partitioning around medoids; supports mixed data
when used with \code{distance = "gower"}.
\item \code{"fuzzy"}: fuzzy c-means clustering.
\item \code{"mclust"}: model-based clustering via Gaussian mixtures.
\item \code{"kmodes"}: clustering for purely categorical data.
\item \code{"kprototypes"}: clustering for mixed data (numeric and
categorical).
}
Default is \code{"ward.D2"}.}

\item{k}{A single integer or an integer vector specifying the number of
clusters. If a vector is provided, clustering is performed for each
value of \code{k}.}

\item{scale_data}{Character string specifying the scaling strategy for
continuous numeric variables (binary variables coded as 0/1 are
automatically excluded from scaling). Options are:
\itemize{
\item \code{"global"}: (default) variables are scaled using pooled
mean and standard deviation computed across all imputations
without stacking them in memory (combinatorial variance formula).
Ensures comparability of distance scales between datasets and
is recommended for consensus clustering.
\item \code{"within"}: each imputed dataset is scaled independently
using its own mean and standard deviation. Useful for
exploratory analyses but may introduce inconsistencies in
distance scales across datasets.
\item \code{"none"}: no scaling is applied.
}
Scaling is automatically disabled for \code{distance = "gower"},
\code{"jaccard"}, or \code{"simple_matching"}, and for
\code{method = "kmodes"} and \code{"kprototypes"}, as these handle
variable scales internally.}

\item{distance}{A character string specifying the distance metric used for
hierarchical clustering and \code{method = "pam"}. Options are:
\itemize{
\item \code{"euclidean"}: (default) standard Euclidean distance.
Suitable for continuous numeric data after scaling.
\item \code{"manhattan"}: sum of absolute differences. Suitable for
continuous numeric data; more robust to outliers than Euclidean.
\item \code{"gower"}: Gower distance via \code{cluster::daisy()}.
Supports mixed data types (numeric, categorical, binary).
When used with hierarchical clustering, numeric variables are
normalized using global min/max computed across all imputations,
ensuring cross-imputation comparability. Requires package
\pkg{cluster}.
\item \code{"jaccard"}: Jaccard dissimilarity. Requires all variables
to be binary (0/1 or two-level factor). Requires package
\pkg{proxy}.
\item \code{"simple_matching"}: Simple Matching Coefficient (SMC).
Requires all variables to be binary. Requires package
\pkg{proxy}.
\item \code{"custom"}: user-defined distance function supplied via
\code{dist_fun}. Must return an object of class \code{dist}.
}
Ignored for \code{method = "kmeans"}, \code{"fuzzy"}, \code{"mclust"},
\code{"kmodes"}, and \code{"kprototypes"}, which compute distances
internally.

For \code{method = "pam"}, all distance options are fully supported
as PAM operates on a precomputed dissimilarity matrix.}

\item{dist_fun}{A function that takes a \code{data.frame} and returns a
\code{dist} object. Required when \code{distance = "custom"};
ignored otherwise.}

\item{dist_args}{A named list of additional arguments passed to the distance
function. For \code{distance = "gower"}, arguments are forwarded to
\code{cluster::daisy()}; for \code{distance = "jaccard"} or
\code{"simple_matching"}, to \code{proxy::dist()}; for
\code{distance = "euclidean"} or \code{"manhattan"}, to
\code{stats::dist()}; also for \code{distance = "custom"}. Default is \code{list()}.}

\item{...}{Additional arguments passed to the underlying clustering
function: \code{stats::kmeans}, \code{e1071::cmeans},
\code{mclust::Mclust}, \code{klaR::kmodes}, or
\code{clustMixType::kproto}. Not used for \code{method = "pam"}, which
operates on a precomputed dissimilarity matrix.}
}
\value{
\itemize{
\item If \code{k} is a single value: a named list of integer vectors,
one per imputed dataset, containing the cluster assignment for
each observation.
\item If \code{k} is a vector: a named list of lists, where each
top-level element corresponds to one value of \code{k} (named
\code{"k2"}, \code{"k3"}, etc.), and each inner element is a
named list of cluster assignment vectors, one per imputed dataset.
}
}
\description{
Applies a clustering algorithm to each completed dataset in a standardized
imputation list (as produced by \code{\link{as_mild_list}}). Supports
hierarchical clustering, k-means, PAM, fuzzy c-means, model-based clustering,
k-modes (for categorical data), and k-prototypes (for mixed data).
Accepts a single value of \code{k} or a range, optionally scales the data
prior to clustering, and supports multiple distance metrics including
Gower, Jaccard, Simple Matching Coefficient, and user-defined distances.
}
\details{
\strong{Scaling strategy and cross-imputation comparability:}

For methods that rely on Euclidean or Manhattan distances
(\code{"kmeans"}, \code{"fuzzy"}, \code{"mclust"}, hierarchical with
\code{distance = "euclidean"} or \code{"manhattan"}), \code{scale_data =
"global"} is strongly recommended. It computes the pooled mean and
standard deviation across all imputations using the combinatorial variance
formula — without creating a stacked copy of the data in memory — and
applies the same transformation to every imputation. This ensures that
distances are on the same scale across datasets, which is essential for
a meaningful co-assignment matrix.

Binary variables coded as 0/1 numeric are automatically excluded from
scaling because standardizing them loses their binary semantics.

\strong{Gower distance and global normalization:}

Gower distance normalizes each numeric variable by its range. When
computed independently per imputation, different imputations may produce
slightly different ranges, leading to incomparable distance scales. When
\code{distance = "gower"} is used with hierarchical clustering or PAM,
the function pre-normalizes numeric variables using global min/max
computed across all imputations (without stacking), so that Gower
distances are on a consistent scale across datasets. Categorical and
factor variables are left untouched and handled normally by
\code{cluster::daisy()}. Additional arguments to \code{cluster::daisy()}
can be passed via \code{dist_args}.

\strong{Hierarchical clustering optimization:}

When \code{method} is hierarchical and \code{k} is a vector, the
distance matrix and dendrogram are computed only once per imputed dataset
and then cut at each requested \code{k}, which avoids redundant
computation.

\strong{Data type compatibility:}
\itemize{
\item Numeric-only: hierarchical, \code{"kmeans"}, \code{"fuzzy"},
\code{"mclust"}.
\item Categorical-only: \code{"kmodes"}.
\item Mixed (numeric + categorical): \code{"pam"} with
\code{distance = "gower"} or \code{"kprototypes"}.
\item Binary-only: hierarchical with \code{distance = "jaccard"} or
\code{"simple_matching"}.
}

Character variables are internally converted to factors.
}
\examples{
# ------------------------------------------------------------
# Example 1: Numeric data, hierarchical clustering
# ------------------------------------------------------------
set.seed(123)
imp_list <- replicate(3, {
  data.frame(x = rnorm(20), y = rnorm(20))
}, simplify = FALSE)

# Single k, global scaling, euclidean distance (default)
res <- cluster_imputations(imp_list, method = "ward.D2", k = 2)
str(res)

# Range of k, manhattan distance
res_multi <- cluster_imputations(imp_list, method = "ward.D2",
                                 k = 2:4, distance = "manhattan")
str(res_multi)

# ------------------------------------------------------------
# Example 2: Mixed data, PAM with Gower distance
# ------------------------------------------------------------
set.seed(123)
imp_mixed <- replicate(3, {
  data.frame(
    x    = rnorm(20),
    grup = factor(sample(letters[1:3], 20, replace = TRUE))
  )
}, simplify = FALSE)

res_pam <- cluster_imputations(imp_mixed, method = "pam",
                               k = 2, distance = "gower")
str(res_pam)

# ------------------------------------------------------------
# Example 3: Binary data, Jaccard distance
# ------------------------------------------------------------
set.seed(123)
imp_bin <- replicate(3, {
  data.frame(
    a = sample(0:1, 20, replace = TRUE),
    b = sample(0:1, 20, replace = TRUE),
    c = sample(0:1, 20, replace = TRUE)
  )
}, simplify = FALSE)

\donttest{
if (requireNamespace("proxy", quietly = TRUE)) {
  res_jac <- cluster_imputations(imp_bin, method = "ward.D2",
                                 k = 2, distance = "jaccard")
  str(res_jac)
}

# ------------------------------------------------------------
# Example 4: Full pipeline with mice
# ------------------------------------------------------------
if (requireNamespace("mice", quietly = TRUE)) {

  df_pre       <- mice::nhanes
  df_pre$age   <- factor(df_pre$age)
  df_pre$hyp   <- factor(df_pre$hyp)

  imp  <- mice::mice(df_pre, m = 3, printFlag = FALSE, seed = 123)
  mild <- as_mild_list(imp)

  # PAM + Gower with global range normalization
  parts <- cluster_imputations(mild, method = "pam",
                               k = 2:4, distance = "gower")
  str(parts)
}
}

}
\references{
Kaufman, L., & Rousseeuw, P. J. (1990).
\emph{Finding Groups in Data: An Introduction to Cluster Analysis}. Wiley.
\doi{10.1002/9780470316801}

Bezdek, J. C. (1981).
\emph{Pattern Recognition with Fuzzy Objective Function Algorithms}.
Plenum Press.

Huang, Z. (1998).
Extensions to the k-means algorithm for clustering large data sets
with categorical values.
\emph{Data Mining and Knowledge Discovery}, \strong{2}(3), 283-304.
\doi{10.1023/A:1009769707641}

Gower, J. C. (1971).
A general coefficient of similarity and some of its properties.
\emph{Biometrics}, \strong{27}(4), 857-871.
\doi{10.2307/2528823}

Scrucca, L., Fop, M., Murphy, T. B., & Raftery, A. E. (2016).
mclust 5: Clustering, classification and density estimation using
Gaussian finite mixture models.
\emph{The R Journal}, \strong{8}(1), 289-317.
\doi{10.32614/RJ-2016-021}
}
\seealso{
\code{\link{as_mild_list}}, \code{\link{consensus_clustering}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Complete_Sequence.R
\name{run_mi_clustering}
\alias{run_mi_clustering}
\title{Run the full multiple-imputation clustering pipeline}
\usage{
run_mi_clustering(
  data,
  method = "ward.D2",
  k,
  scale_data = "global",
  distance = "euclidean",
  dist_fun = NULL,
  dist_args = list(),
  consensus_method = c("classic", "weighted_ari"),
  cluster_method_consensus = "ward.D2",
  pac_lower = 0.1,
  pac_upper = 0.9,
  weights = NULL,
  prefer_stability = TRUE,
  tie_breaker = c("silhouette", "pac", "dunn", "ch", "db", "ari_between",
    "ari_consensus"),
  ...
)
}
\arguments{
\item{data}{An imputation object accepted by \code{\link{as_mild_list}}.
Supported formats include \code{mids} objects from \pkg{mice},
long-format data frames with a \code{.imp} column, \code{amelia} objects,
\code{imputationList} objects from \pkg{mitools}, and plain lists of
\code{data.frame}s.}

\item{method}{A character string specifying the clustering algorithm applied
to each imputed dataset. For hierarchical clustering, accepted values are
\code{"ward.D"}, \code{"ward.D2"}, \code{"single"}, \code{"complete"},
\code{"average"}, \code{"centroid"}, \code{"median"}, and
\code{"mcquitty"}. Additional options are \code{"kmeans"}, \code{"pam"},
\code{"fuzzy"}, \code{"mclust"}, \code{"kmodes"}, and
\code{"kprototypes"}. Default is \code{"ward.D2"}.}

\item{k}{A single integer or an integer vector specifying the number(s) of
clusters to evaluate. If a single value is supplied, the pipeline skips
the \code{\link{choose_best_clustering}} step and returns that solution
directly. If a vector is supplied, all values are evaluated and the
optimal \code{k} is selected via weighted rank aggregation.}

\item{scale_data}{A character string specifying how numeric columns are
standardized prior to clustering. Accepted values are \code{"global"}
(default, pooled mean and standard deviation computed across all
imputations without stacking), \code{"within"} (each imputed dataset
scaled independently), and \code{"none"} (no scaling applied). Scaling
is automatically disabled for \code{distance = "gower"},
\code{"jaccard"}, or \code{"simple_matching"}, and for
\code{method = "kmodes"} and \code{"kprototypes"}.}

\item{distance}{A character string specifying the distance metric used for
hierarchical clustering and \code{method = "pam"}. Options are
\code{"euclidean"} (default), \code{"manhattan"}, \code{"gower"}
(via \code{cluster::daisy()}, supports mixed data), \code{"jaccard"}
(binary data, requires \pkg{proxy}), \code{"simple_matching"} (binary
data, requires \pkg{proxy}), and \code{"custom"} (user-defined function
via \code{dist_fun}). Ignored for \code{method = "kmeans"},
\code{"fuzzy"}, \code{"mclust"}, \code{"kmodes"}, and
\code{"kprototypes"}, which compute distances internally.}

\item{dist_fun}{A function that takes a \code{data.frame} and returns a
\code{dist} object. Required when \code{distance = "custom"};
ignored otherwise.}

\item{dist_args}{A named list of additional arguments passed to the
distance function. For \code{distance = "gower"}, forwarded to
\code{cluster::daisy()}; for \code{"jaccard"} or
\code{"simple_matching"}, to \code{proxy::dist()}; for
\code{"euclidean"} or \code{"manhattan"}, to \code{stats::dist()};
also for \code{"custom"}. Default is \code{list()}.}

\item{consensus_method}{A character string specifying the method used to
build the co-assignment matrix. \code{"classic"} assigns equal weight to
all partitions; \code{"weighted_ari"} weights partitions by their
pairwise ARI centrality. Default is \code{"classic"}.}

\item{cluster_method_consensus}{A character string specifying the
agglomeration method passed to \code{\link[stats]{hclust}} during the
consensus stage. Accepted values are the same as for \code{method} when
using hierarchical clustering. Default is \code{"ward.D2"}.}

\item{pac_lower}{A numeric value between 0 and 1 specifying the lower bound
of the ambiguous assignment region for the PAC metric. Default is
\code{0.1}.}

\item{pac_upper}{A numeric value between 0 and 1 specifying the upper bound
of the ambiguous assignment region for the PAC metric. Default is
\code{0.9}.}

\item{weights}{A named numeric vector specifying the relative importance of
each metric in the rank aggregation used by
\code{\link{choose_best_clustering}}. Expected names are \code{pac},
\code{silhouette}, \code{ari_between}, \code{ari_consensus}, \code{ch},
\code{db}, and \code{dunn}. If \code{NULL} (default), weights are set
automatically based on \code{prefer_stability}. Ignored when \code{k} is
a single value.}

\item{prefer_stability}{Logical. If \code{TRUE} (default), stability
metrics (\code{pac}, \code{ari_between}, \code{ari_consensus}) receive
higher weights during \code{k} selection. If \code{FALSE}, internal
compactness metrics (\code{silhouette}, \code{ch}, \code{db},
\code{dunn}) receive higher weights. Ignored when \code{k} is a single
value or when \code{weights} is provided explicitly.}

\item{tie_breaker}{A character string specifying the secondary metric used
to break ties during \code{k} selection. One of \code{"silhouette"}
(default), \code{"pac"}, \code{"dunn"}, \code{"ch"}, \code{"db"},
\code{"ari_between"}, or \code{"ari_consensus"}. Ignored when \code{k}
is a single value.}

\item{...}{Additional arguments passed to the underlying clustering
function via \code{\link{cluster_imputations}}: \code{stats::kmeans},
\code{e1071::cmeans}, \code{mclust::Mclust}, \code{klaR::kmodes}, or
\code{clustMixType::kproto}. Not used for \code{method = "pam"}.}
}
\value{
An object of class \code{"mi_clustering_result"}: a named list
with the following elements:
\itemize{
\item \code{call}: the matched call.
\item \code{input_k}: the value(s) of \code{k} supplied by the user.
\item \code{clustering_method}: character string with the clustering
method used.
\item \code{consensus_method}: character string with the consensus
method used.
\item \code{scale_data}: A character string specifying how numeric columns
are standardized prior to clustering
\item \code{imputations}: standardized list of imputed datasets, as
returned by \code{\link{as_mild_list}}.
\item \code{partitions}: clustering assignments per imputed dataset,
as returned by \code{\link{cluster_imputations}}.
\item \code{consensus_results}: consensus clustering output, as
returned by \code{\link{consensus_clustering}}.
\item \code{validation_table}: validation metrics per \code{k}, as
returned by \code{\link{validate_clustering}}.
\item \code{best_k}: integer, the selected optimal number of clusters.
\item \code{best_consensus}: integer vector of consensus cluster labels
for the selected \code{k}, one per observation.
\item \code{best_coassignment}: numeric matrix of co-assignment
probabilities for the selected \code{k}.
\item \code{best_consensus_result}: the full consensus result object
for the selected \code{k}.
\item \code{scores_table}: validation table with an additional
\code{score} column when \code{k} is a vector; the plain
validation table when \code{k} is a single value.
\item \code{selection}: the full output of
\code{\link{choose_best_clustering}} when \code{k} is a vector,
or a minimal equivalent list when \code{k} is a single value.
}
}
\description{
Executes the complete multiple-imputation clustering pipeline in a single
call, integrating imputation standardization, per-imputation clustering,
consensus construction, validation, and optimal \code{k} selection.
Designed as a convenience wrapper around \code{\link{as_mild_list}},
\code{\link{cluster_imputations}}, \code{\link{consensus_clustering}},
\code{\link{validate_clustering}}, and \code{\link{choose_best_clustering}}.
When a single value of \code{k} is supplied, the selection step is skipped
and the unique solution is returned directly.
}
\details{
The pipeline proceeds as follows:
\enumerate{
\item \code{\link{as_mild_list}} standardizes the imputation object and
validates data quality.
\item \code{\link{cluster_imputations}} applies the chosen algorithm to
each imputed dataset, optionally scaling the data beforehand.
\item \code{\link{consensus_clustering}} builds a co-assignment matrix
and derives the consensus partition for each requested \code{k}.
\item \code{\link{validate_clustering}} computes internal and stability
metrics for each candidate \code{k}.
\item \code{\link{choose_best_clustering}} selects the optimal \code{k}
via weighted rank aggregation (only when \code{length(k) > 1}).
}

When \code{length(k) == 1}, step 5 is bypassed and the unique solution
is returned directly, with \code{weights} and \code{tie_breaker} set to
\code{NULL} in the \code{selection} element.
}
\examples{
# ------------------------------------------------------------
# Example 1: Simulated list of imputed datasets
# ------------------------------------------------------------
set.seed(123)
make_imputation <- function() {
  data.frame(
    x1 = rnorm(30),
    x2 = rnorm(30),
    x3 = rnorm(30)
  )
}

imp_list <- lapply(1:5, function(i) make_imputation())

# Single k
res <- run_mi_clustering(imp_list, method = "ward.D2", k = 3)
res$best_k
res$best_consensus

# Range of k values (automatic selection)
res_multi <- run_mi_clustering(imp_list, method = "ward.D2", k = 2:4)
res_multi$best_k
res_multi$validation_table

\donttest{
# ------------------------------------------------------------
# Example 2: Full pipeline with mice
# ------------------------------------------------------------
if (requireNamespace("mice", quietly = TRUE)) {

set.seed(123)
imp <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)

# Single k
res <- run_mi_clustering(imp, method = "ward.D2", k = 3)
res$best_k
res$best_consensus

# Range of k values (automatic selection)
res_multi <- run_mi_clustering(imp, method = "ward.D2", k = 2:5)
res_multi$best_k
res_multi$validation_table

# PAM with Gower distance (mixed data)
res_gower <- run_mi_clustering(imp, method = "pam", k = 2:4,
                               distance = "gower")

# ARI-weighted consensus, compactness-focused selection
res_w <- run_mi_clustering(imp, method = "ward.D2", k = 2:5,
                           consensus_method  = "weighted_ari",
                           prefer_stability  = FALSE,
                           tie_breaker       = "dunn")
}
}

}
\references{
Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
Consensus clustering: A resampling-based method for class discovery
and visualization of gene expression microarray data.
\emph{Machine Learning}, \strong{52}(1-2), 91-118.
\doi{10.1023/A:1023949509487}

Rubin, D. B. (1987).
\emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.

Pihur, V., Datta, S., & Datta, S. (2007).
Weighted rank aggregation of cluster validation measures.
\emph{Bioinformatics}, \strong{23}(13), 1607-1615.
\doi{10.1093/bioinformatics/btm158}
}
\seealso{
\code{\link{as_mild_list}}, \code{\link{cluster_imputations}},
\code{\link{consensus_clustering}}, \code{\link{validate_clustering}},
\code{\link{choose_best_clustering}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Validation.R
\name{validate_clustering}
\alias{validate_clustering}
\title{Validate consensus clustering results across multiple imputed datasets}
\usage{
validate_clustering(
  partitions,
  consensus_results,
  pac_lower = 0.1,
  pac_upper = 0.9
)
}
\arguments{
\item{partitions}{A list of cluster assignment vectors (one per imputed
dataset) for a single \code{k}, or a named list of such lists (one per
\code{k} value, named \code{"k2"}, \code{"k3"}, etc.) as returned by
\code{\link{cluster_imputations}}.}

\item{consensus_results}{A list as returned by
\code{\link{consensus_clustering}}, either for a single \code{k} or
for multiple \code{k} values. Must contain at least \code{$consensus}
and \code{$coassignment} for each \code{k}.}

\item{pac_lower}{A numeric value between 0 and 1 specifying the lower
bound of the ambiguous assignment region for the PAC metric. Default
is \code{0.1}.}

\item{pac_upper}{A numeric value between 0 and 1 specifying the upper
bound of the ambiguous assignment region for the PAC metric. Default
is \code{0.9}.}
}
\value{
A \code{data.frame} with one row per evaluated \code{k} and the
following columns:
\itemize{
\item \code{k}: number of clusters evaluated.
\item \code{pac}: Proportion of Ambiguous Clusterings. Lower values
indicate a more stable consensus. Computed from the upper
triangle of the co-assignment matrix.
\item \code{silhouette_mean}: mean silhouette width computed on the
consensus dissimilarity \eqn{1 - C}. Higher values indicate
better-defined clusters.
\item \code{ari_mean_between_imputations}: mean pairwise Adjusted
Rand Index (ARI) across all pairs of imputed partitions.
Higher values indicate greater clustering stability.
\item \code{ari_consensus_mean}: mean ARI between the consensus
partition and each individual imputed partition. Higher values
indicate that the consensus is representative of the ensemble.
\item \code{calinski_harabasz_mean}: Calinski-Harabasz index computed
via \code{fpc::cluster.stats()}. Higher values indicate more
compact and well-separated clusters.
\item \code{davies_bouldin_mean}: Davies-Bouldin index computed using
a medoid-based approach on the consensus dissimilarity. Lower
values indicate better cluster separation.
\item \code{dunn_index}: Dunn index computed via
\code{fpc::cluster.stats()}. Higher values indicate better
cluster compactness and separation.
}
}
\description{
Computes a comprehensive set of internal and stability validation metrics
for consensus clustering results obtained from multiple imputed datasets.
All internal metrics are computed on the consensus dissimilarity matrix
\eqn{1 - C}, where \eqn{C} is the co-assignment matrix, making the
function robust to numeric, categorical, and mixed data types. Supports
both single-k and multi-k evaluation.
}
\details{
All internal validation metrics (silhouette, Calinski-Harabasz,
Davies-Bouldin, Dunn) are computed exclusively on the consensus
dissimilarity matrix \eqn{1 - C} rather than on the original feature
space. This design choice ensures compatibility with any data type and
avoids recomputing distances from the raw imputed datasets.

The Davies-Bouldin index is computed using a custom medoid-based
implementation derived from the consensus dissimilarity, as no standard
R implementation accepts a precomputed dissimilarity matrix directly.

Helper functions prefixed with a dot (e.g., \code{.pac_from_coassignment},
\code{.mean_silhouette_from_diss}) are internal and not exported.
}
\examples{
# ------------------------------------------------------------
# Example 1: Basic validation with simulated partitions
# ------------------------------------------------------------
set.seed(123)

# Simulate partitions (3 imputations, k = 2)
parts <- list(
  imp1 = sample(1:2, 20, replace = TRUE),
  imp2 = sample(1:2, 20, replace = TRUE),
  imp3 = sample(1:2, 20, replace = TRUE)
)

cons <- consensus_clustering(parts, k = 2)
val  <- validate_clustering(parts, cons)

print(val)

\donttest{
# ------------------------------------------------------------
# Example 2: Full pipeline with mice (multiple k)
# ------------------------------------------------------------
if (requireNamespace("mice", quietly = TRUE)) {

  set.seed(123)

  imp  <- mice::mice(mice::nhanes, m = 3, printFlag = FALSE)
  mild <- as_mild_list(imp)

  parts <- cluster_imputations(mild, method = "ward.D2", k = 2:3)
  cons  <- consensus_clustering(parts)

  val <- validate_clustering(parts, cons)

  print(val)
}
}

}
\references{
Senbabaoglu, Y., Michailidis, G., & Li, J. Z. (2014).
Critical limitations of consensus clustering in class discovery.
\emph{Scientific Reports}, \strong{4}, 6207.
\doi{10.1038/srep06207}

Rousseeuw, P. J. (1987).
Silhouettes: A graphical aid to the interpretation and validation
of cluster analysis.
\emph{Journal of Computational and Applied Mathematics}, \strong{20}, 53-65.
\doi{10.1016/0377-0427(87)90125-7}

Calinski, T., & Harabasz, J. (1974).
A dendrite method for cluster analysis.
\emph{Communications in Statistics}, \strong{3}, 1-27.
\doi{10.1080/03610927408827101}

Davies, D. L., & Bouldin, D. W. (1979).
A cluster separation measure.
\emph{IEEE Transactions on Pattern Analysis and Machine Intelligence},
\strong{PAMI-1}, 224-227.
\doi{10.1109/TPAMI.1979.4766909}

Dunn, J. C. (1974).
Well-separated clusters and optimal fuzzy partitions.
\emph{Journal of Cybernetics}, \strong{4}, 95-104.
\doi{10.1080/01969727408546059}

Hubert, L., & Arabie, P. (1985).
Comparing partitions.
\emph{Journal of Classification}, \strong{2}, 193-218.
\doi{10.1007/BF01908075}
}
\seealso{
\code{\link{consensus_clustering}}, \code{\link{choose_best_clustering}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ValidationK.R
\name{plot_validation_metrics}
\alias{plot_validation_metrics}
\title{Plot validation metrics across candidate numbers of clusters}
\usage{
plot_validation_metrics(
  validation_table,
  metrics = c("pac", "silhouette_mean", "ari_mean_between_imputations",
    "ari_consensus_mean", "calinski_harabasz_mean", "davies_bouldin_mean", "dunn_index",
    "score")
)
}
\arguments{
\item{validation_table}{A \code{data.frame} as returned by
\code{\link{validate_clustering}} or \code{$scores_table} from
\code{\link{choose_best_clustering}}. Must contain a column \code{k}
and at least one of the requested metric columns.}

\item{metrics}{A character vector specifying which metrics to plot.
Only metrics present in \code{validation_table} are plotted; others
are silently ignored. Default includes all standard metrics plus
\code{score}:
\code{"pac"}, \code{"silhouette_mean"},
\code{"ari_mean_between_imputations"}, \code{"ari_consensus_mean"},
\code{"calinski_harabasz_mean"}, \code{"davies_bouldin_mean"},
\code{"dunn_index"}, and \code{"score"}.}
}
\value{
Invisibly returns \code{NULL}. The function is called for its
side effect of producing a plot.
}
\description{
Produces a multi-panel line plot displaying one or more clustering
validation metrics as a function of the number of clusters \code{k},
as computed by \code{\link{validate_clustering}} and optionally extended
with a \code{score} column by \code{\link{choose_best_clustering}}. For
each metric, the optimal value is highlighted in red.
}
\details{
The number of panels is determined automatically from the number of
available metrics using \code{ceiling(sqrt(n))} rows and
\code{ceiling(n / nrow)} columns, producing a layout as square as
possible.

The direction of optimality is accounted for when highlighting the best
value: lower is better for \code{pac}, \code{davies_bouldin_mean}, and
\code{score}; higher is better for all remaining metrics. The optimal
point for each metric is highlighted in red.

Original graphical parameters are restored on exit via
\code{\link[graphics]{par}}.

When \code{score} is present (i.e., \code{validation_table} comes from
\code{\link{choose_best_clustering}}), the panel for \code{score} shows
the overall weighted rank aggregation - the \code{k} with the lowest
score is the recommended solution.
}
\examples{
# ------------------------------------------------------------
# Example 1: Validation metrics with simulated data
# ------------------------------------------------------------
val_table <- data.frame(
  k                            = 2:4,
  pac                          = c(0.15, 0.08, 0.20),
  silhouette_mean              = c(0.42, 0.61, 0.38),
  ari_mean_between_imputations = c(0.80, 0.91, 0.75),
  ari_consensus_mean           = c(0.82, 0.93, 0.77),
  calinski_harabasz_mean       = c(120,  198,  105),
  davies_bouldin_mean          = c(0.85, 0.54, 0.97),
  dunn_index                   = c(0.30, 0.48, 0.27)
)

# Plot all metrics
plot_validation_metrics(val_table)

# Plot only stability metrics
plot_validation_metrics(val_table,
                        metrics = c("pac",
                                    "ari_mean_between_imputations",
                                    "ari_consensus_mean"))
\donttest{
# ------------------------------------------------------------
# Example 2: metrics validation with mice
# ------------------------------------------------------------

if (requireNamespace("mice", quietly = TRUE)) {

set.seed(123)
imp    <- mice::mice(mice::nhanes, m = 5, printFlag = FALSE)
mild   <- as_mild_list(imp)
parts  <- cluster_imputations(mild, method = "ward.D2", k = 2:5)
cons   <- consensus_clustering(parts)
val    <- validate_clustering(parts, cons)

# Plot all available metrics (no score column yet)
plot_validation_metrics(val)

# Include score column from choose_best_clustering
best <- choose_best_clustering(val, cons)
plot_validation_metrics(best)

# Plot only stability metrics
plot_validation_metrics(val,
                        metrics = c("pac",
                                    "ari_mean_between_imputations",
                                    "ari_consensus_mean"))
}
}

}
\seealso{
\code{\link{validate_clustering}}, \code{\link{choose_best_clustering}},
\code{\link{plot_consensus_dendrogram}}, \code{\link{plot_consensus_matrix}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cclustr-package.R
\docType{package}
\name{cclustr-package}
\alias{cclustr}
\alias{cclustr-package}
\title{cclustr: Consensus Clustering Methods for Multiple Imputed Data}
\description{
The \pkg{cclustr} package provides a comprehensive framework for performing
clustering analysis on datasets treated with multiple imputation. It
addresses the uncertainty inherent in missing data by finding a stable
consensus partition across multiple imputed realizations.
}
\details{
Standard statistical workflows for imputed data often focus on supervised
learning or univariate statistics. \pkg{cclustr} extends this to
unsupervised learning by implementing a co-assignment matrix approach.
}
\section{Main Workflow}{

To achieve a robust consensus solution, follow these steps:
\enumerate{
\item \strong{Standardization:} Use \code{\link{as_mild_list}} to format
imputed data (e.g., from \pkg{mice}) into a consistent structure.
\item \strong{Clustering:} Apply \code{\link{cluster_imputations}} to run
algorithms (K-means, PAM, K-prototypes, etc.) across all imputations.
\item \strong{Consensus:} Generate the final partition using
\code{\link{consensus_clustering}} based on co-assignment probabilities.
\item \strong{Selection:} Evaluate model quality and optimal cluster
counts with \code{\link{choose_best_clustering}}.
}
}

\section{Supported Data Types}{

\itemize{
\item \strong{Numeric:} Hierarchical clustering, K-means, Fuzzy C-means, and Mclust.
\item \strong{Categorical:} K-modes.
\item \strong{Mixed:} PAM with Gower distance or K-prototypes.
}
}

\references{
Monti, S., Tamayo, P., Mesirov, J., & Golub, T. (2003).
Consensus clustering: A resampling-based method for class discovery
and visualization of gene expression microarray data.
\emph{Machine Learning}, \strong{52}(1-2), 91-118.
\doi{10.1023/A:1023949509487}

Senbabaoglu, Y., Michailidis, G., & Li, J. Z. (2014).
Critical limitations of consensus clustering in class discovery.
\emph{Scientific Reports}, \strong{4}, 6207.
\doi{10.1038/srep06207}

Pihur, V., Datta, S., & Datta, S. (2007).
Weighted rank aggregation of cluster validation measures.
\emph{Bioinformatics}, \strong{23}(13), 1607-1615.
\doi{10.1093/bioinformatics/btm158}

Rubin, D. B. (1987).
\emph{Multiple Imputation for Nonresponse in Surveys}. Wiley.

van Buuren, S., & Groothuis-Oudshoorn, K. (2011).
mice: Multivariate Imputation by Chained Equations in R.
\emph{Journal of Statistical Software}, \strong{45}(3), 1-67.
\doi{10.18637/jss.v045.i03}

Hubert, L., & Arabie, P. (1985).
Comparing partitions.
\emph{Journal of Classification}, \strong{2}, 193-218.
\doi{10.1007/BF01908075}

Rousseeuw, P. J. (1987).
Silhouettes: A graphical aid to the interpretation and validation
of cluster analysis.
\emph{Journal of Computational and Applied Mathematics}, \strong{20}, 53-65.
\doi{10.1016/0377-0427(87)90125-7}

Calinski, T., & Harabasz, J. (1974).
A dendrite method for cluster analysis.
\emph{Communications in Statistics}, \strong{3}, 1-27.
\doi{10.1080/03610927408827101}

Davies, D. L., & Bouldin, D. W. (1979).
A cluster separation measure.
\emph{IEEE Transactions on Pattern Analysis and Machine Intelligence},
\strong{PAMI-1}, 224-227.
\doi{10.1109/TPAMI.1979.4766909}

Dunn, J. C. (1974).
Well-separated clusters and optimal fuzzy partitions.
\emph{Journal of Cybernetics}, \strong{4}, 95-104.
\doi{10.1080/01969727408546059}
}
\seealso{
Useful links:
\itemize{
\item \url{https://github.com/andrews06ml/cclustr}
\item Report bugs at \url{https://github.com/andrews06ml/cclustr/issues}
}
}
\author{
\strong{Maintainer}: Andres Montenegro Lemus \email{andresfemole@gmail.com}

Authors:
\itemize{
\item Anhuar Duran Mendoza \email{asteyduran15@gmail.com}
\item Mario Pacheco Lopez \email{mariopachecolopez@gmail.com}
}
}
\keyword{internal}
